/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#ifndef __CONTAINER_HASHTABLE_H
#define __CONTAINER_HASHTABLE_H

#include "container.allocator.h"

namespace magnum { namespace container {
using magnum::object::Object;
using magnum::types::String;


/*----------------------------------------------------------------------------
The hasher must be a class with the following members:
	class Hasher {
	public:
		uint32		getMax() const;
		uint32		getHash(const Key& key) const;
	};
It is adviseable to use inline implementations for these
methods, especially for the hasher function. The results
of getHash must be smaller or equal to getMax():
	0 <= getHash() <= getMax()
----------------------------------------------------------------------------*/
template<class Key, class Item, class Hasher> class Hashtable : public Object {
DeclareClass(Hashtable, Object);
private:
	struct Node {
		Key			m_Key;
		Item		m_Item;
		Node*		m_Next;
	};

	Allocator<Node> m_Allocator;
	Hasher			m_Hasher;
	Node**			m_Table;
	unsigned int	m_TableSize;

public:
	class Iterator : public IteratorBase<Iterator> {
	public:
		inline Iterator();
		inline				operator Item*();
		inline				operator const Item*() const;
		inline Item&		operator*() const;
		inline Item*		operator->() const;
		inline Iterator&	operator++();
		inline Iterator		operator++(int );
		inline Iterator&	operator--();
		inline Iterator		operator--(int );
		inline bool 		operator==(const Iterator& iter) const;
		inline bool 		operator!=(const Iterator& iter) const;
		inline bool			operator<(const Iterator& iter) const;
		inline const Item&	item() const;
		inline const Key&	key() const;


	};

public:
	inline Hashtable();
	inline ~Hashtable();

	void 				clear();
	bool		 		contains(const Key& key) const;
	void 				insert(const Item& item,const Key& key);
	void 				remove(const Key& key);
	void				remove(const Iterator& );
	
	inline unsigned int	getSize() const;
	inline bool			isEmpty() const;
	
	inline Iterator		begin() const;
	inline Iterator		at(const Key& ) const;
	inline Iterator		end() const;

	const Item&			get(const Key& key) const;
	Item&				get(const Key& key);
	bool				set(const Key&, const Item& elem);

	inline Item&		operator [](const Key& );
	inline const Item&	operator [](const Key& ) const;
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class Key, class Item, class Hasher> class HashtableEnumerator : public Object {
DeclareClass(HashtableEnumerator, Object);
private:
	const Hashtable<Key, Item, Hasher>*		m_Hash;
	unsigned int							m_Index;
	unsigned int							m_TableSize;
	typename Hashtable<Key, Item, Hasher>::Node**	m_Table;
	typename Hashtable<Key, Item, Hasher>::Node*		m_Node;

public:
	inline HashtableEnumerator(const Hashtable<Key, Item, Hasher> *hash);
	inline ~HashtableEnumerator();

	inline void setHash(const Hashtable<Key, Item, Hasher> *hash);
	inline const Hashtable<Key, Item, Hasher> *getHash() const;

	inline bool getFirstItem(Item& );
	inline bool getNextItem(Item& );
	inline bool getCurrentItem(Item& );

	inline Item* getFirstItem();
	inline Item* getNextItem();
	inline Item* getCurrentItem();

	inline bool getFirstKey(Key& );
	inline bool getNextKey(Key& );
	inline bool getCurrentKey(Key& );

	inline const Key* getFirstKey();
	inline const Key* getNextKey();
	inline const Key* getCurrentKey();
};

}; };
#endif
