/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#ifndef __CONTAINER_PRIORITYHEAP_INL
#define __CONTAINER_PRIORITYHEAP_INL

namespace magnum { namespace container {


/*----------------------------------------------------------------------------------
Description:
----------------------------------------------------------------------------------*/
template<class T, class P> 
inline PriorityHeap<T,P>::PriorityHeap() {
	m_Heap.insert(&m_Terminator);
	m_Terminator.m_Index = 0;
}


/*----------------------------------------------------------------------------------
Description:
----------------------------------------------------------------------------------*/
template<class T, class P> 
inline PriorityHeap<T,P>::~PriorityHeap() {
	clear();
}


/*----------------------------------------------------------------------------------
Description:
----------------------------------------------------------------------------------*/
template<class T, class P> 
inline bool PriorityHeap<T,P>::isEmpty() const {
	return (m_Heap.getSize() == 1);
}


/*----------------------------------------------------------------------------------
Description:
----------------------------------------------------------------------------------*/
template<class T, class P> 
inline unsigned int PriorityHeap<T,P>::getSize() const {
	return m_Heap.getSize() - 1;
}


/*----------------------------------------------------------------------------------
Description:
----------------------------------------------------------------------------------*/
template<class T, class P> 
void PriorityHeap<T,P>::clear() {
	unsigned int i;
	for(i = 0; i < m_Heap.getSize() - 1; i++) {
		operator delete(m_Heap[i], m_Allocator);
	}

	m_Heap.clear();
	m_Heap.insert(&m_Terminator);
	m_Terminator.m_Index = 0;
}


/*----------------------------------------------------------------------------------
Description:
----------------------------------------------------------------------------------*/
template<class T, class P> 
inline typename PriorityHeap<T,P>::Iterator PriorityHeap<T,P>::insert(const T& item, const P& prio) {
	Item* nitem = new(m_Allocator) Item(item, prio);
	unsigned int size = m_Heap.getSize() - 1;
	m_Heap[size] = nitem;
	nitem->m_Index = size;
	m_Heap.insert(&m_Terminator);
	m_Terminator.m_Index = size + 1;
	repair_up(size);
	return Iterator(&m_Heap, nitem);
}


/*----------------------------------------------------------------------------------
Description:
----------------------------------------------------------------------------------*/
template<class T, class P> 
inline void PriorityHeap<T,P>::change(const typename PriorityHeap<T,P>::Iterator& iter, const P& prio) {
	if (iter.m_Item->m_Priority < prio) {
		iter.m_Item->m_Priority = prio;
		repair_up(iter.m_Item->m_Index);
	} else {
		iter.m_Item->m_Priority = prio;
		repair_down(iter.m_Item->m_Index);
	}
}


/*----------------------------------------------------------------------------------
Description:
----------------------------------------------------------------------------------*/
template<class T, class P> 
T PriorityHeap<T,P>::remove_idx(unsigned int i) {
	ASSERT(i < m_Heap.getSize() - 1);

	T ritem = m_Heap[i]->m_Item;
	unsigned int last = m_Heap.getSize() - 2; 
	operator delete(m_Heap[i], m_Allocator);
	Item* v = m_Heap[last];
	
	unsigned int j;
	for(;;) {
		j = i+i+1;
		if (j > last)
			break;
		if ((j < last) && (m_Heap[j]->m_Priority < m_Heap[j+1]->m_Priority))
			j++;
		ASSERT(m_Heap[j]->m_Index == j);
		if (v->m_Priority >= m_Heap[j]->m_Priority)
			break;
		m_Heap[i] = m_Heap[j];
		m_Heap[i]->m_Index = i;
		i = j;
	}
	m_Heap[i] = v;
	v->m_Index = i;

	m_Heap.remove(last + 1);
	m_Heap[last] = &m_Terminator;
	m_Terminator.m_Index = last;

	return ritem;
}


/*----------------------------------------------------------------------------------
Description:
----------------------------------------------------------------------------------*/
template<class T, class P> 
inline void PriorityHeap<T,P>::remove(const typename PriorityHeap<T,P>::Iterator& item) {
	remove_idx(item.m_Item->m_Index);
}


/*----------------------------------------------------------------------------------
Description:
----------------------------------------------------------------------------------*/
template<class T, class P> 
inline T PriorityHeap<T,P>::remove() {
	return remove_idx(0);
}


/*----------------------------------------------------------------------------------
Description:
----------------------------------------------------------------------------------*/
template<class T, class P> 
inline const T& PriorityHeap<T,P>::peek() const {
	ASSERT(m_Heap.getSize() > 1);
	return m_Heap[0]->m_Item;
}


/*----------------------------------------------------------------------------------
Description:
----------------------------------------------------------------------------------*/
template<class T, class P> 
inline bool PriorityHeap<T,P>::contains(const T& item) const {
	for (unsigned int i = 0; i < m_Heap.getSize() - 1; i++)
		if (m_Heap[i]->m_Item == item)
			return true;
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class P> 
inline typename PriorityHeap<T,P>::Iterator PriorityHeap<T,P>::begin() const {
	return Iterator(&m_Heap, m_Heap[0]);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class P> 
inline typename PriorityHeap<T,P>::Iterator PriorityHeap<T,P>::end() const {
	return Iterator(&m_Heap, (Item*)&m_Terminator);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, class P> 
inline typename PriorityHeap<T,P>::Iterator PriorityHeap<T,P>::top() const {
	return Iterator(&m_Heap, m_Heap[0]);
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
template<class T, class P> 
inline typename PriorityHeap<T,P>::Iterator PriorityHeap<T,P>::at(const T& item) const {
	for (unsigned int i = 0; i < m_Heap.getSize() - 1; i++)
		if (m_Heap[i]->m_Item == item)
			return Iterator(&m_Heap, m_Heap[i]);
	return Iterator(&m_Heap, &m_Terminator);;
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
template<class T, class P> 
inline bool PriorityHeap<T,P>::operator==(const PriorityHeap<T,P>& heap) const {
	if (m_Size != heap.m_Size)
		return false;
	ASSERT(false);
	return false;
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
template<class T, class P> 
inline bool PriorityHeap<T,P>::operator!=(const PriorityHeap<T,P>& heap) const {
	if (m_Size != set.m_Size)
		return true;
	ASSERT(false);
	return false;
}


/*----------------------------------------------------------------------------------
Description:
----------------------------------------------------------------------------------*/
template<class T, class P> 
void PriorityHeap<T,P>::repair_up(unsigned int i) {
	ASSERT(i < m_Heap.getSize() - 1);

	Item* item = m_Heap[i];
	unsigned int j = (i-1)/2;

	while((i > 0) && (m_Heap[j]->m_Priority < item->m_Priority)) {
		ASSERT(m_Heap[j]->m_Index == j);
		m_Heap[i] = m_Heap[j];
		m_Heap[i]->m_Index = i;
		i = j;
		j = (i-1)/2;
	}

	m_Heap[i] = item;
	item->m_Index = i;
}


/*----------------------------------------------------------------------------------
Description:
----------------------------------------------------------------------------------*/
template<class T, class P> 
void PriorityHeap<T,P>::repair_down(unsigned int i) {
	ASSERT(i < m_Heap.getSize() - 1);

	Item* v = m_Heap[i];
	unsigned int last = m_Heap.getSize() - 2;
	unsigned int j;

	for(;;) {
		j = i+i+1;
		if (j > last)
			break;
		if ((j < last) && (m_Heap[j]->m_Priority < m_Heap[j+1]->m_Priority))
			j++;
		ASSERT(m_Heap[j]->m_Index == j);
		if (v->m_Priority >= m_Heap[j]->m_Priority)
			break;
		m_Heap[i] = m_Heap[j];
		m_Heap[i]->m_Index = i;
		i = j;
	}

	m_Heap[i] = v;
	v->m_Index = i;
}

}; };

#endif
