/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#ifndef __GEOMETRY_BOUNDS_H
#define __GEOMETRY_BOUNDS_H

#include "object.h"
#include "math.matrix.h"
#include "geometry.sphere.h"
#include "geometry.box.h"
#include "geometry.cube.h"
#include "geometry.frustum.h"
#include "geometry.alignedbox.h"
#include "geometry.transform.h"

namespace magnum { namespace geometry {
using magnum::types::String;
using magnum::object::Object;
using magnum::math::Matrix;


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> class Bounds : public Object {
DeclareClass(Bounds, Object);
public:
	enum Type {
		BoundingVoid		= 0,
		BoundingInfinity	= 0x00000001,
		BoundingSphere		= 0x00000002,
		BoundingBox			= 0x00000004,
		AlignedBoundingBox	= 0x00000008,
		BoundingCone		= 0x00000010,
	};

private:
	struct FunctionTable {
		bool			(Bounds<T,dim>::*isInside)(const Vector<T,dim>& ) const;
		void			(Bounds<T,dim>::*translate)(const Vector<T,dim>& );
		void			(Bounds<T,dim>::*rotate)(const Matrix<T,dim>& );
		void			(Bounds<T,dim>::*transform)(const Matrix<T,dim+1>& );
		
		bool			(Bounds<T,dim>::*intersectSphere)(const Sphere<T,dim>& ) const;
		bool			(Bounds<T,dim>::*intersectBox)(const Box<T,dim>& ) const;
		bool			(Bounds<T,dim>::*intersectCube)(const Cube<T,dim>& ) const;
		bool			(Bounds<T,dim>::*intersectFrustum)(const Frustum<T,dim>& ) const;

		void			(Bounds<T,dim>::*toSphere)(Sphere<T,dim>* ) const;
		void			(Bounds<T,dim>::*toBox)(Box<T,dim>* ) const;
		void			(Bounds<T,dim>::*toAlignedBox)(AlignedBox<T,dim>* ) const;
		
		void			(Bounds<T,dim>::*mergeSphere)(const Sphere<T,dim>& );
		void			(Bounds<T,dim>::*mergeBox)(const Box<T,dim>& );
		void			(Bounds<T,dim>::*mergeAlignedBox)(const AlignedBox<T,dim>& );
	};
	
	static FunctionTable g_VoidFunctions;
	static FunctionTable g_InfinityFunctions;
	static FunctionTable g_SphereFunctions;
	static FunctionTable g_BoxFunctions;
	static FunctionTable g_AlignedBoxFunctions;

	FunctionTable*		m_Functions;
	Box<T,dim>			m_Box;
	AlignedBox<T,dim>	m_AlignedBox;
	Sphere<T,dim>		m_Sphere;
	Type				m_Type;
	
	
public:
	inline Bounds();
	inline Bounds(const Sphere<T,dim>& );
	inline Bounds(const Box<T,dim>& );
	inline Bounds(const AlignedBox<T,dim>& );
	
	inline Type				getType() const;

	inline bool				isEmpty() const;
	inline bool				isInfinite() const;
	
	inline void				empty();
	inline void				infinite();
	
	inline bool				isInside(const Vector<T,dim>& ) const;

	inline bool				intersect(const Box<T,dim>& ) const;
	inline bool				intersect(const Sphere<T,dim>& ) const;
	inline bool				intersect(const Cube<T,dim>& ) const;
	inline bool				intersect(const Frustum<T,dim>& ) const;

	void					setBounds(const Bounds<T,dim>& );
	void					setBounds(const Sphere<T,dim>& );
	void					setBounds(const Box<T,dim>& );
	void					setBounds(const AlignedBox<T,dim>& );

	inline bool				bound(const Vector<T,dim>* ,unsigned int n);
	bool					bound(const Vector<T,dim>* ,unsigned int stride,unsigned int n);
	
	void					merge(const Bounds<T,dim>& );
	inline void				merge(const Sphere<T,dim>& );
	inline void				merge(const AlignedBox<T,dim>& );
	inline void				merge(const Box<T,dim>& );
	
	inline Bounds<T,dim>& 	translate(const Vector<T,dim>& );
	inline Bounds<T,dim>& 	rotate(const Matrix<T,dim>& );
	inline Bounds<T,dim>& 	transform(const Matrix<T,dim+1>& );
	inline Bounds<T,dim>& 	transform(const Transform<T,dim>& );

	inline void				toBox(Box<T,dim>* ) const;
	inline void				toSphere(Sphere<T,dim>* ) const;
	inline void				toAlignedBox(AlignedBox<T,dim>* ) const;

	inline const Box<T,dim>& toBox();
	inline const AlignedBox<T,dim>&	toAlignedBox();
	inline const Sphere<T,dim>&	toSphere();
	
	inline const Bounds<T,dim>& operator=(const Bounds<T,dim>& );
	inline const Bounds<T,dim>& operator=(const Sphere<T,dim>& );
	inline const Bounds<T,dim>& operator=(const Box<T,dim>& );
	inline const Bounds<T,dim>& operator=(const AlignedBox<T,dim>& );
	
private:
	void				void_translate(const Vector<T,dim>& );
	void				void_rotate(const Matrix<T,dim>& );
	void				void_transform(const Matrix<T,dim+1>& );
	bool				void_isInside(const Vector<T,dim>& ) const;
	bool				void_intersectSphere(const Sphere<T,dim>& ) const;
	bool				void_intersectBox(const Box<T,dim>& ) const;
	bool				void_intersectCube(const Cube<T,dim>& ) const;
	bool				void_intersectFrustum(const Frustum<T,dim>& ) const;
	void				void_toSphere(Sphere<T,dim>* ) const;
	void				void_toBox(Box<T,dim>* ) const;
	void				void_toAlignedBox(AlignedBox<T,dim>* ) const;
	void				void_mergeSphere(const Sphere<T,dim>& );
	void				void_mergeBox(const Box<T,dim>& );
	void				void_mergeAlignedBox(const AlignedBox<T,dim>& );

	void				infinity_translate(const Vector<T,dim>& );
	void				infinity_rotate(const Matrix<T,dim>& );
	void				infinity_transform(const Matrix<T,dim+1>& );
	bool				infinity_isInside(const Vector<T,dim>& ) const;
	bool				infinity_intersectSphere(const Sphere<T,dim>& ) const;
	bool				infinity_intersectBox(const Box<T,dim>& ) const;
	bool				infinity_intersectCube(const Cube<T,dim>& ) const;
	bool				infinity_intersectFrustum(const Frustum<T,dim>& ) const;
	void				infinity_toSphere(Sphere<T,dim>* ) const;
	void				infinity_toBox(Box<T,dim>* ) const;
	void				infinity_toAlignedBox(AlignedBox<T,dim>* ) const;
	void				infinity_mergeSphere(const Sphere<T,dim>& );
	void				infinity_mergeBox(const Box<T,dim>& );
	void				infinity_mergeAlignedBox(const AlignedBox<T,dim>& );
	
	void				sphere_translate(const Vector<T,dim>& );
	void				sphere_rotate(const Matrix<T,dim>& );
	void				sphere_transform(const Matrix<T,dim+1>& );
	bool				sphere_isInside(const Vector<T,dim>& ) const;
	bool				sphere_intersectSphere(const Sphere<T,dim>& ) const;
	bool				sphere_intersectBox(const Box<T,dim>& ) const;
	bool				sphere_intersectCube(const Cube<T,dim>& ) const;
	bool				sphere_intersectFrustum(const Frustum<T,dim>& ) const;
	void				sphere_toSphere(Sphere<T,dim>* ) const;
	void				sphere_toBox(Box<T,dim>* ) const;
	void				sphere_toAlignedBox(AlignedBox<T,dim>* ) const;
	void				sphere_mergeSphere(const Sphere<T,dim>& );
	void				sphere_mergeBox(const Box<T,dim>& );
	void				sphere_mergeAlignedBox(const AlignedBox<T,dim>& );

	void				box_translate(const Vector<T,dim>& );
	void				box_rotate(const Matrix<T,dim>& );
	void				box_transform(const Matrix<T,dim+1>& );
	bool				box_isInside(const Vector<T,dim>& ) const;
	bool				box_intersectSphere(const Sphere<T,dim>& ) const;
	bool				box_intersectBox(const Box<T,dim>& ) const;
	bool				box_intersectCube(const Cube<T,dim>& ) const;
	bool				box_intersectFrustum(const Frustum<T,dim>& ) const;
	void				box_toSphere(Sphere<T,dim>* ) const;
	void				box_toBox(Box<T,dim>* ) const;
	void				box_toAlignedBox(AlignedBox<T,dim>* ) const;
	void				box_mergeSphere(const Sphere<T,dim>& );
	void				box_mergeBox(const Box<T,dim>& );
	void				box_mergeAlignedBox(const AlignedBox<T,dim>& );

	void				alignedBox_translate(const Vector<T,dim>& );
	void				alignedBox_rotate(const Matrix<T,dim>& );
	void				alignedBox_transform(const Matrix<T,dim+1>& );
	bool				alignedBox_isInside(const Vector<T,dim>& ) const;
	bool				alignedBox_intersectSphere(const Sphere<T,dim>& ) const;
	bool				alignedBox_intersectBox(const Box<T,dim>& ) const;
	bool				alignedBox_intersectCube(const Cube<T,dim>& ) const;
	bool				alignedBox_intersectFrustum(const Frustum<T,dim>& ) const;
	void				alignedBox_toSphere(Sphere<T,dim>* ) const;
	void				alignedBox_toBox(Box<T,dim>* ) const;
	void				alignedBox_toAlignedBox(AlignedBox<T,dim>* ) const;
	void				alignedBox_mergeSphere(const Sphere<T,dim>& );
	void				alignedBox_mergeBox(const Box<T,dim>& );
	void				alignedBox_mergeAlignedBox(const AlignedBox<T,dim>& );
};

// Some Transformations for some common objects
template<class T, int dim> inline Bounds<T,dim> operator*(const Transform<T,dim>& ,const Bounds<T,dim>& );


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> class Bounds2 : public Bounds<T,2> {
//DeclareClass(Bounds2, Bounds);
public:
	inline Bounds2() {};
	inline Bounds2(const Sphere<T,2>& s) : Bounds<T,2>(s) {};
	inline Bounds2(const Box<T,2>& b) : Bounds<T,2>(b) {};
	inline Bounds2(const AlignedBox<T,2>& b) : Bounds<T,2>(b) {};
	inline Bounds2(const Bounds<T,2>& b) : Bounds<T,2>(b) {};
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> class Bounds3 : public Bounds<T,3> {
//DeclareClass(Bounds3, Bounds);
public:
	inline Bounds3() {};
	inline Bounds3(const Sphere<T,3>& s) : Bounds<T,3>(s) {};
	inline Bounds3(const Box<T,3>& b) : Bounds<T,3>(b) {};
	inline Bounds3(const AlignedBox<T,3>& b) : Bounds<T,3>(b) {};
	inline Bounds3(const Bounds<T,3>& b) : Bounds<T,3>(b) {};
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> class Bounds4 : public Bounds<T,4> {
//DeclareClass(Bounds4, Bounds);
public:
	inline Bounds4() {};
	inline Bounds4(const Sphere<T,4>& s) : Bounds<T,4>(s) {};
	inline Bounds4(const Box<T,4>& b) : Bounds<T,4>(b) {};
	inline Bounds4(const AlignedBox<T,4>& b) : Bounds<T,4>(b) {};
	inline Bounds4(const Bounds<T,4>& b) : Bounds<T,4>(b) {};
};

}; };

#endif
