/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#ifndef __GEOMETRY_BOUNDS_INL
#define __GEOMETRY_BOUNDS_INL

namespace magnum { namespace geometry {


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Bounds<T,dim>::Bounds() {
	m_Functions = &g_VoidFunctions;
	m_Type = BoundingVoid;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Bounds<T,dim>::Bounds(const Sphere<T,dim>& s) {
	setBounds(s);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Bounds<T,dim>::Bounds(const AlignedBox<T,dim>& b) {
	setBounds(b);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Bounds<T,dim>::Bounds(const Box<T,dim>& b) {
	setBounds(b);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline typename Bounds<T,dim>::Type Bounds<T,dim>::getType() const {
	return m_Type;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline void Bounds<T,dim>::empty() {
	m_Functions = &g_VoidFunctions;
	m_Type = BoundingVoid;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline void Bounds<T,dim>::infinite() {
	m_Functions = &g_InfinityFunctions;
	m_Type = BoundingInfinity;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline bool Bounds<T,dim>::bound(const Vector<T,dim>* v,unsigned int n) {
	return bound(v, sizeof(Vector<T,dim>), n);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Bounds<T,dim>::bound(const Vector<T,dim>* v,unsigned int stride,unsigned int n) {
	AlignedBox<T,dim> box;
	box.bound(v, stride, n);
	setBounds(box);
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::setBounds(const Bounds<T,dim>& bounds) {
	m_Type = bounds.m_Type;
	m_Sphere = bounds.m_Sphere;
	m_Box = bounds.m_Box;
	m_AlignedBox = bounds.m_AlignedBox;
	m_Functions = bounds.m_Functions;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::setBounds(const Sphere<T,dim>& sphere) {
	m_Sphere = sphere;
	m_Type = BoundingSphere;
	m_Functions = &g_SphereFunctions;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::setBounds(const Box<T,dim>& box) {
	m_Box = box;
	m_Type = BoundingBox;
	m_Functions = &g_BoxFunctions;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::setBounds(const AlignedBox<T,dim>& box) {
	m_AlignedBox = box;
	m_Type = AlignedBoundingBox;
	m_Functions = &g_AlignedBoxFunctions;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline bool Bounds<T,dim>::isEmpty() const {
	return (m_Type == BoundingVoid);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline bool Bounds<T,dim>::isInfinite() const {
	return (m_Type == BoundingInfinity);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline bool Bounds<T,dim>::isInside(const Vector<T,dim>& v) const {
	return (this->*m_Functions->isInside)(v);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline bool Bounds<T,dim>::intersect(const Box<T,dim>& box) const {
	return (this->*m_Functions->intersectBox(box));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline bool Bounds<T,dim>::intersect(const Sphere<T,dim>& s) const {
	return (this->*m_Functions->intersectSphere(sphere));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline bool Bounds<T,dim>::intersect(const Cube<T,dim>& cube) const {
	return (this->*m_Functions->intersectCube(cube));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline bool Bounds<T,dim>::intersect(const Frustum<T,dim>& f) const {
	return (this->*m_Functions->intersectFrustum(f));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Bounds<T,dim>& Bounds<T,dim>::translate(const Vector<T,dim>& v) {
	(this->*m_Functions->translate)(v);
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Bounds<T,dim>& Bounds<T,dim>::rotate(const Matrix<T,dim>& mtx) {
	(this->*m_Functions->rotate)(mtx);
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Bounds<T,dim>& Bounds<T,dim>::transform(const Matrix<T,dim+1>& mtx) {
	(this->*m_Functions->transform)(mtx);
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Bounds<T,dim>& Bounds<T,dim>::transform(const Transform<T,dim>& t) {
	(this->*m_Functions->transform)(t.getTransform());
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Sphere<T,dim>& Bounds<T,dim>::toSphere() {
	if (m_Type != BoundingSphere) {
		(this->*m_Functions->toSphere)(&m_Sphere);
		m_Type = BoundingSphere;
	}
	return m_Sphere
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline void Bounds<T,dim>::toSphere(Sphere<T,dim>* s) const {
	(this->*m_Functions->toSphere)(s);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Box<T,dim>& Bounds<T,dim>::toBox() {
	if (m_Type != BoundingBox) {
		(this->*m_Functions->toBox)(&m_Box);
		m_Type = BoundingBox;
	}
	return m_Box;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline void Bounds<T,dim>::toBox(Box<T,dim>* b) const {
	(this->*m_Functions->toBox)(b);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const AlignedBox<T,dim>& Bounds<T,dim>::toAlignedBox() {
	if (m_Type != AlignedBoundingBox) {
		(this->*m_Functions->toAlignedBox)(&m_AlignedBox);
		m_Type = AlignedBoundingBox;
	}
	return m_AlignedBox;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline void Bounds<T,dim>::toAlignedBox(AlignedBox<T,dim>* b) const {
	(this->*m_Functions->toAlignedBox)(b);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::merge(const Bounds<T,dim>& b) {
	switch(b.getType()) {
		case BoundingVoid:
			return;
		case BoundingInfinity:
			infinite();
			return;
		case BoundingSphere:
			(this->*m_Functions->mergeSphere)(b.m_Sphere);
			return;
		case BoundingBox:
			(this->*m_Functions->mergeBox)(b.m_Box);
			return;
		case AlignedBoundingBox:
			(this->*m_Functions->mergeAlignedBox)(b.m_AlignedBox);
			return;
		default:
			ASSERT(false);
			return;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline void Bounds<T,dim>::merge(const Sphere<T,dim>& s) {
	(this->*m_Functions->mergeSphere)(s);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline void Bounds<T,dim>::merge(const AlignedBox<T,dim>& b) {
	(this->*m_Functions->mergeAlignedBox)(b);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline void Bounds<T,dim>::merge(const Box<T,dim>& b) {
	(this->*m_Functions->mergeBox)(b);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Bounds<T,dim>&	Bounds<T,dim>::operator=(const Bounds<T,dim>& b) {
	setBounds(b);
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Bounds<T,dim>&	Bounds<T,dim>::operator=(const Sphere<T,dim>& s) {
	setBounds(s);
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Bounds<T,dim>&	Bounds<T,dim>::operator=(const Box<T,dim>& b) {
	setBounds(b);
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline const Bounds<T,dim>&	Bounds<T,dim>::operator=(const AlignedBox<T,dim>& b) {
	setBounds(b);
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
inline Bounds<T,dim> operator*(const Transform<T,dim>& t, const Bounds<T,dim>& b) {
	Bounds<T,dim> r(b);
	r.transform(t);
	return r;
}


/*----------------------------------------------------------------------------
DUMMY DUMMY DUMMY DUMMY DUMMY DUMMY DUMMY DUMMY DUMMY DUMMY DUMMY DUMMY DUMMY
DUMMY DUMMY DUMMY DUMMY DUMMY DUMMY DUMMY DUMMY DUMMY DUMMY DUMMY DUMMY DUMMY
DUMMY DUMMY DUMMY DUMMY DUMMY DUMMY DUMMY DUMMY DUMMY DUMMY DUMMY DUMMY DUMMY
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
typename Bounds<T,dim>::FunctionTable Bounds<T,dim>::g_VoidFunctions = {
		void_isInside,
		void_translate,
		void_rotate,
		void_transform,
		void_intersectSphere,
		void_intersectBox,
		void_intersectCube,
		void_intersectFrustum,
		void_toSphere,
		void_toBox,
		void_toAlignedBox,
		void_mergeSphere,
		void_mergeBox,
		void_mergeAlignedBox };

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::void_translate(const Vector<T,dim>& ) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::void_rotate(const Matrix<T,dim>& ) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::void_transform(const Matrix<T,dim+1>& ) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Bounds<T,dim>::void_isInside(const Vector<T,dim>& ) const {
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Bounds<T,dim>::void_intersectSphere(const Sphere<T,dim>& ) const {
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Bounds<T,dim>::void_intersectBox(const Box<T,dim>& ) const {
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Bounds<T,dim>::void_intersectCube(const Cube<T,dim>& ) const {
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Bounds<T,dim>::void_intersectFrustum(const Frustum<T,dim>& ) const {
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::void_toSphere(Sphere<T,dim>* s) const {
	s->setOrigin(Vector<T,dim>::Null);
	s->setRadius(T(0));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::void_toBox(Box<T,dim>* b) const {
	b->setOrigin(Vector<T,dim>::Null);
	for (unsigned int i = 0; i < dim; i++) {
		b->setAxis(i, Vector<T,dim>::Null);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::void_toAlignedBox(AlignedBox<T,dim>* b) const {
	b->setMinimum(Vector<T,dim>::Null);
	b->setMaximum(Vector<T,dim>::Null);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::void_mergeSphere(const Sphere<T,dim>& s) {
	m_Sphere = s;
	m_Type = BoundingSphere;
	m_Functions = &g_SphereFunctions;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::void_mergeBox(const Box<T,dim>& b) {
	m_Box = b;
	m_Type = BoundingBox;
	m_Functions = &g_BoxFunctions;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::void_mergeAlignedBox(const AlignedBox<T,dim>& b) {
	m_AlignedBox = b;
	m_Type = AlignedBoundingBox;
	m_Functions = &g_AlignedBoxFunctions;
}


/*----------------------------------------------------------------------------
INFINITY INFINITY INFINITY INFINITY INFINITY INFINITY INFINITY INFINITY INFINITY 
INFINITY INFINITY INFINITY INFINITY INFINITY INFINITY INFINITY INFINITY INFINITY 
INFINITY INFINITY INFINITY INFINITY INFINITY INFINITY INFINITY INFINITY INFINITY 
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
typename Bounds<T,dim>::FunctionTable Bounds<T,dim>::g_InfinityFunctions = {
		infinity_isInside,
		infinity_translate,
		infinity_rotate,
		infinity_transform,
		infinity_intersectSphere,
		infinity_intersectBox,
		infinity_intersectCube,
		infinity_intersectFrustum,
		infinity_toSphere,
		infinity_toBox,
		infinity_toAlignedBox,
		infinity_mergeSphere,
		infinity_mergeBox,
		infinity_mergeAlignedBox };



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::infinity_translate(const Vector<T,dim>& ) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::infinity_rotate(const Matrix<T,dim>& ) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::infinity_transform(const Matrix<T,dim+1>& ) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Bounds<T,dim>::infinity_isInside(const Vector<T,dim>& ) const {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Bounds<T,dim>::infinity_intersectSphere(const Sphere<T,dim>& ) const {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Bounds<T,dim>::infinity_intersectBox(const Box<T,dim>& ) const {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Bounds<T,dim>::infinity_intersectCube(const Cube<T,dim>& ) const {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Bounds<T,dim>::infinity_intersectFrustum(const Frustum<T,dim>& ) const {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::infinity_toSphere(Sphere<T,dim>* ) const {
	ASSERT(false);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::infinity_toBox(Box<T,dim>* ) const {
	ASSERT(false);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::infinity_toAlignedBox(AlignedBox<T,dim>* ) const {
	ASSERT(false);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::infinity_mergeSphere(const Sphere<T,dim>& s) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::infinity_mergeBox(const Box<T,dim>& b) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::infinity_mergeAlignedBox(const AlignedBox<T,dim>& b) {
}




/*----------------------------------------------------------------------------
SPHERE SPHERE SPHERE SPHERE SPHERE SPHERE SPHERE SPHERE SPHERE SPHERE SPHERE 
SPHERE SPHERE SPHERE SPHERE SPHERE SPHERE SPHERE SPHERE SPHERE SPHERE SPHERE 
SPHERE SPHERE SPHERE SPHERE SPHERE SPHERE SPHERE SPHERE SPHERE SPHERE SPHERE 
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
typename Bounds<T,dim>::FunctionTable Bounds<T,dim>::g_SphereFunctions = {
		sphere_isInside,
		sphere_translate,
		sphere_rotate,
		sphere_transform,
		sphere_intersectSphere,
		sphere_intersectBox,
		sphere_intersectCube,
		sphere_intersectFrustum,
		sphere_toSphere,
		sphere_toBox,
		sphere_toAlignedBox,
		sphere_mergeSphere,
		sphere_mergeBox,
		sphere_mergeAlignedBox };
		
/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::sphere_translate(const Vector<T,dim>& v) {
	m_Sphere.setOrigin(m_Sphere.getOrigin() + v);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::sphere_rotate(const Matrix<T,dim>& mtx) {
	m_Sphere.setOrigin(mtx * m_Sphere.getOrigin());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::sphere_transform(const Matrix<T,dim+1>& mtx) {
	Vector<T,dim> origin;
	for (unsigned int i = 0; i < dim; i++) {
		origin[i] = mtx[i][dim];
		for (unsigned int j = 0; j < dim; j++)
			origin[i] += mtx[i][j] * m_Sphere.getOrigin()[j];
	}
	m_Sphere.setOrigin(origin);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Bounds<T,dim>::sphere_isInside(const Vector<T,dim>& v) const {
	return m_Sphere.isInside(v);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Bounds<T,dim>::sphere_intersectSphere(const Sphere<T,dim>& s) const {
	return m_Sphere.intersect(s);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Bounds<T,dim>::sphere_intersectBox(const Box<T,dim>& b) const {
	return m_Sphere.intersect(b);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Bounds<T,dim>::sphere_intersectCube(const Cube<T,dim>& c) const {
	return m_Sphere.intersect(c);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Bounds<T,dim>::sphere_intersectFrustum(const Frustum<T,dim>& ) const {
	ASSERT(false);
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::sphere_toSphere(Sphere<T,dim>* s) const {
	*s = m_Sphere;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::sphere_toBox(Box<T,dim>* box) const {
	Vector<T,dim> axis;
	Vector<T,dim> origin = m_Sphere.getOrigin();
	axis.null();
	for (unsigned int i = 0; i < dim; i++) {
		origin[i] -= m_Sphere.getRadius();
		axis[i] = m_Sphere.getRadius()*T(2);
		box->setAxis(i, axis);
		axis[i] = T(0);
	}
	box->setOrigin(origin);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::sphere_toAlignedBox(AlignedBox<T,dim>* box) const {
	Vector<T,dim> vec;
	for (unsigned int i = 0; i < dim; i++) {
		vec[i] = m_Sphere.getRadius();
	}
	box->setMinimum(m_Sphere.getOrigin() - vec);
	box->setMaximum(m_Sphere.getOrigin() + vec);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::sphere_mergeSphere(const Sphere<T,dim>& s) {
	m_Sphere.merge(s);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::sphere_mergeBox(const Box<T,dim>& ) {
	ASSERT(false);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::sphere_mergeAlignedBox(const AlignedBox<T,dim>& ) {
	ASSERT(false);
}



/*----------------------------------------------------------------------------
BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX
BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX
BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX BOX
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
typename Bounds<T,dim>::FunctionTable Bounds<T,dim>::g_BoxFunctions = {
		box_isInside,
		box_translate,
		box_rotate,
		box_transform,
		box_intersectSphere,
		box_intersectBox,
		box_intersectCube,
		box_intersectFrustum,
		box_toSphere,
		box_toBox,
		box_toAlignedBox,
		box_mergeSphere,
		box_mergeBox,
		box_mergeAlignedBox };
		
/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::box_translate(const Vector<T,dim>& v) {
	m_Box.setOrigin(m_Box.getOrigin() + v);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::box_rotate(const Matrix<T,dim>& mtx) {
	m_Box.setOrigin(mtx * m_Box.getOrigin());
	for (unsigned int i = 0; i < dim; i++)
		m_Box.setAxis(i, mtx * m_Box.getAxis(i));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::box_transform(const Matrix<T,dim+1>& ) {
	ASSERT(false);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Bounds<T,dim>::box_isInside(const Vector<T,dim>& v) const {
	return m_Box.isInside(v);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Bounds<T,dim>::box_intersectSphere(const Sphere<T,dim>& s) const {
	return m_Box.intersect(s);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Bounds<T,dim>::box_intersectBox(const Box<T,dim>& b) const {
	return m_Box.intersect(b);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Bounds<T,dim>::box_intersectCube(const Cube<T,dim>& c) const {
	return m_Box.intersect(c);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Bounds<T,dim>::box_intersectFrustum(const Frustum<T,dim>& ) const {
	ASSERT(false);
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::box_toSphere(Sphere<T,dim>* s) const {
	Vector<T,dim> v[1 << dim];
	m_Box.getVertices(v);
	s->bound(v, 1 << dim);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::box_toBox(Box<T,dim>* box) const {
	*box = m_Box;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::box_toAlignedBox(AlignedBox<T,dim>* box) const {
	Vector<T,dim> vmin, vmax;
	vmin = vmax = m_Box.getOrigin();
	unsigned int i,j;
	for (i = 1; i < (1 << dim); i++) {
		Vector<T,dim> vec = m_Box.getOrigin();;
		for (j = 0; j < dim; j++)
			if(i & (1 << j))
				vec += m_Box.getAxis(j);
		for (j = 0; j < dim; j++) {
			if (vmin[j] > vec[j])
				vmin[j] = vec[j];
			else if (vmax[j] < vec[j])
				vmax[j] = vec[j];
		}
	}
	box->setMinimum(vmin);
	box->setMaximum(vmax);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::box_mergeSphere(const Sphere<T,dim>& ) {
	ASSERT(false);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::box_mergeBox(const Box<T,dim>& ) {
	ASSERT(false);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::box_mergeAlignedBox(const AlignedBox<T,dim>& ) {
	ASSERT(false);
}




/*----------------------------------------------------------------------------
ALIGNEDBOX ALIGNEDBOX ALIGNEDBOX ALIGNEDBOX ALIGNEDBOX ALIGNEDBOX ALIGNEDBOX 
ALIGNEDBOX ALIGNEDBOX ALIGNEDBOX ALIGNEDBOX ALIGNEDBOX ALIGNEDBOX ALIGNEDBOX 
ALIGNEDBOX ALIGNEDBOX ALIGNEDBOX ALIGNEDBOX ALIGNEDBOX ALIGNEDBOX ALIGNEDBOX 
----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
typename Bounds<T,dim>::FunctionTable Bounds<T,dim>::g_AlignedBoxFunctions = {
		alignedBox_isInside,
		alignedBox_translate,
		alignedBox_rotate,
		alignedBox_transform,
		alignedBox_intersectSphere,
		alignedBox_intersectBox,
		alignedBox_intersectCube,
		alignedBox_intersectFrustum,
		alignedBox_toSphere,
		alignedBox_toBox,
		alignedBox_toAlignedBox,
		alignedBox_mergeSphere,
		alignedBox_mergeBox,
		alignedBox_mergeAlignedBox };		
		
/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::alignedBox_translate(const Vector<T,dim>& v) {
	m_AlignedBox.setOrigin(m_AlignedBox.getOrigin() + v);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::alignedBox_rotate(const Matrix<T,dim>& mtx) {
	m_Box.setOrigin(mtx * m_AlignedBox.getMinimum());

	// Construct new axes
	Vector<T,dim> extents = m_AlignedBox.getMaximum() - m_AlignedBox.getMinimum();
	for (unsigned int i = 0; i < dim; i++)
		m_Box.setAxis(i, extents[i] * mtx.getColumn(i));

	// Set correct type and function table
	m_Type = BoundingBox;
	m_Functions = &g_BoxFunctions;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::alignedBox_transform(const Matrix<T,dim+1>& mtx) {
	// Transform origin
	Vector<T,dim> origin;
	for (unsigned int i = 0; i < dim; i++) {
		origin[i] = mtx[i][dim];
		for (unsigned int j = 0; j < dim; j++)
			origin[i] += mtx[i][j] * m_AlignedBox.getMinimum()[j];
	}
	m_Box.setOrigin(origin);
	
	// Construct new axes
	Vector<T,dim> extents = m_AlignedBox.getMaximum() - m_AlignedBox.getMinimum();
	for (unsigned int i = 0; i < dim; i++) {
		Vector<T,dim> axis;
		for (unsigned int j = 0; j < dim; j++)
			axis[j] = extents[i] * mtx[j][i] + mtx[dim][j];
		m_Box.setAxis(i, axis);
	}
	
	// Set correct type and function table
	m_Type = BoundingBox;
	m_Functions = &g_BoxFunctions;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Bounds<T,dim>::alignedBox_isInside(const Vector<T,dim>& v) const {
	return m_AlignedBox.isInside(v);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Bounds<T,dim>::alignedBox_intersectSphere(const Sphere<T,dim>& s) const {
	return m_AlignedBox.intersect(s);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Bounds<T,dim>::alignedBox_intersectBox(const Box<T,dim>& b) const {
	return m_AlignedBox.intersect(b);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Bounds<T,dim>::alignedBox_intersectCube(const Cube<T,dim>& c) const {
	return m_AlignedBox.intersect(c);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
bool Bounds<T,dim>::alignedBox_intersectFrustum(const Frustum<T,dim>& ) const {
	ASSERT(false);
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::alignedBox_toSphere(Sphere<T,dim>* s) const {
	s->setOrigin(m_AlignedBox.getOrigin());
	s->setRadius((m_AlignedBox.getMaximum() - m_AlignedBox.getMinimum()).length()/(2));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::alignedBox_toBox(Box<T,dim>* box) const {
	Vector<T,dim> diff = m_AlignedBox.getMaximum() - m_AlignedBox.getMinimum();
	Vector<T,dim> axis;
	axis.null();
	for (unsigned int i = 0; i < dim; i++) {
		axis[i] = diff[i];
		box->setAxis(i, axis);
		axis[i] = T(0);
	}
	box->setOrigin(m_AlignedBox.getMinimum());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::alignedBox_toAlignedBox(AlignedBox<T,dim>* box) const {
	*box = m_AlignedBox;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::alignedBox_mergeSphere(const Sphere<T,dim>& ) {
	ASSERT(false);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::alignedBox_mergeBox(const Box<T,dim>& ) {
	ASSERT(false);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim> 
void Bounds<T,dim>::alignedBox_mergeAlignedBox(const AlignedBox<T,dim>& b) {
	m_AlignedBox.merge(b);
}



}; };

#endif
