/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#ifndef __GEOMETRY_BOX_INL
#define __GEOMETRY_BOX_INL

namespace magnum { namespace geometry {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline void Box<T,dim>::setOrigin(const Vector<T,dim>& o) {
	m_Origin = o;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline const Vector<T,dim>& Box<T,dim>::getOrigin() const {
	return m_Origin;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline void	Box<T,dim>::setAxes(const Vector<T,dim>* e) {
	for (unsigned int i = 0; i < dim; i++) {
		m_Axes[i] = e[i];
		m_Axes2[i] = m_Axes[i] * m_Axes[i];
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline const Vector<T,dim>* Box<T,dim>::getAxes() const {
	return m_Axes;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline void	Box<T,dim>::setAxis(unsigned int i, const Vector<T,dim>& a) {
	ASSERT(i < dim);
	m_Axes[i] = a;
	m_Axes2[i] = m_Axes[i] * m_Axes[i];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline const Vector<T,dim>& Box<T,dim>::getAxis(unsigned int i) const {
	ASSERT(i < dim);
	return m_Axes[i];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline bool Box<T,dim>::isEmpty() const {
	for(unsigned int i = 0; i < dim; i++) {
		if (m_Axes2[i] == T(0))
			return true;
	}
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline void Box<T,dim>::empty() {
	for(unsigned int i = 0; i < dim; i++) {
		m_Axes[i] = Vector<T,dim>::Null;
		m_Axes2[i] = T(0);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
inline bool Box<T,dim>::isInside(const Vector<T,dim>& v) const {
	Vector<T,dim> p = v - m_Origin;
	for (unsigned int i = 0; i < dim; i++) {
		T val = p*m_Axes[i];
		if ((val < -m_Axes2[i]) || (val > m_Axes2[i]))
			return false;
	}
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
T Box<T,dim>::getDistance(const Vector<T,dim>& o) const {
	// Find point within box nearest to center of sphere
	Vector<T,dim> p;
	p.null();
	for (unsigned int i = 0; i < dim; i++) {
		T t = o*m_Axes[i];
		if (t >= m_Axes2[i]) {
			p += m_Axes[i];
		} else if (t <= -m_Axes2[i]) {
			p -= m_Axes[i];
		} else {
			t /= m_Axes2[i];
			p += m_Axes[i]*t;
		}
	}

	return (o - p).length();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
bool Box<T,dim>::intersect(const Sphere<T,dim>& sphere) const {
	// Find point within box nearest to center of sphere
	Vector<T,dim> p;
	Vector<T,dim> o = sphere.getOrigin() - m_Origin;
	p.null();
	for (unsigned int i = 0; i < dim; i++) {
		T t = o*m_Axes[i];
		if (t >= m_Axes2[i]) {
			p += m_Axes[i];
		} else if (t <= -m_Axes2[i]) {
			p -= m_Axes[i];
		} else {
			t /= m_Axes2[i];
			p += m_Axes[i]*t;
		}
	}

	return ((o - p).length2() <= sphere.getRadius() * sphere.getRadius());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
bool Box<T,dim>::intersect(const Cube<T,dim>& cube) const {
	unsigned int i,j;
	T radius = cube.getRadius();
	Vector<T,dim> origin = cube.getOrigin();
	// Case 1: Project all axes onto cube axes
	for (i = 0; i < dim; i++) {
		T l = abs(m_Axes[0][i]);
		for (j = 1; j < dim; j++)
			l += abs(m_Axes[j][i]);
		if ((m_Origin[i] - l >= origin[i] + radius) 
		 || (m_Origin[i] + l <= origin[i] - radius))
			return false;
	}
	
	// Case 2: Project all cube axes onto box
	Vector<T,dim> diff = origin - m_Origin;
	for (i = 0; i < dim; i++) {
		T l = abs(m_Axes[i][0]);
		for (j = 1; j < dim; j++)	
			l += abs(m_Axes[i][j]);
		l *= radius;
		T p = abs(diff * m_Axes[i]);
		if (p >= l + m_Axes2[i])
			return false;
	}
	
	if (dim == 3) {
		// Finally test mixtures of box axes and cube axes
		for (i = 0; i < 3; i++) {
			Vector<T,dim> axis;
			T l;
			axis[0] = T(0);
			axis[1] = m_Axes[i][2];
			axis[2] = -m_Axes[i][1];
			l = radius * (abs(axis[1]) + abs(axis[2]));
			l += abs(axis[1]*m_Axes[0][1] + axis[2]*m_Axes[0][2]);
			l += abs(axis[1]*m_Axes[1][1] + axis[2]*m_Axes[1][2]);
			l += abs(axis[1]*m_Axes[2][1] + axis[2]*m_Axes[2][2]);
			if (abs(axis[1]*diff[1] + axis[2]*diff[2]) >= l)
				return false;

			axis[0] = -m_Axes[i][2];
			axis[1] = T(0);
			axis[2] = m_Axes[i][0];
			l = radius * (abs(axis[0]) + abs(axis[2]));
			l += abs(axis[0]*m_Axes[0][0] + axis[2]*m_Axes[0][2]);
			l += abs(axis[0]*m_Axes[1][0] + axis[2]*m_Axes[1][2]);
			l += abs(axis[0]*m_Axes[2][0] + axis[2]*m_Axes[2][2]);
			if (abs(axis[0]*diff[0] + axis[2]*diff[2]) >= l)
				return false;

			axis[0] = m_Axes[i][1];
			axis[1] = -m_Axes[i][0];
			axis[2] = T(0);
			l = radius * (abs(axis[0]) + abs(axis[1]));
			l += abs(axis[0]*m_Axes[0][0] + axis[1]*m_Axes[0][1]);
			l += abs(axis[0]*m_Axes[1][0] + axis[1]*m_Axes[1][1]);
			l += abs(axis[0]*m_Axes[2][0] + axis[1]*m_Axes[2][1]);
			if (abs(axis[0]*diff[0] + axis[1]*diff[1]) >= l)
				return false;
		}
	} else if (dim >= 4) {
		ASSERT(false);
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
bool Box<T,dim>::intersect(const AlignedBox<T,dim>& box) const {
	unsigned int i,j;
	Vector<T,dim> mi = box.getMinimum();
	Vector<T,dim> ma = box.getMaximum();
	// Case 1: Project all axes onto aligned box axes
	for (i = 0; i < dim; i++) {
		T l = abs(m_Axes[0][i]);
		for (j = 1; j < dim; j++)
			l += abs(m_Axes[j][i]);
		if ((m_Origin[i] - l >= ma[i]) 
		 || (m_Origin[i] + l <= mi[i]))
			return false;
	}
	
	// Case 2: Project all aligned box axes onto box
	Vector<T,dim> diff = (mi + ma) - T(2)*m_Origin;
	Vector<T,dim> size = ma - mi;
	for (i = 0; i < dim; i++) {
		T l = abs(m_Axes[i][0] * size[0]);
		for (j = 1; j < dim; j++)	
			l += abs(m_Axes[i][j] * size[j]);
		T p = abs(diff * m_Axes[i]);
		if (p >= l + T(2)*m_Axes2[i])
			return false;
	}
	
	if (dim == 3) {
		// Finally test mixtures of box axes and cube axes
		for (i = 0; i < 3; i++) {
			Vector<T,dim> axis;
			T l,k;
			axis[0] = T(0);
			axis[1] = m_Axes[i][2];
			axis[2] = -m_Axes[i][1];
			k = abs(size[1]*abs(axis[1]) + abs(size[2]*axis[2]));
			l = abs(axis[1]*m_Axes[0][1] + axis[2]*m_Axes[0][2]);
			l += abs(axis[1]*m_Axes[1][1] + axis[2]*m_Axes[1][2]);
			l += abs(axis[1]*m_Axes[2][1] + axis[2]*m_Axes[2][2]);
			if (abs(axis[1]*diff[1] + axis[2]*diff[2]) >= k + T(2)*l)
				return false;

			axis[0] = -m_Axes[i][2];
			axis[1] = T(0);
			axis[2] = m_Axes[i][0];
			k = abs(size[0]*abs(axis[0]) + abs(size[2]*axis[2]));
			l = abs(axis[0]*m_Axes[0][0] + axis[2]*m_Axes[0][2]);
			l += abs(axis[0]*m_Axes[1][0] + axis[2]*m_Axes[1][2]);
			l += abs(axis[0]*m_Axes[2][0] + axis[2]*m_Axes[2][2]);
			if (abs(axis[0]*diff[0] + axis[2]*diff[2]) >= k + T(2)*l)
				return false;

			axis[0] = m_Axes[i][1];
			axis[1] = -m_Axes[i][0];
			axis[2] = T(0);
			k = abs(size[0]*abs(axis[0]) + abs(size[1]*axis[1]));
			l = abs(axis[0]*m_Axes[0][0] + axis[1]*m_Axes[0][1]);
			l += abs(axis[0]*m_Axes[1][0] + axis[1]*m_Axes[1][1]);
			l += abs(axis[0]*m_Axes[2][0] + axis[1]*m_Axes[2][1]);
			if (abs(axis[0]*diff[0] + axis[1]*diff[1]) >= k + T(2)*l)
				return false;
		}
	} else if (dim >= 4) {
		ASSERT(false);
	}
	
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
bool Box<T,dim>::intersect(const Box<T,dim>& box) const {
	Vector<T,dim> diff = box.m_Origin - m_Origin;
	unsigned int i,j;
	// Case 1: Project all axes onto aligned box axes
	for (i = 0; i < dim; i++) {
		T l = abs(m_Axes[i] * box.m_Axes[0]);
		for (j = 1; j < dim; j++)
			l += abs(m_Axes[i] * box.m_Axes[j]);
		T p = abs(diff * m_Axes[i]);
		if (p >= l + m_Axes2[i])
			return false;
	}
	
	// Case 2: Project all aligned box axes onto box
	for (i = 0; i < dim; i++) {
		T l = abs(m_Axes[0] * box.m_Axes[i]);
		for (j = 1; j < dim; j++)
			l += abs(m_Axes[j] * box.m_Axes[i]);
		T p = abs(diff * box.m_Axes[i]);
		if (p >= l + box.m_Axes2[i])
			return false;
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T, int dim>
void Box<T,dim>::getVertices(Vector<T,dim>* cv) const {
	if (dim == 2) {
		cv[0] = m_Origin - m_Axes[0] - m_Axes[1];
		cv[1] = m_Origin + m_Axes[0] - m_Axes[1];
		cv[2] = m_Origin - m_Axes[0] + m_Axes[1];
		cv[3] = m_Origin + m_Axes[0] + m_Axes[1];
	} else if (dim == 3) {
		cv[0] = m_Origin - m_Axes[0] - m_Axes[1] - m_Axes[2];
		cv[1] = m_Origin + m_Axes[0] - m_Axes[1] - m_Axes[2];
		cv[2] = m_Origin - m_Axes[0] + m_Axes[1] - m_Axes[2];
		cv[3] = m_Origin + m_Axes[0] + m_Axes[1] - m_Axes[2];
		cv[4] = m_Origin - m_Axes[0] - m_Axes[1] + m_Axes[2];
		cv[5] = m_Origin + m_Axes[0] - m_Axes[1] + m_Axes[2];
		cv[6] = m_Origin - m_Axes[0] + m_Axes[1] + m_Axes[2];
		cv[7] = m_Origin + m_Axes[0] + m_Axes[1] + m_Axes[2];
	} else {
		unsigned int i,j;
		for (i = 0; i < (1 << dim); i++) {
			*cv = m_Origin;
			for (j = 0; j < dim; j++) {
				if (i & (1 << j))
					*cv += m_Axes[j];
				else
					*cv -= m_Axes[j];
			}
			cv++;
		}
	}	
}

}; };

#endif
