/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "io.memfile.h"
#include "io.exception.h"

#include "io.memfile.inl"


namespace magnum { namespace io {
using magnum::math::min;
using magnum::math::max;


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
MemoryFile::MemoryFile(unsigned long bs) {
	m_Position = 0;
	m_Size = 0;
	m_Blocksize = bs;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
MemoryFile::~MemoryFile() {
	while(m_Blocks.getSize() > 0) {
		void* data = m_Blocks[m_Blocks.getSize() - 1];
		delete[] (byte *)data;
		m_Blocks.remove(m_Blocks.getSize() - 1);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned long MemoryFile::write(const void *buf, unsigned long size) {
	unsigned long posInCurrentBlock = m_Position % m_Blocksize;
	unsigned long freeInCurrentBlock = (m_Blocksize - posInCurrentBlock) % m_Blocksize;
	unsigned long bytesToWrite;
	unsigned long bytesWritten = 0;

	bytesToWrite = min(freeInCurrentBlock, size);
	if (bytesToWrite > 0) {
		memcpy(((byte *)m_Blocks[m_Position / m_Blocksize]) + posInCurrentBlock, buf, bytesToWrite);
		m_Position += bytesToWrite;
		m_Size = max(m_Size, m_Position);
		size -= bytesToWrite;
		buf = ((byte *)buf) + bytesToWrite;
		bytesWritten += bytesToWrite;
	}

	if (m_Position + size > m_Size)
		append(m_Position + size - m_Size);

	while(size > 0) {
		bytesToWrite = min((unsigned long)m_Blocksize, size);
		memcpy(m_Blocks[m_Position / m_Blocksize], buf, bytesToWrite);
		m_Position += bytesToWrite;
		size -= bytesToWrite;
		buf = ((byte *)buf) + bytesToWrite;
		bytesWritten += bytesToWrite;
	}

	return bytesWritten;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned long MemoryFile::read(void *buf, unsigned long size) {
	unsigned long posInCurrentBlock = m_Position % m_Blocksize;
	unsigned long leftInCurrentBlock = (m_Blocksize - m_Position) % m_Blocksize;
	unsigned long bytesToRead;
	unsigned long bytesRead = 0;

	size = min(size, m_Size - m_Position);
	bytesToRead = min(leftInCurrentBlock, size);
	if (bytesToRead > 0) {
		memcpy(buf, ((byte *)m_Blocks[m_Position / m_Blocksize]) + posInCurrentBlock, bytesToRead);
		m_Position += bytesToRead;
		size -= bytesToRead;
		buf = ((byte *)buf) + bytesToRead;
		bytesRead += bytesToRead;
	}

	while(size > 0) {
		bytesToRead = min((unsigned long)m_Blocksize, size);
		memcpy(buf, m_Blocks[m_Position / m_Blocksize], bytesToRead);
		m_Position += bytesToRead;
		size -= bytesToRead;
		buf = ((byte *)buf) + bytesToRead;
		bytesRead += bytesToRead;
	}

	return bytesRead;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool MemoryFile::flush() {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned long MemoryFile::setPosition(long where,PositionMode mode) {
	switch(mode) {
		case Absolute:
			if (where < 0)
				throw IllegalPositionException();
			if (where > (long)m_Size)
				throw IllegalPositionException();
			m_Position = where;
			break;
		case Current:
			if (m_Position + where < 0)
				throw IllegalPositionException();
			if (m_Position + where > m_Size)
				throw IllegalPositionException();
			m_Position += where;
			break;
		case End:
			if (where > 0)
				throw IllegalPositionException();
			if (m_Size + where < 0)
				throw IllegalPositionException();
			m_Position = m_Size + where;
		default:
			throw IllegalPositionModeException();
	}

	return m_Position;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned long MemoryFile::getPosition() const {
	return m_Position;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool MemoryFile::rewind() {
	m_Position = 0;
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned long MemoryFile::getSize() const {
	return m_Size;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned long MemoryFile::setSize(unsigned long newSize) {
	if (newSize < m_Size)
		truncate(m_Size - newSize);
	if (newSize > m_Size)
		append(newSize - m_Size);
	return m_Size;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool MemoryFile::isWriteable() const {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool MemoryFile::isReadable() const {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Timestamp MemoryFile::getCreationTime() const {
	return m_Timestamp;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Timestamp MemoryFile::getModificationTime() const {
	return m_Timestamp;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Timestamp MemoryFile::getAccessTime() const {
	return m_Timestamp;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool MemoryFile::append(unsigned long addSize) {
	unsigned long freeInLastBlock = m_Blocks.getSize()*m_Blocksize - m_Size;
	if (addSize <= freeInLastBlock) {
		m_Size += addSize;
		return true;
	}

	addSize -= freeInLastBlock;
	m_Size += freeInLastBlock;

	for(;;) {
		void *newBlock = new byte[m_Blocksize];
		if (newBlock == NULL)
			throw DiskFullException("Not enough free memory for requested operation.");
		m_Blocks.insert(newBlock);
		if (addSize > (unsigned long)m_Blocksize) {
			m_Size += m_Blocksize;
			addSize -= m_Blocksize;
		} else {
			m_Size += addSize;
			return true;
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool MemoryFile::truncate(unsigned long delSize) {
	unsigned long usedInLastBlock = (m_Blocksize - (m_Size % m_Blocksize)) % m_Blocksize;
	if (delSize < usedInLastBlock) {
		m_Size -= delSize;
		return true;
	}

	delSize -= usedInLastBlock;
	m_Size -= usedInLastBlock;
	void* data = m_Blocks[m_Blocks.getSize() - 1];
	delete[] (byte *)data;
	m_Blocks.remove(m_Blocks.getSize() - 1);

	for(;;) {
		if (delSize >= (unsigned long)m_Blocksize) {
			data = m_Blocks[m_Blocks.getSize() - 1];
			delete[] (byte *)data;
			m_Blocks.remove(m_Blocks.getSize() - 1);
			m_Size -= m_Blocksize;
		} else {
			m_Size -= delSize;
			return true;
		}
	}
}

}; };
