/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#ifndef __MATH_LINEARSOLVER_INL
#define __MATH_LINEARSOLVER_INL

namespace magnum { namespace math {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
template<int dim>
inline bool LinearSolver<T>::solve(const Matrix<T,dim>& mtx,const Vector<T,dim>& b,Vector<T,dim>* x) {
	return solve(dim, &mtx[0][0], (const T*)b, (T*)x);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
template<int dim>
inline bool LinearSolver<T>::solve(const Matrix<T,dim>& mtx,Vector<T,dim>* b) {
	return solve(dim, &mtx[0][0], (T*)b);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
template<int dim>
inline bool LinearSolver<T>::invert(const Matrix<T,dim>& mtx,Matrix<T,dim>* inv) {
	return invert(dim, &mtx[0][0], &inv[0][0]);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool solve(const Matrix2<T>& mtx,const Vector2<T>& b,Vector2<T>* x) {
	return solve2(&mtx[0][0], (const T*)b, (T*)x);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool solve(const Matrix2<T>& mtx,Vector2<T>* b) {
	return solve2(&mtx[0][0], (T*)b);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool solve(const Matrix3<T>& mtx,const Vector3<T>& b,Vector3<T>* x) {
	return solve3(&mtx[0][0], (const T*)b, (T*)x);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool solve(const Matrix3<T>& mtx,Vector3<T>* b) {
	return solve3(&mtx[0][0], (T*)b);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool solve(const Matrix4<T>& mtx,const Vector4<T>& b,Vector4<T>* x) {
	return solve4(&mtx[0][0], (const T*)b, (T*)x);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool solve(const Matrix4<T>& mtx,Vector4<T>* b) {
	return solve3(&mtx[0][0], (T*)b);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool LinearSolver<T>::solve(unsigned int dim, const T* omtx, const T* b, T* x) {
	CopyElements(x, b, dim);
	return solve(dim, omtx, b, x);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
bool LinearSolver<T>::solve(unsigned int dim, const T* omtx, T* x) {
	T* inv = new T[dim*dim];
	ASSERT(inv);

	CopyElements(inv, omtx, dim*dim);

    bool* pivoted = new bool[dim];
    ASSERT( pivoted );
    memset(pivoted,0,dim*sizeof(bool));

    unsigned int i0, i1, i2, k;
    unsigned int row = 0, col = 0;

    // elimination by full pivoting
    for (i0 = 0; i0 < dim; i0++) {
        // search matrix (excluding pivoted rows) for maximum absolute entry
        T maxv = T(0);
        for (i1 = 0; i1 < dim; i1++) {
            if ( pivoted[i1] ) 
				continue;

            for (i2 = 0; i2 < dim; i2++) {
                if ( pivoted[i2] )
					continue;
                T absv = abs(inv[i1*dim + i2]);
                if ( absv > maxv ) {
                    maxv = absv;
                    row = i1;
                    col = i2;
                }
            }
        }

        if ( maxv == (T)0.0 )
			goto error;

        pivoted[col] = true;

        // swap rows so that A[col][col] contains the pivot entry
        if ( row != col ) {
			T tmp;
			for (k = 0; k < dim; k++) {
				tmp = inv[row*dim + k];
				inv[row*dim + k] = inv[col*dim + k];
				inv[col*dim + k] = tmp;
			}
			tmp = x[row];
			x[row] = x[col];
			x[col] = tmp;
		}

        // scale the row so that the pivot entry is 1
        T scale = ((T)1.0) / inv[col*dim + col];
        inv[col*dim + col] = (T)1.0;
        for (i2 = 0; i2 < dim; i2++)
            inv[col*dim + i2] *= scale;
		x[col] *= scale;

        // zero out the pivot column locations in the other rows
        for (i1 = 0; i1 < dim; i1++) {
            if ( i1 == col )
				continue;
            T tmp = inv[i1*dim + col];
            inv[i1*dim + col] = T(0);
            for (i2 = 0; i2 < dim; i2++)
                inv[i1*dim + i2] -= inv[col*dim + i2]*tmp;
			x[i1] -= x[col]*tmp;
        }
    }

    delete[] pivoted;
	delete[] inv;
    return true;

error:
    delete[] pivoted;
	delete[] inv;
    return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
bool LinearSolver<T>::invert(unsigned int dim, const T* mtx, T* inv) {
	CopyElements(inv, mtx, dim*dim);

    unsigned int* colidx = new unsigned int[dim];
    ASSERT( colidx );

    unsigned int* rowidx = new unsigned int[dim];
    ASSERT( rowidx );

    bool* pivoted = new bool[dim];
    ASSERT( pivoted );
    memset(pivoted,0,dim*sizeof(bool));

    unsigned int i0, i1, i2, row = 0, col = 0;

    // elimination by full pivoting
    for (i0 = 0; i0 < dim; i0++) {
        // search matrix (excluding pivoted rows) for maximum absolute entry
        T maxv = T(0);
        for (i1 = 0; i1 < dim; i1++) {
            if ( pivoted[i1] ) 
				continue;

            for (i2 = 0; i2 < dim; i2++) {
                if ( pivoted[i2] )
					continue;
                T absv = abs(inv[i1*dim + i2]);
                if ( absv > maxv ) {
                    maxv = absv;
                    row = i1;
                    col = i2;
                }
            }
        }

        if ( maxv == (T)0.0 )
			goto error;

        pivoted[col] = true;

        // swap rows so that A[col][col] contains the pivot entry
        if ( row != col ) {
			T tmp;
			for (i1 = 0; i1 < dim; i1++) {
				tmp = inv[row*dim + i1];
				inv[row*dim + i1] = inv[col*dim + i1];
				inv[col*dim + i1] = tmp;
			}
		}

        // keep track of the permutations of the rows
        rowidx[i0] = row;
        colidx[i0] = col;

        // scale the row so that the pivot entry is 1
        T scale = ((T)1.0) / inv[col*dim + col];
        inv[col*dim + col] = (T)1.0;
        for (i2 = 0; i2 < dim; i2++)
            inv[col*dim + i2] *= scale;

        // zero out the pivot column locations in the other rows
        for (i1 = 0; i1 < dim; i1++) {
            if ( i1 == col )
				continue;
            T tmp = inv[i1*dim + col];
            inv[i1*dim + col] = T(0);
            for (i2 = 0; i2 < dim; i2++)
                inv[i1*dim + i2] -= inv[col*dim + i2]*tmp;
        }
    }

    // reorder rows so that A[][] stores the inverse of the original matrix
    for (i1 = dim-1; i1 < dim; i1--) {
        if ( rowidx[i1] == colidx[i1] )
			continue;
        for (i2 = 0; i2 < dim; i2++) {
            T tmp = inv[i2*dim + rowidx[i1]];
            inv[i2*dim + rowidx[i1]] = inv[i2*dim + colidx[i1]];
            inv[i2*dim + colidx[i1]] = tmp;
        }
    }

    delete[] colidx;
    delete[] rowidx;
    delete[] pivoted;
    return true;

error:
    delete[] colidx;
    delete[] rowidx;
    delete[] pivoted;
    return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
#if 0
template<class T> 
bool LinearSolver<T>::invert(unsigned int dim, const T* omtx, T* inv) {
	unsigned int i,j,k;
	
	T* mtx = new T[dim*dim];
	CopyElements(mtx, omtx, dim*dim);
	
	for (i = 0; i < dim; i++)
		for (j = 0; j < dim; j++)
		 inv[i*dim+j] = (i == j) ? T(1) : T(0);
		 
	for (j = 0; j < dim; j++) {
		int p = getPivot(dim, mtx, j):

		// Help the compiler optimizing by
		// precaalculating pointers to the rows
		T* mtx_jdim = &mtx[j*dim];
		T* inv_jdim = &inv[j*dim];
		T* mtx_pdim = &mtx[p*dim];
		T* inv_pdim = &inv[p*dim];
		if (mtx_pdim[j] == T(0))
			goto error;
			
		// Swap rows p and j
		if (p != j) {
			T ta, tb;
			for (k = 0; k < dim; k++) {
				ta = mtx_pdim[k]
				tb = inv_pdim[k]
				mtx_pdim[k] = mtx_jdim[k];
				inv_pdim[k] = inv_jdim[k];
				mtx_jdim[k] = ta;
				inv_jdim[k] = tb;
			}
		}
		
		T div = T(1.0) / mtx [j*dim + j];
		for (k = j+1; k < dim; k++) {
			mtx_jdim[k] *= div;
			inv_jdim[k] *= div;
		}
		
		for (i = 0; i < dim; i++) {
			if (i == j)
				continue;
			T* mtx_idim = &mtx[i*dim];
			T* inv_idim = &inv[i*dim];
			T mtx_idimj = mtx_idim[j];
			for (k = j+1; k < dim; k++) {
				mtx_idim[k] -= mtx_idimj * mtx_jdim[k];
				inv_idim[k] -= mtx_idimj * inv_jdim[k];
			}
		}
	}
	
	delete[] mtx;
	return true;
	
error:
	delete[] mtx;
	return false;
}
#endif


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
int LinearSolver<T>::getPivot(unsigned int dim, const T* mtx, unsigned int r) {
	int p = r;
	T K = big number;
	T max, a;
	
	for (unsigned int i = r; i < dim; i++) {
		if (mtx[i*dim+r] == 0)
			continue;
			
		max = T(0);
		for (unsigned int j = r+1; j < dim; j++) {
			a = abs(mtx[i*dim + j]);
			if (a > max)
				max = a;
		}
		
		a = max / abs(mtx[i*dim + r]);
		if (a < K) {
			K = a;
			p = i;
		}
	}
	
	return p;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool LinearSolver<T>::solve2(const T* mtx, const T* B, T* X) {
    T det = mtx[0]*mtx[3] - mtx[1]*mtx[2];
    if ( abs(det) == T(0) )
        return false;

    det = ((T)1.0)/det;
    X[0] = (mtx[3]*B[0] - mtx[1]*B[1])*det;
    X[1] = (mtx[0]*B[1] - mtx[2]*B[0])*det;
    return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool LinearSolver<T>::solve2(const T* mtx, T* B) {
    T det = mtx[0]*mtx[3] - mtx[1]*mtx[2];
    if ( abs(det) == T(0) )
        return false;

    det = ((T)1.0)/det;
    T tmp = (mtx[3]*B[0] - mtx[1]*B[1])*det;
    B[1] = (mtx[0]*B[1] - mtx[2]*B[0])*det;
	B[0] = tmp;
    return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline bool LinearSolver<T>::invert2(const T* mtx, T* inv) {
    T det = mtx[0]*mtx[3] - mtx[1]*mtx[2];
    if ( abs(det) == T(0) )
        return false;

    det = ((T)1.0)/det;
	inv[0] = mtx[3]*det;
	inv[1] = -mtx[1]*det;
	inv[2] = mtx[0]*det;
	inv[3] = -mtx[2]*det;
    return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template <class T> 
bool LinearSolver<T>::solve3 (const T* mtx, const T* B, T* X) {
    T inv[3*3 + 3];

    inv[0][0] = mtx[1*3 + 1]*mtx[2*3 + 2] - mtx[1*3 + 2]*mtx[2*3 + 1];
    inv[0][1] = mtx[0*3 + 2]*mtx[2*3 + 1] - mtx[0*3 + 1]*mtx[2*3 + 2];
    inv[0][2] = mtx[0*3 + 1]*mtx[1*3 + 2] - mtx[0*3 + 2]*mtx[1*3 + 1];
    inv[1][0] = mtx[1*3 + 2]*mtx[2*3 + 0] - mtx[1*3 + 0]*mtx[2*3 + 2];
    inv[1][1] = mtx[0*3 + 0]*mtx[2*3 + 2] - mtx[0*3 + 2]*mtx[2*3 + 0];
    inv[1][2] = mtx[0*3 + 2]*mtx[1*3 + 0] - mtx[0*3 + 0]*mtx[1*3 + 2];
    inv[2][0] = mtx[1*3 + 0]*mtx[2*3 + 1] - mtx[1*3 + 1]*mtx[2*3 + 0];
    inv[2][1] = mtx[0*3 + 1]*mtx[2*3 + 0] - mtx[0*3 + 0]*mtx[2*3 + 1];
    inv[2][2] = mtx[0*3 + 0]*mtx[1*3 + 1] - mtx[0*3 + 1]*mtx[1*3 + 0];
    T det = mtx[0*3 + 0]*inv[0][0] + mtx[0*3 + 1]*inv[1][0] + mtx[0*3 + 2]*inv[2][0];
    if ( abs(det) == T(0) )
        return false;

    det = ((T)1.0)/det;
 
    X[0] = (inv[0*3 + 0]*B[0] + inv[0*3 + 1]*B[1] + inv[0*3 + 2]*B[2]) * det;
    X[1] = (inv[1*3 + 0]*B[0] + inv[1*3 + 1]*B[1] + inv[1*3 + 2]*B[2]) * det;
    X[2] = (inv[2*3 + 0]*B[0] + inv[2*3 + 1]*B[1] + inv[2*3 + 2]*B[2]) * det;
    return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template <class T> 
bool LinearSolver<T>::solve3 (const T* mtx, T* B) {
    T inv[3*3 + 3];

    inv[0][0] = mtx[1*3 + 1]*mtx[2*3 + 2] - mtx[1*3 + 2]*mtx[2*3 + 1];
    inv[0][1] = mtx[0*3 + 2]*mtx[2*3 + 1] - mtx[0*3 + 1]*mtx[2*3 + 2];
    inv[0][2] = mtx[0*3 + 1]*mtx[1*3 + 2] - mtx[0*3 + 2]*mtx[1*3 + 1];
    inv[1][0] = mtx[1*3 + 2]*mtx[2*3 + 0] - mtx[1*3 + 0]*mtx[2*3 + 2];
    inv[1][1] = mtx[0*3 + 0]*mtx[2*3 + 2] - mtx[0*3 + 2]*mtx[2*3 + 0];
    inv[1][2] = mtx[0*3 + 2]*mtx[1*3 + 0] - mtx[0*3 + 0]*mtx[1*3 + 2];
    inv[2][0] = mtx[1*3 + 0]*mtx[2*3 + 1] - mtx[1*3 + 1]*mtx[2*3 + 0];
    inv[2][1] = mtx[0*3 + 1]*mtx[2*3 + 0] - mtx[0*3 + 0]*mtx[2*3 + 1];
    inv[2][2] = mtx[0*3 + 0]*mtx[1*3 + 1] - mtx[0*3 + 1]*mtx[1*3 + 0];
    T det = mtx[0*3 + 0]*inv[0][0] + mtx[0*3 + 1]*inv[1][0] + mtx[0*3 + 2]*inv[2][0];
    if ( abs(det) == T(0) )
        return false;

    det = ((T)1.0)/det;
 
    T tmp0 = (inv[0*3 + 0]*B[0] + inv[0*3 + 1]*B[1] + inv[0*3 + 2]*B[2]) * det;
    T tmp1 = (inv[1*3 + 0]*B[0] + inv[1*3 + 1]*B[1] + inv[1*3 + 2]*B[2]) * det;
    T tmp2 = (inv[2*3 + 0]*B[0] + inv[2*3 + 1]*B[1] + inv[2*3 + 2]*B[2]) * det;
	B[0] = tmp0;
	B[1] = tmp1;
	B[2] = tmp2;
    return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template <class T> 
bool LinearSolver<T>::invert3 (const T* mtx, T* inv) {
    T inv[3*3 + 3];

    inv[0*3 + 0] = mtx[1*3 + 1]*mtx[2*3 + 2] - mtx[1*3 + 2]*mtx[2*3 + 1];
    inv[1*3 + 0] = mtx[1*3 + 2]*mtx[2*3 + 0] - mtx[1*3 + 0]*mtx[2*3 + 2];
    inv[2*3 + 0] = mtx[1*3 + 0]*mtx[2*3 + 1] - mtx[1*3 + 1]*mtx[2*3 + 0];
    T det = mtx[0*3 + 0]*inv[0*3 + 0] + mtx[0*3 + 1]*inv[1*3 + 0] + mtx[0*3 + 2]*inv[2*3 + 0];
    if ( abs(det) == T(0) )
        return false;

    det = ((T)1.0)/det;

	inv[0*3 + 0] *= det;
	inv[1*3 + 0] *= det;
	inv[2*3 + 0] *= det;

    inv[0*3 + 1] = (mtx[0*3 + 2]*mtx[2*3 + 1] - mtx[0*3 + 1]*mtx[2*3 + 2]) * det;
    inv[0*3 + 2] = (mtx[0*3 + 1]*mtx[1*3 + 2] - mtx[0*3 + 2]*mtx[1*3 + 1]) * det;
    inv[1*3 + 1] = (mtx[0*3 + 0]*mtx[2*3 + 2] - mtx[0*3 + 2]*mtx[2*3 + 0]) * det;
    inv[1*3 + 2] = (mtx[0*3 + 2]*mtx[1*3 + 0] - mtx[0*3 + 0]*mtx[1*3 + 2]) * det;
    inv[2*3 + 1] = (mtx[0*3 + 1]*mtx[2*3 + 0] - mtx[0*3 + 0]*mtx[2*3 + 1]) * det;
    inv[2*3 + 2] = (mtx[0*3 + 0]*mtx[1*3 + 1] - mtx[0*3 + 1]*mtx[1*3 + 0]) * det;

    return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template <class T> 
bool LinearSolver<T>::solve4(const T* mtx, const T* b, T* x) {
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template <class T> 
bool LinearSolver<T>::solve4(const T* mtx, T* b) {
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template <class T> 
bool LinearSolver<T>::invert4(const T* mtx, T* inv) {
	T tmp[12]; /* temp array for pairs */
	T dst[16];
	T det; /* determinant */

	/* calculate pairs for first 8 elements (cofactors) */
	tmp[0] = mtx[10] * mtx[15];
	tmp[1] = mtx[14] * mtx[11];
	tmp[2] = mtx[6] * mtx[15];
	tmp[3] = mtx[14] * mtx[7];
	tmp[4] = mtx[6] * mtx[11];
	tmp[5] = mtx[10] * mtx[7];
	tmp[6] = mtx[2] * mtx[15];
	tmp[7] = mtx[14] * mtx[3];
	tmp[8] = mtx[2] * mtx[11];
	tmp[9] = mtx[10] * mtx[3];
	tmp[10] = mtx[2] * mtx[7];
	tmp[11] = mtx[6] * mtx[3];

	/* calculate first 8 elements (cofactors) */
	dst[0] = tmp[0]*mtx[5] + tmp[3]*mtx[9] + tmp[4]*mtx[13];
	dst[0] -= tmp[1]*mtx[5] + tmp[2]*mtx[9] + tmp[5]*mtx[13];
	dst[1] = tmp[1]*mtx[1] + tmp[6]*mtx[9] + tmp[9]*mtx[13];
	dst[1] -= tmp[0]*mtx[1] + tmp[7]*mtx[9] + tmp[8]*mtx[13];
	dst[2] = tmp[2]*mtx[1] + tmp[7]*mtx[5] + tmp[10]*mtx[13];
	dst[2] -= tmp[3]*mtx[1] + tmp[6]*mtx[5] + tmp[11]*mtx[13];
	dst[3] = tmp[5]*mtx[1] + tmp[8]*mtx[5] + tmp[11]*mtx[9];
	dst[3] -= tmp[4]*mtx[1] + tmp[9]*mtx[5] + tmp[10]*mtx[9];
	dst[4] = tmp[1]*mtx[4] + tmp[2]*mtx[8] + tmp[5]*mtx[12];
	dst[4] -= tmp[0]*mtx[4] + tmp[3]*mtx[8] + tmp[4]*mtx[12];
	dst[5] = tmp[0]*mtx[0] + tmp[7]*mtx[8] + tmp[8]*mtx[12];
	dst[5] -= tmp[1]*mtx[0] + tmp[6]*mtx[8] + tmp[9]*mtx[12];
	dst[6] = tmp[3]*mtx[0] + tmp[6]*mtx[4] + tmp[11]*mtx[12];
	dst[6] -= tmp[2]*mtx[0] + tmp[7]*mtx[4] + tmp[10]*mtx[12];
	dst[7] = tmp[4]*mtx[0] + tmp[9]*mtx[4] + tmp[10]*mtx[8];
	dst[7] -= tmp[5]*mtx[0] + tmp[8]*mtx[4] + tmp[11]*mtx[8];

	/* calculate pairs for second 8 elements (cofactors) */
	tmp[0] = mtx[8]*mtx[13];
	tmp[1] = mtx[12]*mtx[9];
	tmp[2] = mtx[4]*mtx[13];
	tmp[3] = mtx[12]*mtx[5];
	tmp[4] = mtx[4]*mtx[9];
	tmp[5] = mtx[8]*mtx[5];
	tmp[6] = mtx[0]*mtx[13];
	tmp[7] = mtx[12]*mtx[1];
	tmp[8] = mtx[0]*mtx[9];
	tmp[9] = mtx[8]*mtx[1];
	tmp[10] = mtx[0]*mtx[5];
	tmp[11] = mtx[4]*mtx[1];

	/* calculate second 8 elements (cofactors) */
	dst[8] = tmp[0]*mtx[7] + tmp[3]*mtx[11] + tmp[4]*mtx[15];
	dst[8] -= tmp[1]*mtx[7] + tmp[2]*mtx[11] + tmp[5]*mtx[15];
	dst[9] = tmp[1]*mtx[3] + tmp[6]*mtx[11] + tmp[9]*mtx[15];
	dst[9] -= tmp[0]*mtx[3] + tmp[7]*mtx[11] + tmp[8]*mtx[15];
	dst[10] = tmp[2]*mtx[3] + tmp[7]*mtx[7] + tmp[10]*mtx[15];
	dst[10]-= tmp[3]*mtx[3] + tmp[6]*mtx[7] + tmp[11]*mtx[15];
	dst[11] = tmp[5]*mtx[3] + tmp[8]*mtx[7] + tmp[11]*mtx[11];
	dst[11]-= tmp[4]*mtx[3] + tmp[9]*mtx[7] + tmp[10]*mtx[11];
	dst[12] = tmp[2]*mtx[10] + tmp[5]*mtx[14] + tmp[1]*mtx[6];
	dst[12]-= tmp[4]*mtx[14] + tmp[0]*mtx[6] + tmp[3]*mtx[10];
	dst[13] = tmp[8]*mtx[14] + tmp[0]*mtx[2] + tmp[7]*mtx[10];
	dst[13]-= tmp[6]*mtx[10] + tmp[9]*mtx[14] + tmp[1]*mtx[2];
	dst[14] = tmp[6]*mtx[6] + tmp[11]*mtx[14] + tmp[3]*mtx[2];
	dst[14]-= tmp[10]*mtx[14] + tmp[2]*mtx[2] + tmp[7]*mtx[6];
	dst[15] = tmp[10]*mtx[10] + tmp[4]*mtx[2] + tmp[9]*mtx[6];
	dst[15]-= tmp[8]*mtx[6] + tmp[11]*mtx[10] + tmp[5]*mtx[2];

	/* calculate determinant */
	det=mtx[0]*dst[0]+mtx[4]*dst[1]+mtx[8]*dst[2]+mtx[12]*dst[3];
	/* calculate matrix inverse */
	det = 1/det;
	for ( int i = 0; i < 4; i++) {
		inv[i*4 + 0] = dst[i*4]*det;
		inv[i*4 + 1] = dst[i*4 + 1]*det;
		inv[i*4 + 2] = dst[i*4 + 2]*det;
		inv[i*4 + 3] = dst[i*4 + 3]*det;
	}
}

}; };

#endif
