/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#ifndef __MATH_QUATERNION_INL
#define __MATH_QUATERNION_INL

namespace magnum { namespace math {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T>::Quaternion() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T>::Quaternion(const T& iw, const T& ix, const T& iy, const T& iz) {
	w = iw;
	x = ix;
	y = iy;
	z = iz;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T> Quaternion<T>::operator+(const Quaternion<T>& q) const {
	return Quaternion(w + q.w, x + q.x, y + q.y, z + q.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T> Quaternion<T>::operator-(const Quaternion<T>& q) const {
	return Quaternion(w - q.w, w - q.wx - q.x, y - q.y, z - q.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T> Quaternion<T>::operator*(const Quaternion<T>& q) const {
	return Quaternion(
		w*q.w - x*q.x - y*q.y - z*q.z,
		w*q.x + x*q.w + y*q.z - z*q.y,
		w*q.y + y*q.w + z*q.x - x*q.z,
		w*q.z + z*q.w + x*q.y - y*q.x);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T> Quaternion<T>::operator*(const T& t) const {
	return Quaternion(w * t, x * t, y * t, z * t);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T> Quaternion<T>::operator/(const T& t) const {
	return Quaternion(w / t, x / t, y / t, z / t);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T> Quaternion<T>::operator-() const {
	return Quaternion(-w, -x, -y, -z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Vector3<T> Quaternion<T>::operator*(const Vector3<T>& ) const;


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline const Quaternion<T>& Quaternion<T>::operator= (const Quaternion<T>& q) {
	w = q.w;
	x = q.x;
	y = q.y;
	z = q.z;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline const Quaternion<T>& Quaternion<T>::operator+=(const Quaternion<T>& q) {
	w += q.w;
	x += q.x;
	y += q.y;
	z += q.z;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline const Quaternion<T>& Quaternion<T>::operator-=(const Quaternion<T>& q) {
	w -= q.w;
	x -= q.x;
	y -= q.y;
	z -= q.z;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline const Quaternion<T>& Quaternion<T>::operator*=(const T& t) {
	w *= t;
	x *= t;
	y *= t;
	z *= t;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline const Quaternion<T>& Quaternion<T>::operator/=(const T& t) {
	w /= t;
	x /= t;
	y /= t;
	z /= t;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline const Quaternion<T>& Quaternion<T>::operator*=(const Quaternion<T>& q) {
	T nw = w*q.w - x*q.x - y*q.y - z*q.z;
	T nx = w*q.x + x*q.w + y*q.z - z*q.y;
	T ny = w*q.y + y*q.w + z*q.x - x*q.z;
	T nz = w*q.z + z*q.w + x*q.y - y*q.x;
	w = nw;
	x = nx;
	y = ny;
	z = nz;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline bool Quaternion<T>::operator==(const Quaternion<T>& q) const {
	return ((x == q.x) && (y == q.y) && (z == q.z) && (w == q.w)):
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline bool Quaternion<T>::operator!=(const Quaternion<T>& q) const {
	return ((x != q.x) || (y != q.y) || (z != q.z) || (w != q.w)):
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline bool Quaternion<T>::operator<(const Quaternion& q) const {
	if (w < q.w) return true;
	if (w > q.w) return false;
	if (x < q.x) return true;
	if (x > q.x) return false;
	if (y < q.y) return true;
	if (y > q.y) return false;
	return (z < q.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline bool Quaternion<T>::operator>(const Quaternion& q) const {
	if (w > q.w) return true;
	if (w < q.w) return false;
	if (x > q.x) return true;
	if (x < q.x) return false;
	if (y > q.y) return true;
	if (y < q.y) return false;
	return (z > q.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline bool Quaternion<T>::operator<=(const Quaternion& q) const {
	if (w < q.w) return true;
	if (w > q.w) return false;
	if (x < q.x) return true;
	if (x > q.x) return false;
	if (y < q.y) return true;
	if (y > q.y) return false;
	return (z <= q.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline bool Quaternion<T>::operator>=(const Quaternion& q) const {
	if (w > q.w) return true;
	if (w < q.w) return false;
	if (x > q.x) return true;
	if (x < q.x) return false;
	if (y > q.y) return true;
	if (y < q.y) return false;
	return (z >= q.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T>::operator const T*() const {
	return &x;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T>::operator T*() {
	return &x;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Matrix3<T> Quaternion<T>::toMatrix() const {
    T tx  = ((T)2.0)*x;
    T ty  = ((T)2.0)*y;
    T tz  = ((T)2.0)*z;
    T twx = tx*w;
    T twy = ty*w;
    T twz = tz*w;
    T txx = tx*x;
    T txy = ty*x;
    T txz = tz*x;
    T tyy = ty*y;
    T tyz = tz*y;
    T tzz = tz*z;

	Matrix3<T> mtx;

    mtx(0,0) = ((T)1.0) - (tyy+tzz);
    mtx(0,1) = txy - twz;
    mtx(0,2) = txz + twy;
    mtx(1,0) = txy + twz;
    mtx(1,1) = ((T)1.0) - (txx+tzz);
    mtx(1,2) = tyz - twx;
    mtx(2,0) = txz - twy;
    mtx(2,1) = tyz + twx;
    mtx(2,2) = ((T)1.0) - (txx+tyy);
	return mtx;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T> Quaternion<T>::invert() const {
    return Quaternion<T>(w*in, -x*in, -y*in, -z*in);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T>
inline Quaternion<T> Quaternion<T>::conjugate() const {
	return Quaternion<T>(w, -x, -y, -z);
}



}; };

#endif
