/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#ifndef __MATH_VECTOR2_INL
#define __MATH_VECTOR2_INL

namespace magnum {namespace math {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> Vector2<T> Vector2<T>::Null = Vector2<T>(T(0), T(0));


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector2<T>::Vector2() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector2<T>::Vector2(const T* s) {
	x = s[0];
	y = s[1];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector2<T>::Vector2(const Vector<T,2>& s) {
	x = s[0];
	y = s[1];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector2<T>::Vector2(const T& t1, const T& t2) {
	x = t1;
	y = t2;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector2<T> Vector2<T>::operator+(const Vector2<T> &u) const {
	return Vector2(x + u.x, y + u.y);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector2<T> Vector2<T>::operator+(const T& c) const {
	return Vector2(x + c, y + c);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector2<T> Vector2<T>::operator-(const Vector2<T> &u) const {
	return Vector2(x - u.x, y - u.y);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector2<T> Vector2<T>::operator-(const T& c) const {
	return Vector2(x - c, y - c);
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector2<T> Vector2<T>::operator-() const {
	return Vector2(-x, -y);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector2<T> Vector2<T>::operator*(const T& s) const {
	return Vector2(s*x, s*y);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector2<T> operator*(const T& s,const Vector2<T>& v) {
	return Vector2<T>(s*v.x, s*v.y);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector2<T> operator*(const Matrix<T,3>& m,const Vector2<T>& v) {
	return Vector2<T>(
		m[0][0]*v.x + m[0][1]*v.y + m[0][2],
		m[1][0]*v.x + m[1][1]*v.y + m[1][2]);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector2<T> operator*(const Matrix2<T>& mtx, const Vector2<T>& v) {
	return Vector2<T>(
		v.x*mtx[0][0] + v.y*mtx[0][1], 
		v.x*mtx[1][0] + v.y*mtx[1][1]);
}

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline T Vector2<T>::operator*(const Vector2<T>& v) const {
	return x*v.x + y*v.y;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector2<T> Vector2<T>::operator/(const T& s) const {
	return Vector2(x / s, y / s);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline const Vector2<T>& Vector2<T>::operator+= (const T& v) {
	x += v;
	y += v;
	return *this;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline const Vector2<T>& Vector2<T>::operator+= (const Vector2<T>& v) {
	x += v.x;
	y += v.y;
	return *this;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline const Vector2<T>& Vector2<T>::operator-= (const T& s) {
	x -= s;
	y -= s;
	return *this;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline const Vector2<T>& Vector2<T>::operator-= (const Vector2<T>& v) {
	x -= v.x;
	y -= v.y;
	return *this;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline const Vector2<T>& Vector2<T>::operator*= (const T& s) {
	x *= s;
	y *= s;
	return *this;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline const Vector2<T>& Vector2<T>::operator/= (const T& s) {
	x /= s;
	y /= s;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline const Vector2<T>& Vector2<T>::operator=(const Vector<T,2>& v) {
	x = v.vec[0];
	y = v.vec[1];
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline bool Vector2<T>::operator==(const Vector2<T>& v) const {
	return ((x == v.x) && (y == v.y));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline bool Vector2<T>::operator!=(const Vector2<T>& v) const {
	return ((x != v.x) || (y != v.y));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline bool Vector2<T>::operator<(const Vector2<T>& v) const {
	if (x < v.x) return true;
	if (x > v.x) return false;
	return (y < v.y);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline bool Vector2<T>::operator>(const Vector2<T>& v) const {
	if (x > v.x) return true;
	if (x < v.x) return false;
	return (y > v.y);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline bool Vector2<T>::operator<=(const Vector2<T>& v) const {
	if (x < v.x) return true;
	if (x > v.x) return false;
	return (y <= v.y);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline bool Vector2<T>::operator>=(const Vector2<T>& v) const {
	if (x > v.x) return true;
	if (x < v.x) return false;
	return (y >= v.y);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector2<T>::operator const T*() const {
	return &x;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector2<T>::operator T*() {
	return &x;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector2<T>::operator Vector<T,2>() const {
	return Vector<T,2>(&x);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector2<T>& Vector2<T>::null() {
	x = y = T(0);
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector2<T>& Vector2<T>::normalize() {
	T l = x*x + y*y;
	if (l > T(0)) {
		l = 1/sqrt(l);
		x *= l;
		y *= l;
	}
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline T Vector2<T>::length() const {
	return (T)sqrt(x*x + y*y);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline T Vector2<T>::length2() const {
	return (x*x + y*y);
}

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline T Vector2<T>::max() const {
	return (x < y) ? y : x;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline T Vector2<T>::min() const {
	return (x < y) ? x : y;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline T Vector2<T>::maxabs() const {
	T xa = abs(x);
	T ya = abs(y);
	return (xa < ya) ? ya : xa;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline T Vector2<T>::minabs() const {
	T xa = abs(x);
	T ya = abs(y);
	return (xa < ya) ? xa : ya;
}


}; };

#endif
