/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#ifndef __MATH_VECTOR3_INL
#define __MATH_VECTOR3_INL

namespace magnum {namespace math {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> Vector3<T> Vector3<T>::Null = Vector3<T>(T(0), T(0), T(0));


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector3<T>::Vector3() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector3<T>::Vector3(const T* s) {
	x = s[0];
	y = s[1];
	z = s[2];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector3<T>::Vector3(const Vector<T,3>& s) {
	x = s[0];
	y = s[1];
	z = s[2];
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector3<T>::Vector3(const T& t1, const T& t2, const T& t3) {
	x = t1;
	y = t2;
	z = t3;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector3<T>::Vector3(const Vector2<T>& v) {
	x = v.x;
	y = v.y;
	z = T(0);
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector3<T> Vector3<T>::operator+(const Vector3<T> &u) const {
	return Vector3(x + u.x, y + u.y, z + u.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector3<T> Vector3<T>::operator+(const T& c) const {
	return Vector3(x + c, y + c, z + c);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector3<T> Vector3<T>::operator-(const Vector3<T> &u) const {
	return Vector3(x - u.x, y - u.y, z - u.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector3<T> Vector3<T>::operator-(const T& c) const {
	return Vector3(x - c, y - c, z - c);
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector3<T> Vector3<T>::operator-() const {
	return Vector3(-x, -y, -z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector3<T> Vector3<T>::operator*(const T& s) const {
	return Vector3(s*x, s*y, s*z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector3<T> operator*(const T& s,const Vector3<T>& v) {
	return Vector3<T>(s*v.x, s*v.y, s*v.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector3<T> operator*(const Matrix<T,4>& m,const Vector3<T>& v) {
	return Vector3<T>(
		m[0][0]*v.x + m[0][1]*v.y + m[0][2]*v.z + m[0][3],
		m[1][0]*v.x + m[1][1]*v.y + m[1][2]*v.z + m[1][3],
		m[2][0]*v.x + m[2][1]*v.y + m[2][2]*v.z + m[2][3]);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector3<T> operator*(const Matrix3<T>& mtx, const Vector3<T>& v) {
	return Vector3<T>(
		v.x*mtx[0][0] + v.y*mtx[0][1] + v.z*mtx[0][2], 
		v.x*mtx[1][0] + v.y*mtx[1][1] + v.z*mtx[1][2],
		v.x*mtx[2][0] + v.y*mtx[2][1] + v.z*mtx[2][2]);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline T Vector3<T>::operator*(const Vector3<T>& v) const {
	return x*v.x + y*v.y + z*v.z;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector3<T> Vector3<T>::operator/(const T& s) const {
	return Vector3(x / s, y / s, z / s);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector3<T> Vector3<T>::operator%(const Vector3<T>& s) const {
	return Vector3(y*s.z - z*s.y, z*s.x - x*s.z, x*s.y - y*s.x);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline const Vector3<T>& Vector3<T>::operator+= (const T& s) {
	x += s;
	y += s;
	z += s;
	return *this;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline const Vector3<T>& Vector3<T>::operator+= (const Vector3<T>& v) {
	x += v.x;
	y += v.y;
	z += v.z;
	return *this;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline const Vector3<T>& Vector3<T>::operator-= (const T& s) {
	x -= s;
	y -= s;
	z -= s;
	return *this;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline const Vector3<T>& Vector3<T>::operator-= (const Vector3<T>& v) {
	x -= v.x;
	y -= v.y;
	z -= v.z;
	return *this;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline const Vector3<T>& Vector3<T>::operator*= (const T& s) {
	x *= s;
	y *= s;
	z *= s;
	return *this;
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline const Vector3<T>& Vector3<T>::operator/= (const T& s) {
	x /= s;
	y /= s;
	z /= s;
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline const Vector3<T>& Vector3<T>::operator=(const Vector<T,3>& v) {
	x = v.vec[0];
	y = v.vec[1];
	z = v.vec[2];
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline const Vector3<T>& Vector3<T>::operator=(const Vector<T,2>& v) {
	x = v.vec[0];
	y = v.vec[1];
	z = T(0);
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline const Vector3<T>& Vector3<T>::operator=(const Vector2<T>& v) {
	x = v.x;
	y = v.y;
	z = T(0);
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline bool Vector3<T>::operator==(const Vector3<T>& v) const {
	return ((x == v.x) && (y == v.y) && (z == v.z));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline bool Vector3<T>::operator!=(const Vector3<T>& v) const {
	return ((x != v.x) || (y != v.y) || (z != v.z));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline bool Vector3<T>::operator<(const Vector3<T>& v) const {
	if (x < v.x) return true;
	if (x > v.x) return false;
	if (y < v.y) return true;
	if (y > v.y) return false;
	return (z < v.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline bool Vector3<T>::operator>(const Vector3<T>& v) const {
	if (x > v.x) return true;
	if (x < v.x) return false;
	if (y > v.y) return true;
	if (y < v.y) return false;
	return (z > v.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline bool Vector3<T>::operator<=(const Vector3<T>& v) const {
	if (x < v.x) return true;
	if (x > v.x) return false;
	if (y < v.y) return true;
	if (y > v.y) return false;
	return (z <= v.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline bool Vector3<T>::operator>=(const Vector3<T>& v) const {
	if (x > v.x) return true;
	if (x < v.x) return false;
	if (y > v.y) return true;
	if (y < v.y) return false;
	return (z >= v.z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector2<T>	Vector3<T>::xy() const {
	return Vector2<T>(x,y);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector2<T>	Vector3<T>::xz() const {
	return Vector2<T>(x,z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector2<T>	Vector3<T>::yz() const {
	return Vector2<T>(y,z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector3<T>::operator const T*() const {
	return &x;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector3<T>::operator T*() {
	return &x;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector3<T>::operator Vector<T,3>() const {
	return Vector<T,3>(&x);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline Vector3<T>& Vector3<T>::null() {
	x = y = z = T(0);
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> 
inline Vector3<T>& Vector3<T>::normalize() {
	T l = x*x + y*y + z*z;
	if (l > T(0)) {
		l = T(1)/sqrt(l);
		x *= l;
		y *= l;
		z *= l;
	}
	return *this;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline T Vector3<T>::length() const {
	return (T)sqrt(x*x + y*y + z*z);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline T Vector3<T>::length2() const {
	return (x*x + y*y + z*z);
}

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline T Vector3<T>::max() const {
	if (x > y)
		return (x > z) ? x : z;
	else
		return (y > z) ? y : z;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline T Vector3<T>::min() const {
	if (x < y)
		return (x < z) ? x : z;
	else
		return (y < z) ? y : z;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline T Vector3<T>::maxabs() const {
	T xa = abs(x);
	T ya = abs(y);
	T za = abs(z);
	if (xa > ya)
		return (xa > za) ? xa : za;
	else
		return (ya > za) ? ya : za;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> inline T Vector3<T>::minabs() const {
	T xa = abs(x);
	T ya = abs(y);
	T za = abs(z);
	if (xa < ya)
		return (xa < za) ? xa : za;
	else
		return (ya < za) ? ya : za;
}


}; };

#endif
