/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#ifndef __NET_SOCKET_H
#define __NET_SOCKET_H

#include "io.inputstream.h"
#include "io.outputstream.h"
#include "net.address.h"

# if defined(__FORWINDOWS__)
#	define SHUT_RD		SD_RECEIVE
#	define SHUT_WR		SD_SEND
#	define SHUT_RDWR	SD_BOTH
#	pragma comment (lib, "ws2_32.lib")
#else
#	include <unistd.h>
#	include <errno.h>
#	include <sys/socket.h>
#	include <sys/ioctl.h>
#	include <sys/time.h>
#	include <netinet/in.h>
#	include <netinet/tcp.h>
#	define SOCKET		int
#	define SOCKET_ERROR (-1)
# endif


namespace magnum { namespace net {
using magnum::object::Object;
using magnum::types::String;
using magnum::io::InputStream;
using magnum::io::OutputStream;
using magnum::io::StreamFactory;

class DatagramSocket;
class MulticastSocket;
class ServerSocket;
class Socket;


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
class Socket : public StreamFactory {
DeclareClass(Socket, StreamFactory);
class SocketInputStream;
class SocketOutputStream;
friend class DatagramSocket;
friend class MulticastSocket;
friend class ServerSocket;
friend class SocketInputStream;
friend class SocketOutputStream;
private:
	class Startup {
		public:
			Startup();
			~Startup();
	};

	class SocketInputStream : public InputStream {
		DeclareClass(SocketInputStream, InputStream);
		friend class Socket;
		private:
			Socket*	m_Socket;

		public:
			unsigned long read(void *buffer, unsigned long size);
			unsigned long skip(unsigned long size);
			unsigned long available();
			inline bool	close();
	};

	class SocketOutputStream : public OutputStream {
		DeclareClass(SocketOutputStream, OutputStream);
		friend class Socket;
		private:
			Socket*	m_Socket;

		public:
			unsigned long write(const void *buffer, unsigned long size);
			bool		flush();
			inline bool	close();
	};


	friend class Socket::Startup;

	static Startup			m_Startup;

#if defined(__FORWINDOWS__)
	static WSAPROTOCOL_INFO* m_ProtocolInfo;
	static unsigned int		m_ProtocolInfoCount;
#endif

	SocketInputStream		m_InputStream;
	SocketOutputStream		m_OutputStream;
	Address					m_RemoteAddress;
	unsigned short			m_RemotePort;
	Address					m_LocalAddress;
	unsigned short			m_LocalPort;
	unsigned int			m_SelectIndex;
	bool					m_Connected;
	bool					m_Bound;

protected:
	SOCKET					m_Socket;

public:
	enum WaitMode {
		WaitRead = 1,
		WaitWrite = 2,
	};

	static const char* const ExceptionUnknown;
	static const char* const ExceptionNotInitialised;
	static const char* const ExceptionNetDown;
	static const char* const ExceptionNotConnected;
	static const char* const ExceptionInterrupted;
	static const char* const ExceptionBlocked;
	static const char* const ExceptionNetReset;
	static const char* const ExceptionDescriptor;
	static const char* const ExceptionAddressFault; 
	static const char* const ExceptionShutdown;
	static const char* const ExceptionWouldBlock;
	static const char* const ExceptionMessageSize;
	static const char* const ExceptionNotBound;
	static const char* const ExceptionAddressInUse;
	static const char* const ExceptionAlreadyConnected;
	static const char* const ExceptionConnectionAborted; 
	static const char* const ExceptionConnectionReset;
	static const char* const ExceptionConnectionClosed;
	static const char* const ExceptionConnectionRefused;
	static const char* const ExceptionNoRouteToHost;
	static const char* const ExceptionTimedOut;
	static const char* const ExceptionIOPending; 
	static const char* const ExceptionIOAborted;
	static const char* const ExceptionUnknownHost;
	static const char* const ExceptionInvalidAddress;
	static const char* const ExceptionAddressNotAvailable;
	static const char* const ExceptionBrokenPipe;


protected:
	Socket(unsigned int protocol);
	Socket();

public:
	~Socket();

	bool					bind(const Address& lcoalAdr, unsigned int localPort);
	bool					bind(const String& lcoalAdr, unsigned int localPort);
	bool					bind(unsigned int localPort);
	bool					bind(const Address& lcoalAdr);
	bool					bind(const String& lcoalAdr);

	bool					wait(WaitMode mode);
	bool					wait(WaitMode mode, unsigned int millis);
	 
	bool					connect(const Address& remoteAdr, unsigned int port);
	inline bool				connect(const String& remoteAdr, unsigned int port);
	bool					disconnect();

	inline bool				isConnected() const;
	inline bool				isBound() const;

	inline const Address&	getAddress() const;
	inline unsigned int		getPort() const;

	inline unsigned int		getTimeToLive() const;
	inline void				setTimeToLive(unsigned int ttl);

	inline const Address&	getLocalAddress() const;
	inline unsigned int		getLocalPort() const;

	inline unsigned int		getSendBufferSize() const;
	inline void				setSendBufferSize(unsigned int size);

	inline unsigned int		getReceiveBufferSize() const;
	inline void				setReceiveBufferSize(unsigned int size);

	inline unsigned int		getSendTimeout() const;
	inline void				setSendTimeout(unsigned int to);

	inline unsigned int		getReceiveTimeout() const;
	inline void				setReceiveTimeout(unsigned int to);

	inline InputStream*		getInputStream();
	void					shutdownInput();
	inline OutputStream*	getOutputStream();
	void					shutdownOutput();

private:
	static inline void		throwSocketException();
	static void				throwSocketException(int err);
	static inline void		throwBindException();
	static void				throwBindException(int err);
};


}; };
#endif
