/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "freetype.fontfactory.h"
#include "freetype.font.h"
#include "freetype.descriptor.h"
#include "freetype.exception.h"
#include "io.directory.h"
#include "io.sysfile.h"
#include "container.map.h"

#include "freetype.font.inl"
#include "freetype.glyph.inl"
#include "freetype.fontfactory.inl"
#include "freetype.library.inl"
#include "freetype.descriptor.inl"
#include "base.string.inl"
#include "container.iterator.inl"
#include "container.allocator.inl"
#include "container.array.inl"
#include "container.map.inl"
#include "graphics.point.inl"
#include "graphics.rectangle.inl"
#include "io.directory.inl"


namespace magnum { namespace freetype {
using magnum::io::Directory;
using magnum::io::DirectoryNode;
using magnum::io::SystemFile;
using magnum::container::Map;

/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
FontFactory::FontFactory() {
	setDirectory(".");
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
void FontFactory::setDirectory(const char* d) {
	m_Directory = d;
	unsigned int len = m_Directory.getLength();
	if (len > 0) {
		if (m_Directory[len-1] != '\\')
			m_Directory += "\\";
	}
	
	m_FontInfo.clear();

	// Read in directory
	Directory dir;
	dir.open(m_Directory + "*");
	
	Directory::Iterator iter;
	for (iter = dir.begin(); iter != dir.end(); ++iter) {
		if(iter->isDirectory())
			continue;
		readFontInfo(m_Directory + iter->getName());
	}
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
Font* FontFactory::create(const Descriptor& desc) {
	Array<FontInfo>::Iterator iter, best;
	best = m_FontInfo.end();
	int bestPoints = 0;
	int points;
	
	String family(desc.getFamily().toLowerCase());
	String style(desc.getStyle().toLowerCase());

	for (iter = m_FontInfo.begin(); iter != m_FontInfo.end(); ++iter) {
		points = 0;
		if (iter->m_Family == family)
			points += 100;
		if (iter->m_Style == style)
			points += 50;
		if (points >= bestPoints) {
			best = iter;
			bestPoints = points;
		}
	}

	if (best == m_FontInfo.end())
		throw NoMatchException();
		
	File* file = SystemFile::open(best->m_File);
	if (file == NULL)
		throw FormatException();
		
	Font* font = create(desc, file);
	file->release();
	
	if (font == NULL)
		throw FormatException();
		
	return font;
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
Font* FontFactory::create(const Descriptor& indesc, File* file) {
	Descriptor desc(indesc);
	Dimension<int> dim;

	// Lock library
	FT_Library lib = Library::lock();
	ASSERT(lib != NULL);

	// Open new face
	FT_Open_Args oargs;
	FT_Error error;
	FT_StreamRec stream;
	FT_Face face;
	oargs.flags = FT_OPEN_STREAM;
	oargs.stream = &stream;
	stream.base = NULL;
	stream.size = file->getSize();
    stream.pos = 0;
    stream.descriptor.pointer = file;
    stream.read = &streamRead;
    stream.close = &streamClose;
	error = FT_Open_Face(lib, &oargs, 0, &face);
	if (error != 0) {
		Library::unlock();
		throw FreetypeException();
	}
		
	// Set font name
	ASSERT(face->family_name != NULL)
	desc.setFamily(face->family_name);
	if (face->style_name != NULL)
		desc.setStyle(face->style_name);
	else
		desc.setStyle("");
		
	// Select charmap:
	FT_CharMap charmap;
	int n;
	for (n = 0; n < face->num_charmaps; n++ ) {
		charmap = face->charmaps[n]; 
		if (charmap->encoding_id == desc.getEncoding()) { 
			error = FT_Set_Charmap(face, charmap);
			ASSERT(error == 0);
			break; 
		} 
	}
	if (n == face->num_charmaps) {
		desc.setEncoding(Encoding::None);
	}

	// Set font size
	error = FT_Set_Pixel_Sizes(face, (FT_UInt)desc.getWidth(), (FT_UInt)desc.getHeight());
	ASSERT(error == 0);

	// Estimate needed image size
	FT_GlyphSlot slot = face->glyph; 
	int charcode;
	long maxheight = 0;
	long maxwidth = 0;
	Map<int, int> glyphmap;
	int glyphcnt = 0;
	for (charcode = 0; charcode < 256; charcode++) {
		int gid = FT_Get_Char_Index( face, (char)charcode );
		if (!glyphmap.contains(gid)) {
			glyphmap.insert(gid, glyphcnt);
			glyphcnt++;
		}
		error = FT_Load_Glyph( face, gid, FT_LOAD_TARGET_NORMAL );
		if (error)
			continue;
		
		if (maxheight < slot->metrics.height)
			maxheight = slot->metrics.height;
		if (maxwidth < slot->metrics.width)
			maxwidth = slot->metrics.width;
	}
	
	// Allocate new font object
	maxwidth = ((maxwidth + 63) >> 6) + 1;
	maxheight = ((maxheight + 63) >> 6) + 1;
	dim.setWidth(maxwidth * 16);
	dim.setHeight((glyphcnt + 15)/16 * maxheight);
	Font* font;
	font = new Font(desc, dim);
	ASSERT(font != NULL);

	// Clear image
	byte* data;
	unsigned int pitch;
	data = (byte *)font->getBitmap()->lock();
	pitch = font->getBitmap()->getLinesize();
	memset(data, 0, pitch*font->getBitmap()->getHeight());

	// Render glyphs into image
	Map<int, int>::Iterator iter;
	for (iter = glyphmap.begin(); iter != glyphmap.end(); ++iter) {
		// Load glyph
		int gid = iter.key();
		error = FT_Load_Glyph( face, gid, FT_LOAD_TARGET_NORMAL );
		if (error)
			continue;
			
		// convert to an anti-aliased bitmap
		error = FT_Render_Glyph(face->glyph, ft_render_mode_normal ); 
		if (error) 
			continue;
			
		ASSERT((slot->bitmap.width <= maxwidth) && (slot->bitmap.rows <= maxheight));
		
		// Get glyph location
		gid = iter.item();
		int posx = (gid & 0xf) * maxwidth;
		int posy = (gid >> 4) * maxheight;

		// Copy glyph into bitmap
		byte* dst = data + posy*pitch + posx;
		copyGlyph(&slot->bitmap, dst, pitch);
	}	
	font->getBitmap()->unlock();

	// Get Glyph locations for all characters
	for (charcode = 0; charcode < 256; charcode++) {
		// Load glyph
		int gid = FT_Get_Char_Index( face, (char)charcode );
		error = FT_Load_Glyph( face, gid, FT_LOAD_TARGET_NORMAL );
		if (error)
			continue;
			
		// convert to an anti-aliased bitmap
		error = FT_Render_Glyph(face->glyph, ft_render_mode_normal ); 
		if (error) 
			continue;
			
		// Get glyph location
		gid = glyphmap[gid];
		int posx = (gid & 0xf) * maxwidth;
		int posy = (gid >> 4) * maxheight;
		
		// Set glyph-information
		font->m_Glyphs[charcode].setSize((float)slot->metrics.horiAdvance / 64.0f, (float)slot->metrics.vertAdvance / 64.0f);
		font->m_Glyphs[charcode].setBearing((float)slot->metrics.horiBearingX / 64.0f, (float)slot->metrics.horiBearingY / 64.0f);
		font->m_Glyphs[charcode].setLocation(posx, posy, slot->bitmap.width, slot->bitmap.rows);
	}	
	
	FT_Done_Face(face);
	Library::unlock();
	return font;
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
bool FontFactory::readFontInfo(const char* fname) {
	File* file;
	try {
		file = SystemFile::open(fname);
	} catch(...) {
		return false;
	}

	// Lock library
	FT_Library lib = Library::lock();
	ASSERT(lib != NULL);

	// Open new face
	FT_Open_Args oargs;
	FT_Error error;
	FT_StreamRec stream;
	FT_Face face;
	oargs.flags = FT_OPEN_STREAM;
	oargs.stream = &stream;
	stream.base = NULL;
	stream.size = file->getSize();
    stream.pos = 0;
    stream.descriptor.pointer = file;
    stream.read = &streamRead;
    stream.close = &streamClose;
	error = FT_Open_Face(lib, &oargs, 0, &face);
	if (error != 0) {
		file->release();
		Library::unlock();
		return false;
	}
		
	// Set font info
	FontInfo info;
	ASSERT(face->family_name != NULL)
	info.m_File = fname;
	info.m_Family = String(face->family_name).toLowerCase();
	if (face->style_name != NULL)
		info.m_Style = String(face->style_name).toLowerCase();
	else
		info.m_Style = "";

	// Close file
	FT_Done_Face(face);
	file->release();

	// Insert Font Info
	m_FontInfo.insert(info);
	Library::unlock();
		
	return true;		
}		


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
void FontFactory::copyGlyphGray1(const FT_Bitmap* bmp, byte* dst, unsigned int pitch) {
	ASSERT(bmp->pixel_mode == FT_PIXEL_MODE_MONO);
	int i,j;
	byte* src = (byte*)bmp->buffer;
	for (i = 0; i < bmp->rows; i++) {
		for (j = 0; j < bmp->width; j++) {
			if ((src[j >> 3] & (1 << (7 - (j & 0x7)))) != 0)
				dst[j] = 255;
			else
				dst[j] = 0;
		}
		dst += pitch;
		src += bmp->pitch;
	}
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
void FontFactory::copyGlyphGray2(const FT_Bitmap* bmp, byte* dst, unsigned int pitch) {
	ASSERT(bmp->pixel_mode == FT_PIXEL_MODE_GRAY2);
	int i,j;
	byte* src = (byte*)bmp->buffer;
	byte tmp;
	int shift;
	for (i = 0; i < bmp->rows; i++) {
		for (j = 0; j < bmp->width; j++) {
			shift = 2*(j & 0x3);
			tmp = (src[j >> 2] & (3 << (6 - shift))) >> shift;
			dst[j] = (byte)(255.0f*(float)tmp/3.0f);
		}
		dst += pitch;
		src += bmp->pitch;
	}
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
void FontFactory::copyGlyphGray4(const FT_Bitmap* bmp, byte* dst, unsigned int pitch) {
	ASSERT(bmp->pixel_mode == FT_PIXEL_MODE_GRAY4);
	int i,j;
	byte* src = (byte*)bmp->buffer;
	byte tmp;
	for (i = 0; i < bmp->rows; i++) {
		for (j = 0; j < bmp->width; j++) {
			if (j & 1)
				tmp = src[j >> 1] >> 4;
			else 
				tmp = src[j >> 1] & 0x0f;
			dst[j] = (byte)(255.0f*(float)tmp/15.0f);
		}
		dst += pitch;
		src += bmp->pitch;
	}
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
void FontFactory::copyGlyphGrayN(const FT_Bitmap* bmp, byte* dst, unsigned int pitch) {
	ASSERT(bmp->pixel_mode == FT_PIXEL_MODE_GRAY);
	int i,j;
	byte* src = (byte*)bmp->buffer;
	byte tmp;
	for (i = 0; i < bmp->rows; i++) {
		for (j = 0; j < bmp->width; j++) {
			tmp = src[j];
			dst[j] = (byte)(255.0f*(float)tmp / (float)bmp->num_grays);
		}
		dst += pitch;
		src += bmp->pitch;
	}
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
void FontFactory::copyGlyph(const FT_Bitmap* bmp, byte* dst, unsigned int pitch) {
	ASSERT(bmp != NULL);
	ASSERT(dst != NULL);
	ASSERT(pitch > 0);
	switch(bmp->pixel_mode) {
		case FT_PIXEL_MODE_MONO:
			copyGlyphGray1(bmp, dst, pitch);
			return;
		case FT_PIXEL_MODE_GRAY:
			copyGlyphGrayN(bmp, dst, pitch);
			return;
		case FT_PIXEL_MODE_GRAY2:
			copyGlyphGray2(bmp, dst, pitch);
			return;
		case FT_PIXEL_MODE_GRAY4:
			copyGlyphGray4(bmp, dst, pitch);
			return;
		case FT_PIXEL_MODE_LCD:
		case FT_PIXEL_MODE_LCD_V:
		default:
			ASSERT(false);
			return;
	}
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
unsigned long FontFactory::streamRead(FT_Stream stream,unsigned long offset,unsigned char* buffer,unsigned long size) {
	((File*)stream->descriptor.pointer)->setPosition(offset, File::Absolute);
	return ((File*)stream->descriptor.pointer)->read(buffer, size);
}


/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
void FontFactory::streamClose(FT_Stream stream) {
	// ((File*)stream->descriptor.pointer)->release();
}

}; };
