/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "graphics.image.h"
#include "graphics.filter.byteorder.h"

#include "io.filter.inl"
#include "graphics.image.inl"
#include "graphics.dimension.inl"
#include "graphics.filter.inl"


namespace magnum {namespace graphics {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Image* FilterByteorder::process(const Image* src) {
	ASSERT(src != NULL);

	Image* img;
	int bits;

	switch(src->getType()) {
		case Image::RGB8:
			img = new Image(Image::BGR8, src->getWidth(), src->getHeight());
			bits = 24;
			break;
		case Image::BGR8:
			img = new Image(Image::RGB8, src->getWidth(), src->getHeight());
			bits = 24;
			break;
		case Image::RGBA8:
			img = new Image(Image::BGRA8, src->getWidth(), src->getHeight());
			bits = 32;
			break;
		case Image::BGRA8:
			img = new Image(Image::RGBA8, src->getWidth(), src->getHeight());
			bits = 32;
			break;
		default:
			return NULL;
	}

	if (img == NULL)
		return NULL;
	if (bits == 24) {
		if (!swap24(img, src)) {
			img->release();
			return NULL;
		}
	} else {
		if (!swap32(img, src)) {
			img->release();
			return NULL;
		}
	}

	return img;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool FilterByteorder::filter(Image* dstImg, const Image* srcImg) {
	ASSERT((dstImg != NULL) && (srcImg != NULL));

	if(dstImg == srcImg)
		return false;

	if ((dstImg->getWidth() != srcImg->getWidth()) 
	  || (dstImg->getHeight() != srcImg->getHeight()))
		return false;

	switch(srcImg->getType()) {
		case Image::RGB8:
			if (dstImg->getType() != Image::BGR8)
				return false;
			else
				return swap24(dstImg, srcImg);
		case Image::BGR8:
			if (dstImg->getType() != Image::RGB8)
				return false;
			else
				return swap24(dstImg, srcImg);
		case Image::RGBA8:
			if (dstImg->getType() != Image::BGRA8)
				return false;
			else
				return swap32(dstImg, srcImg);
		case Image::BGRA8:
			if (dstImg->getType() != Image::RGBA8)
				return false;
			else
				return swap32(dstImg, srcImg);
		default:
			return false;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool FilterByteorder::swap24(Image* dstImg, const Image* srcImg) {
	byte* src = (byte *)((Image *)srcImg)->lock();
	if (src == NULL) 
		return false;
	byte* dst = (byte *)dstImg->lock();
	if (dst == NULL) {
		((Image *)srcImg)->unlock();
		return false;
	}

	int srcLs = srcImg->getLinesize();
	int dstLs = dstImg->getLinesize();

	int height = srcImg->getHeight();
	int width = srcImg->getWidth();

	for (int i = 0; i < height; i++) {
		for (int j = 0; j < width; j++) {
			dst[3*j] = src[3*j+2];
			dst[3*j+1] = src[3*j+1];
			dst[3*j+2] = src[3*j];
		}
		src += srcLs;
		dst += dstLs;
	}

	dstImg->unlock();
	((Image *)srcImg)->unlock();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool FilterByteorder::swap32(Image* dstImg, const Image* srcImg) {
	byte* src = (byte *)((Image *)srcImg)->lock();
	if (src == NULL) {
		return false;
	}
	byte* dst = (byte *)dstImg->lock();
	if (dst == NULL) {
		((Image *)srcImg)->unlock();
		return false;
	}

	int srcLs = srcImg->getLinesize();
	int dstLs = dstImg->getLinesize();

	int height = srcImg->getHeight();
	int width = srcImg->getWidth();

	for (int i = 0; i < height; i++) {
		for (int j = 0; j < width; j++) {
			dst[4*j] = src[4*j+2];
			dst[4*j+1] = src[4*j+1];
			dst[4*j+2] = src[4*j];
			dst[4*j+3] = src[4*j+3];
		}
		src += srcLs;
		dst += dstLs;
	}

	dstImg->unlock();
	((Image *)srcImg)->unlock();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool FilterByteorder::swap24(Image* img) {
	ASSERT(img != NULL);

	byte* src = (byte *)img->lock();
	if (src == NULL)
		return false;

	int ls = img->getLinesize();
	int height = img->getHeight();
	int width = img->getWidth();

	byte tmp;
	for (int i = 0; i < height; i++) {
		for (int j = 0; j < width; j++) {
			tmp = src[3*j];
			src[3*j] = src[3*j+2];
			src[3*j+2] = tmp;
		}
		src += ls;
	}

	img->unlock();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool FilterByteorder::swap32(Image* img) {
	ASSERT(img != NULL);

	byte* src = (byte *)img->lock();
	if (src == NULL)
		return false;

	int ls = img->getLinesize();
	int height = img->getHeight();
	int width = img->getWidth();

	byte tmp;
	for (int i = 0; i < height; i++) {
		for (int j = 0; j < width; j++) {
			tmp = src[4*j];
			src[4*j] = src[4*j+2];
			src[4*j+2] = tmp;
		}
		src += ls;
	}

	img->unlock();
	return true;
}



}; };
