/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "graphics3d.compiler.h"
#include "graphics3d.renderstream.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.material.h"
#include "graphics3d.shader.h"
#include "graphics3d.appearance.h"
#include "graphics3d.context.h"
#include "graphics3d.pixelprogram.h"
#include "graphics3d.vertexprogram.h"

#include "object.inl"
#include "base.string.inl"
#include "object.pointer.inl"
#include "container.array.inl"
#include "container.set.inl"
#include "container.iterator.inl"
#include "container.allocator.inl"
#include "system.assembler.inl"
#include "system.program.inl"
#include "graphics3d.renderstream.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.material.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.context.inl"
#include "graphics3d.mode.inl"
#include "graphics3d.texture.inl"
#include "graphics3d.appearance.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.pixelprogram.inl"
#include "graphics3d.vertexprogram.inl"


namespace magnum {namespace graphics3d {
using magnum::container::Set;
using magnum::object::Pointer;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Compiler::Compiler() {
	clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Compiler::~Compiler() {
	clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Compiler::clear() {
	unsigned int i;

    m_CurrentMaterial = NULL;
    m_CurrentShader = NULL;
    m_CurrentMode = NULL;
    m_CurrentIndexBuffer = NULL;
	m_CurrentPixelProgram = NULL;
	m_CurrentVertexProgram = NULL;


	for (i = 0; i < Attributes::MaxAttributes; i++)
	    m_CurrentVertexBuffer[i] = NULL;

	for (i = 0; i < Texture::MaxTextures; i++)
		m_CurrentTexture[i] = NULL;	

	m_Resources.clear();
	m_Assembler.clear();

	// Get Context...
	m_Assembler.asmPush(Assembler::R0);
	m_Assembler.asmCall((void (*)())RenderStream::getContext);
	m_Assembler.asmMove(Assembler::R0, Assembler::Result);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Compiler::finish() {
	m_Assembler.asmPop(Assembler::R0);
	m_Assembler.asmReturn();
	m_Assembler.finish();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
RenderStream* Compiler::compile() {
	RenderStream* stream = new RenderStream();
	
	// Create Program
	stream->m_Program = m_Assembler.compile();

	// Copy all Resources
	magnum::container::Set<ResPtr>::Iterator iter, end;
	end = m_Resources.end();
	iter = m_Resources.begin();
	while(iter != end) {
		stream->m_Resources.insert(*iter);
		iter++;
	}
	stream->m_Resources.shrink();

	return stream;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Compiler::selectVertexBuffer(const VertexBuffer* vb, unsigned int attribs) {
	ASSERT(vb != NULL);
	attribs &= vb->getAttributes();
	unsigned int change = 0;

	for (unsigned int i = 0; i < Attributes::MaxAttributes; i++) {
		if ((attribs & (1 << i)) == 0)
			continue;
		if (m_CurrentVertexBuffer[i] != vb) {
			change |= (1 << i);
			m_CurrentVertexBuffer[i] = vb;
		}
	}

	if (change == 0)
		return;

	m_Resources.insert(ResPtr((Resource*)vb));
	vb->compileSelect(m_Assembler, change);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Compiler::selectIndexBuffer(const IndexBuffer* ib) {
	ASSERT(ib != NULL);

	if (m_CurrentIndexBuffer == ib)
		return;

	m_CurrentIndexBuffer = ib;

	ib->compileSelect(m_Assembler);
	m_Resources.insert(ResPtr((Resource*)ib));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Compiler::selectMaterial(const Material* mat) {
	ASSERT(mat != NULL);

	if (m_CurrentMaterial == mat)
		return;

	m_CurrentMaterial = mat;

	m_Resources.insert(ResPtr((Resource*)mat));
	mat->compileSelect(m_Assembler);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Compiler::deselectMaterial() {
	if (m_CurrentMaterial == NULL)
		return;

	m_CurrentMaterial = NULL;

	m_Assembler.asmCall(Material::selectNone);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Compiler::selectTexture(const Texture* tex,unsigned int stage) {
	ASSERT(tex != NULL);
	ASSERT(stage < Texture::MaxTextures);

	if (m_CurrentTexture[stage] == tex)
		return;

	m_CurrentTexture[stage] = tex;

	m_Resources.insert(ResPtr((Resource*)tex));
	tex->compileSelect(m_Assembler, stage);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Compiler::deselectTexture(unsigned int stage) {
	ASSERT(stage < Texture::MaxTextures);

	if (m_CurrentTexture[stage] == NULL)
		return;

	m_CurrentTexture[stage] = NULL;

	m_Assembler.asmCall(Texture::selectNone, stage);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Compiler::selectShader(const Shader* shader) {
	ASSERT(shader != NULL);

	if (m_CurrentShader == shader)
		return;

	m_CurrentShader = shader;

	m_Resources.insert(ResPtr((Resource*)shader));
	shader->compileSelect(m_Assembler);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Compiler::deselectShader() {
	if (m_CurrentShader == NULL)
		return;

	m_CurrentShader = NULL;
	m_Assembler.asmCall(Shader::selectNone);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Compiler::selectMode(const Mode* mode) {
	ASSERT(mode != NULL);

	if (m_CurrentMode == mode)
		return;

	m_CurrentMode = mode;

	m_Resources.insert(ResPtr((Resource*)mode));
	m_Assembler.asmCall(Mode::forceSelect, mode);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Compiler::deselectMode() {
	if (m_CurrentMode == NULL)
		return;

	m_CurrentMode = NULL;
	
	m_Assembler.asmCall(Mode::selectNone);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Compiler::selectAppearance(const Appearance* app) {
	ASSERT(app != NULL);
	if (app->getShader() != NULL)
		selectShader(app->getShader());
	if (app->getMaterial() != NULL)
		selectMaterial(app->getMaterial());
	for (unsigned int i = 0; i < Texture::MaxTextures; i++) {
		if (app->getTexture(i) != NULL)
			selectTexture(app->getTexture(i), i);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Compiler::deselectAppearance() {
	deselectShader();
	deselectMaterial();
	for (unsigned int i = 0; i < Texture::MaxTextures; i++) {
		deselectTexture(i);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Compiler::renderVertices(Mesh::Primitive p, unsigned int start, unsigned int size) {
	m_Assembler.asmCall((void (*)(Mesh::Primitive, unsigned int, unsigned int))
		RenderStream::onRenderVertices, p, start, size);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Compiler::renderIndices(Mesh::Primitive p, unsigned int start, unsigned int size) {
	m_Assembler.asmCall((void (*)(Mesh::Primitive, unsigned int, unsigned int))
		RenderStream::onRenderIndices, p, start, size);
}

}; };
