/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "graphics3d.context.h"
#include "graphics3d.framebuffer.h"
#include "graphics3d.texture1d.h"
#include "graphics3d.texture2d.h"
#include "graphics3d.vertexformat.h"

#include "object.inl"
#include "base.string.inl"
#include "graphics.point.inl"
#include "graphics.rectangle.inl"
#include "graphics.dimension.inl"
#include "graphics.image.inl"
#include "graphics3d.framebuffer.inl"
#include "graphics3d.texture.inl"
#include "graphics3d.texture1d.inl"
#include "graphics3d.texture2d.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.context.inl"


namespace magnum { namespace graphics3d {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
FrameBuffer::FrameBuffer() {
	m_RenderTarget = NULL;
#if defined(__FORWINDOWS__)
	m_DeviceContext = wglGetCurrentDC();
	m_WindowHandle = WindowFromDC(m_DeviceContext);
	RECT rect;
	GetWindowRect(m_WindowHandle, &rect);
	m_Size.setWidth(rect.right - rect.left);
	m_Size.setHeight(rect.bottom - rect.top);
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameBuffer::setSize(const Dimension<int>& size) {
#if defined (__FORWINDOWS__)
	SetWindowPos(m_WindowHandle, NULL, 0, 0, size.getWidth(), size.getHeight(), SWP_NOZORDER | SWP_NOCOPYBITS | SWP_NOMOVE | SWP_NOSENDCHANGING);
	RECT rect;
	GetWindowRect(m_WindowHandle, &rect);
	m_Size.setWidth(rect.right - rect.left);
	m_Size.setHeight(rect.bottom - rect.top);
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameBuffer::copyTo(Texture2D* tex,unsigned int level, const Point<int>& p,const Rectangle<int>& r) const {
	ASSERT(tex != NULL);
	ASSERT(level < tex->getLevels());
	int oldtex;
	glActiveTextureARB(GL_TEXTURE0_ARB);
	glGetIntegerv(GL_TEXTURE_2D, &oldtex);
	glBindTexture(GL_TEXTURE_2D, tex->m_OglTexture);
	glCopyTexSubImage2D(GL_TEXTURE_2D, level, 
		p.getX(), tex->getHeight() - (p.getY() + r.getHeight()), 
		r.getX(), m_Size.getHeight() - (r.getY() + r.getHeight()), 
		r.getWidth(), r.getHeight());
	glBindTexture(GL_TEXTURE_2D, oldtex);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameBuffer::copyTo(Image* img, const Point<int>& p,const Rectangle<int>& r) const {
	ASSERT(img != NULL);
	ASSERT((img->getWidth() - p.getX()) >= r.getWidth());
	ASSERT((img->getHeight() - p.getX()) >= r.getHeight());

	int type;
	int format;
	int psize;

	switch (img->getType()) {
		case Image::RGB8:
			type = GL_UNSIGNED_BYTE;
			format = GL_RGB;
			psize = 3;
			break;
		case Image::RGBA8:
			type = GL_UNSIGNED_BYTE;
			format = GL_RGBA;
			psize = 4;
			break;
		case Image::BGR8:
			type = GL_UNSIGNED_BYTE;
			format = GL_BGR_EXT;
			psize = 3;
			break;
		case Image::BGRA8:
			type = GL_UNSIGNED_BYTE;
			format = GL_BGRA_EXT;
			psize = 4;
			break;
		case Image::Index8:
			type = GL_UNSIGNED_BYTE;
			format = 1;
			psize = 1;
			break;
		default:
			return;
	}

	void* data = img->lock();
	glPixelStorei(GL_PACK_SWAP_BYTES , false);
	glPixelStorei(GL_PACK_ROW_LENGTH, img->getWidth());
	glPixelStorei(GL_PACK_SKIP_PIXELS, p.getX());
	glPixelStorei(GL_PACK_SKIP_ROWS, img->getHeight() - (p.getY() + r.getHeight()));
	glPixelStorei(GL_PACK_ALIGNMENT, 1); 
	glReadPixels(r.getX(), m_Size.getHeight() - (r.getY() + r.getHeight()), 
		r.getWidth(), r.getHeight(), 
		format, type, data);

	// Now we still have to flip the image
	int i;
	int len = psize * r.getWidth();
	int lsize = img->getLinesize();
	byte* upos = (byte*)data + p.getY() * lsize + p.getX()*psize;
	byte* lpos = upos + (r.getHeight() - 1) * lsize;
	byte flipbuffer[4096*4];
	for (i = 0; i < (r.getHeight() >> 1); i++) {
		memcpy(flipbuffer, lpos, len);
		memcpy(lpos, upos, len);
		memcpy(upos, flipbuffer, len);
		lpos -= lsize;
		upos += lsize;
	}

	img->unlock();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameBuffer::copyFrom(const Image* img, const Point<int>& p,const Rectangle<int>& r) {
	ASSERT(img != NULL);

	void* data = ((Image *)img)->lock();
	glPixelZoom(1.0f, -1.0f);
	glPixelStorei(GL_UNPACK_SWAP_BYTES , false);
	glPixelStorei(GL_UNPACK_ROW_LENGTH, img->getWidth());
	glPixelStorei(GL_UNPACK_SKIP_PIXELS, p.getX());
	glPixelStorei(GL_UNPACK_SKIP_ROWS, img->getHeight() - (p.getY() + r.getHeight()));
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1); 

	int format, type;
	switch (img->getType()) {
		case Image::RGB8:
			format = GL_RGB;
			type = GL_UNSIGNED_BYTE;
			break;
		case Image::Index8:
			format = GL_COLOR_INDEX;
			type = GL_UNSIGNED_BYTE;
			break;
		default:
			format = GL_COLOR_INDEX;
			type = GL_UNSIGNED_BYTE;
	}

	int mmode;
	glGetIntegerv(GL_MATRIX_MODE, &mmode);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glRasterPos2i(r.getX(), m_Size.getHeight() - (r.getY() + r.getHeight()));
	glDrawPixels(r.getWidth(), r.getHeight(), format, type, data);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glMatrixMode(mmode);
	((Image *)img)->unlock();
}

}; };
