/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "graphics3d.framemodel.h"
#include "graphics3d.mesh.h"
#include "graphics3d.meshgroup.h"
#include "graphics3d.material.h"
#include "graphics3d.shader.h"
#include "graphics3d.appearance.h"
#include "graphics3d.renderstream.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.context.h"
#include "graphics3d.bounds.h"

#include "object.pointer.inl"
#include "math.matrix.inl"
#include "math.vector.inl"
#include "container.array.inl"
#include "container.iterator.inl"
#include "geometry.bounds.inl"
#include "geometry.sphere.inl"
#include "geometry.box.inl"
#include "geometry.alignedbox.inl"
#include "graphics3d.framemodel.inl"
#include "graphics3d.meshgroup.inl"
#include "graphics3d.mesh.inl"
#include "graphics3d.renderstream.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.appearance.inl"
#include "graphics3d.matrixstack.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.parameter.inl"
#include "graphics3d.bounds.inl"


namespace magnum { namespace graphics3d {
using magnum::geometry::AlignedBox3;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameModel::finish() {
	ParameterFormat fmt;
	fmt.insert(Parameters::Position, Types::Matrix4f, 0);
	clearParameters();
	publishParameters(fmt);
	
	clearJoints();
	
	AlignedBox3<float> box, tmpbox;
	box.setMaximum(Vector3<float>::Null);
	box.setMinimum(Vector3<float>::Null);
	Array<Group>::Iterator giter;
	for (giter = m_Groups.begin(); giter != m_Groups.end(); giter++) {
		VertexBuffer* vb;
		for (unsigned int i = 0; i < giter->m_Frames.getSize(); i++) {
			vb = giter->m_Frames[i];
			if (fitAlignedBox(vb, &tmpbox)) {
				box.merge(tmpbox);
			}
		}
	}
	publishBounds(Bounds3<float>(box));
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameModel::getGroups(Array<MeshGroup*>* groups) const {
	magnum::container::Array<Group>::Iterator giter, gend;
	gend = m_Groups.end();

	for (giter = m_Groups.begin(); giter != gend; giter++) {
		groups->insert(giter->m_MeshGroup);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameModel::getMeshes(Array<Mesh*>* meshes) const {
	magnum::container::Array<Group>::Iterator giter, gend;
	gend = m_Groups.end();
	MeshGroup::Iterator miter, mend;

	for (giter = m_Groups.begin(); giter != gend; giter++) {
		mend = giter->m_MeshGroup->end();
		for (miter = giter->m_MeshGroup->begin(); miter != mend; miter++) {
			meshes->insert(*miter);
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameModel::getFrames(const MeshGroup* group,Array<VertexBuffer*>* frames) const {
	magnum::container::Array<Group>::Iterator giter, gend;
	gend = m_Groups.end();
	magnum::container::Array<VertexBufferP>::Iterator fiter, fend;

	for (giter = m_Groups.begin(); giter != gend; giter++) {
		if (giter->m_MeshGroup != group)
			continue;

		fend = giter->m_Frames.end();
		for (fiter = giter->m_Frames.begin(); fiter != fend; fiter++) {
			frames->insert(*fiter);
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned int FrameModel::insertGroup(MeshGroup* g) {
	Group group;
	group.m_MeshGroup = g;
	return m_Groups.insert(group);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool FrameModel::removeGroup(MeshGroup* mg) {
	magnum::container::Array<Group>::Iterator giter, gend;
	gend = m_Groups.end();

	for (giter = m_Groups.begin(); giter != gend; giter++) {
		if (giter->m_MeshGroup == mg) {
			m_Groups.remove(giter);
			return true;
		}
	}

	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameModel::insertFrame(MeshGroup* mg,VertexBuffer* vb,unsigned int frame) {
	magnum::container::Array<Group>::Iterator giter, gend;
	gend = m_Groups.end();

	for (giter = m_Groups.begin(); giter != gend; giter++) {
		if (giter->m_MeshGroup == mg) {
			giter->m_Frames.set(frame, vb);
			return;
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameModel::removeFrame(MeshGroup* mg,VertexBuffer* vb) {
	magnum::container::Array<Group>::Iterator giter, gend;
	gend = m_Groups.end();

	for (giter = m_Groups.begin(); giter != gend; giter++) {
		if (giter->m_MeshGroup == mg)
			break;
	}

	if (giter == gend)
		return;

	for (unsigned int i = 0; i < giter->m_Frames.getSize(); i++) {
		if (giter->m_Frames[i] == vb)
			giter->m_Frames[i] = NULL;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameModel::removeFrame(MeshGroup* mg, unsigned int i) {
	magnum::container::Array<Group>::Iterator giter, gend;
	gend = m_Groups.end();

	for (giter = m_Groups.begin(); giter != gend; giter++) {
		if (giter->m_MeshGroup == mg) {
			giter->m_Frames[i] = NULL;
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void FrameModel::onRender(const ParameterBuffer* pbuf) const {
	ASSERT(pbuf != NULL);
	MatrixStack* ms = getContext()->getMatrixStack();
	ms->selectMode(MatrixStack::ModelView);
	// ms->set(*(const Matrix4<float>*)pbuf->getBuffer());
	
	magnum::container::Array<Group>::Iterator giter, gend;
	gend = m_Groups.end();
	MeshGroup::Iterator miter, mend;

	for (giter = m_Groups.begin(); giter != gend; giter++) {
		giter->m_Frames[0]->select();
		mend = giter->m_MeshGroup->end();
		for (miter = giter->m_MeshGroup->begin(); miter != mend; miter++) {
			(*miter)->render();
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool FrameModel::onOptimize() {
	return false;
}

}; };
