/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "graphics3d.indexmesh.h"
#include "graphics3d.context.h"
#include "graphics3d.material.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.shader.h"
#include "graphics3d.pixelprogram.h"
#include "graphics3d.vertexprogram.h"
#include "graphics3d.appearance.h"

#include "object.inl"
#include "object.pointer.inl"
#include "base.string.inl"
#include "system.thread.inl"
#include "math.vector.inl"
#include "math.vector3.inl"
#include "math.matrix.inl"
#include "math.range.inl"
#include "geometry.trianglemesh.inl"
#include "geometry.bounds.inl"
#include "geometry.sphere.inl"
#include "geometry.box.inl"
#include "geometry.alignedbox.inl"
#include "container.array.inl"
#include "graphics3d.mesh.inl"
#include "graphics3d.vertexmesh.inl"
#include "graphics3d.indexmesh.inl"
#include "graphics3d.context.inl"
#include "graphics3d.material.inl"
#include "graphics3d.texture.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.vertexformat.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.pixelprogram.inl"
#include "graphics3d.vertexprogram.inl"
#include "graphics3d.appearance.inl"


namespace magnum { namespace graphics3d {
using magnum::math::Vector;
using magnum::geometry::Bounds;
using magnum::geometry::AlignedBox3;


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> void triangulateTriangleStrip(TriangleMesh3<float>* mesh, const T* indices, unsigned int numIndices) {
	T a,b,c;
	a = *indices++;
	b = *indices++;
	unsigned int tri = 0;
	for (unsigned int i = 0; i < numIndices - 2; i++) {
		c = *indices++;
		if ((a != b) && (a != c) && (b != c))
			mesh->setTriangle(tri++, a,b,c);
		if (i & 1) {
			b = c;
		} else {
			a = c;
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> void triangulateTriangleList(TriangleMesh3<float>* mesh, const T* indices, unsigned int numIndices) {
	T a,b,c;
	for (unsigned int i = 0; i < numIndices/3; i++) {
		a = *indices++;
		b = *indices++;
		c = *indices++;
		mesh->setTriangle(i, a,b,c);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
template<class T> void triangulateTriangleFan(TriangleMesh3<float>* mesh, const T* indices, unsigned int numIndices) {
	T a,b,c;
	a = *indices++;
	b = *indices++;
	for (unsigned int i = 0; i < numIndices - 2; i++) {
		c = *indices++;
		mesh->setTriangle(i, a,b,c);
		b = c;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void IndexMesh::onRender() const {
	ASSERT(m_IndexBuffer != NULL);

	selectVertexBuffers();
	selectIndexBuffer();

	if (getAppearance() != NULL)
		getAppearance()->select();

	if (getRange().getSize()) 
		getContext()->renderIndices(getPrimitive(), getRange().getStart(), getRange().getSize());
	else
		getContext()->renderIndices(getPrimitive(), m_IndexBuffer->getSize());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool IndexMesh::onBound(Bounds3<float>* bounds) const {
	ASSERT(bound != NULL);
	unsigned int i;

	// Search appropriate vertex buffer
	VertexBuffer* vb = NULL;
	for (i = 0; i < VertexBuffer::MaxBuffers; i++) {
		if (getVertexBuffer(i)->getAttributes() & Attributes::Vertex) {
			vb = getVertexBuffer(i);
			break;
		}
	}
	if (i == VertexBuffer::MaxBuffers)
		return false;

	// Find correct attribute and offset
	unsigned int offset;
	unsigned int type;
	unsigned int count;
	for (i = 0; i < VertexFormat::MaxComponents; i++) {
		if (vb->getAttribute(i) & Attributes::Vertex) {
			offset = vb->getOffset(i);
			type = vb->getType(i);
			count = vb->getCount(i);
			break;
		}
	}
	if (i == VertexFormat::MaxComponents)
		return false;
	
	// Get offset, stride etc	
	unsigned int numVertices = vb->getSize();
	unsigned int incVertices = vb->getStride();
	void* vertices = vb->lock(VertexBuffer::ReadOnly);
	if (vertices == NULL)
		return NULL;

	// Adjust buffer for offset and range	
	vertices = (void*)((byte *)vertices + offset);
//	if (m_Range.getSize() > 0) {
//		vertices = (void*)((byte *)vertices + m_Range.getStart()*incVertices);
//		numVertices = m_Range.getSize();	
//	}
	
	switch(type) {
		case Types::Byte3:
		case Types::Byte4: {
			AlignedBox3<byte> box;
			box.bound((Vector<byte,3>*)vertices, incVertices, numVertices);
			bounds->setBounds(AlignedBox3<float>(box.getMinimum(), box.getMaximum()));
			break;		
		}
		case Types::Integer3:
		case Types::Integer4: {
			AlignedBox3<int> box;
			box.bound((Vector<int,3>*)vertices, incVertices, numVertices);
			bounds->setBounds(AlignedBox3<float>(box.getMinimum(), box.getMaximum()));
			break;		
		}
		case Types::Float3:
		case Types::Float4: {
			AlignedBox3<float> box;
			box.bound((Vector<float,3>*)vertices, incVertices, numVertices);
			bounds->setBounds(AlignedBox3<float>(box.getMinimum(), box.getMaximum()));
			break;		
		}
		case Types::Double3:
		case Types::Double4: {
			AlignedBox3<double> box;
			box.bound((Vector<double,3>*)vertices, incVertices, numVertices);
			bounds->setBounds(AlignedBox3<float>(box.getMinimum(), box.getMaximum()));
			break;		
		}
		
		default:
			vb->unlock();
			return false;
	}
	
	vb->unlock();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
TriangleMesh3<float>* IndexMesh::onTriangulate() const {
	unsigned int i;
	float x,y,z;

	// Check for existence of index buffer
	if (getIndexBuffer() == NULL)
		return NULL;

	// Search appropriate vertex buffer
	VertexBuffer* vb = NULL;
	for (i = 0; i < VertexBuffer::MaxBuffers; i++) {
		if (getVertexBuffer(i)->getAttributes() & Attributes::Vertex) {
			vb = getVertexBuffer(i);
			break;
		}
	}
	if (i == VertexBuffer::MaxBuffers)
		return NULL;

	// Find correct attribute and offset
	unsigned int offset;
	unsigned int type;
	unsigned int count;
	for (i = 0; i < VertexFormat::MaxComponents; i++) {
		if (vb->getAttribute(i) & Attributes::Vertex) {
			offset = vb->getOffset(i);
			type = (vb->getType(i) >> Types::TypeShift) & Types::TypeMask;
			count = vb->getCount(i);
			break;
		}
	}
	if (i == VertexFormat::MaxComponents)
		return NULL;

	// Copy vertices
	unsigned int numVertices = vb->getSize();
	unsigned int incVertices = vb->getStride();
	void* vertices = vb->lock(VertexBuffer::ReadOnly);
	if (vertices == NULL)
		return NULL;
		
	vertices = (void*)((byte *)vertices + offset);
	
	// Create Triangle mesh
	TriangleMesh3<float>* mesh = new TriangleMesh3<float>();
	

#define copyVertex(T) \
		x = (float)((T*)vertices)[0]; \
		y = (count >= 2) ? (float)((T*)vertices)[1] : 0.0f; \
		z = (count >= 3) ? (float)((T*)vertices)[2] : 0.0f; \
		vertices = (void*)((byte *)vertices + incVertices); \
		mesh->setVertex(i, Vector3<float>(x,y,z));
#define copyVertices(T) \
		mesh->setVertices(numVertices); \
		for (i = 0; i < numVertices; i++) { \
			copyVertex(T); \
		}
	switch(type) {
		case ((Types::Byte >> Types::TypeShift) & Types::TypeMask):
			copyVertices(int8);
			break;
		case ((Types::UByte >> Types::TypeShift) & Types::TypeMask):
			copyVertices(uint8);
			break;
		case ((Types::Short >> Types::TypeShift) & Types::TypeMask):
			copyVertices(int16);
			break;
		case ((Types::UShort >> Types::TypeShift) & Types::TypeMask):
			copyVertices(uint16);
			break;
		case ((Types::Int >> Types::TypeShift) & Types::TypeMask):
			copyVertices(int32);
			break;
		case ((Types::UInt >> Types::TypeShift) & Types::TypeMask):
			copyVertices(uint32);
			break;
		case ((Types::Float >> Types::TypeShift) & Types::TypeMask):
			copyVertices(float);
			break;
		case ((Types::Double >> Types::TypeShift) & Types::TypeMask):
			copyVertices(double);
			break;
		default:
			break;
	}
	vb->unlock();

	// Copy indices
	IndexBuffer* ib = getIndexBuffer();
	unsigned int numIndices = ib->getSize();
	void* indices = ib->lock(IndexBuffer::ReadOnly);
	if (indices == NULL)
		return NULL;
		
	// Adjust for subrange
	if (getRange().getSize() > 0) {
		numIndices = getRange().getSize();
		if (ib->getType() == IndexBuffer::Index8)
			indices = (void *)((byte *)indices + getRange().getStart());
		else if (ib->getType() == IndexBuffer::Index16)
			indices = (void *)((word *)indices + getRange().getStart());
		else if (ib->getType() == IndexBuffer::Index32)
			indices = (void *)((dword *)indices + getRange().getStart());
	}

	switch(getPrimitive()) {
		case TriangleList:
			if (ib->getType() == IndexBuffer::Index8)
				triangulateTriangleList(mesh, (byte*)indices, numIndices);
			else if (ib->getType() == IndexBuffer::Index16)
				triangulateTriangleList(mesh, (word*)indices, numIndices);
			else if (ib->getType() == IndexBuffer::Index32)
				triangulateTriangleList(mesh, (dword*)indices, numIndices);
			break;
		case TriangleStrip:
			if (ib->getType() == IndexBuffer::Index8)
				triangulateTriangleStrip(mesh, (byte*)indices, numIndices);
			else if (ib->getType() == IndexBuffer::Index16)
				triangulateTriangleStrip(mesh, (word*)indices, numIndices);
			else if (ib->getType() == IndexBuffer::Index32)
				triangulateTriangleStrip(mesh, (dword*)indices, numIndices);
			break;
		case TriangleFan:
			if (ib->getType() == IndexBuffer::Index8)
				triangulateTriangleFan(mesh, (byte*)indices, numIndices);
			else if (ib->getType() == IndexBuffer::Index16)
				triangulateTriangleFan(mesh, (word*)indices, numIndices);
			else if (ib->getType() == IndexBuffer::Index32)
				triangulateTriangleFan(mesh, (dword*)indices, numIndices);
			break;
		case QuadStrip:
		case QuadList:
		default:
			break;
	}
	ib->unlock();

	return mesh;
}


}; };
