/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "math.consts.h"
#include "graphics3d.meshmerger.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.extensions.h"
#include "graphics3d.context.h"
#include "graphics3d.indexmesh.h"
#include "graphics3d.vertexmesh.h"

#include "object.inl"
#include "base.string.inl"
#include "math.range.inl"
#include "container.iterator.inl"
#include "container.map.inl"
#include "container.tuple.inl"
#include "container.list.inl"
#include "container.allocator.inl"
#include "graphics3d.meshmerger.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.indexmesh.inl"
#include "graphics3d.vertexmesh.inl"
#include "graphics3d.vertexformat.inl"

#include <typeinfo.h>


namespace magnum { namespace graphics3d {
using namespace magnum::container;
using namespace magnum::math;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
MeshMerger::MeshMerger() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
MeshMerger::~MeshMerger() {
	clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void MeshMerger::clear() {
	// Release all buffers
	List<VbInfo>::Iterator viter;
	for (viter = m_VbInfo.begin(); viter != m_VbInfo.end(); viter++) {
		viter->m_Buffer->release();
		viter->m_Buffer = NULL;
	}
	List<IbInfo>::Iterator iiter;
	for (iiter = m_IbInfo.begin(); iiter != m_IbInfo.end(); iiter++) {
		iiter->m_Buffer->release();
		iiter->m_Buffer = NULL;
	}

	// Clean up
	m_VbMapping.clear();
	m_IbMapping.clear();
	m_VbInfo.clear();
	m_IbInfo.clear();
	m_Meshes.clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void MeshMerger::add(Mesh* mesh) {
	IndexBuffer* ib;
	unsigned int i, shift;
	VbTuple vbt;

	// Collect some data
	if (typeid(*mesh) == typeid(IndexMesh)) {
		ib = ((IndexMesh*)mesh)->getIndexBuffer();
		for (i = 0; i < VertexBuffer::MaxBuffers; i++)
			vbt[i] = ((VertexMesh*)mesh)->getVertexBuffer(i);
		shift = addVertexBuffers(vbt);
		addIndexBuffer(ib, shift);
		m_Meshes.insert(mesh);
	} else if (typeid(*mesh) == typeid(VertexMesh)) {
		for (i = 0; i < VertexBuffer::MaxBuffers; i++)
			vbt[i] = ((VertexMesh*)mesh)->getVertexBuffer(i);
		addVertexBuffers(vbt);
		m_Meshes.insert(mesh);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void MeshMerger::add(const Tuple<VertexBuffer*, VertexBuffer::MaxBuffers>& vbt) {
	addVertexBuffers(vbt);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void MeshMerger::add(const Tuple<VertexBuffer*, VertexBuffer::MaxBuffers>& vbt,IndexBuffer* ib) {
	unsigned int ishift;
	ishift = addVertexBuffers(vbt);
	addIndexBuffer(ib, ishift);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void MeshMerger::add(VertexBuffer* vb) {
	VbTuple vbt;
	for (unsigned int i = 1; i < VertexBuffer::MaxBuffers; i++)
		vbt[i] = NULL;
	vbt[0] = vb;

	addVertexBuffers(vbt);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void MeshMerger::add(VertexBuffer* vb,IndexBuffer* ib) {
	VbTuple vbt;
	for (unsigned int i = 1; i < VertexBuffer::MaxBuffers; i++)
		vbt[i] = NULL;
	vbt[0] = vb;

	unsigned int ishift;
	ishift = addVertexBuffers(vbt);
	addIndexBuffer(ib, ishift);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void MeshMerger::finish() {
	// merge buffers
	mergeIndexBuffers();
	mergeVertexBuffers();

	// Walk through all jobs and replace buffers
	List<Mesh*>::Iterator iter;
	for (iter = m_Meshes.begin(); iter != m_Meshes.end(); iter++) {
		processMesh(*iter);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
MeshMerger::IndexMapping MeshMerger::getMapping(const IndexBuffer* ib) {
	IndexMapping mapping;
	Map<IndexBuffer*, IbMapping>::Iterator iter;

	iter = m_IbMapping.at((IndexBuffer*)ib);
	if (iter == m_IbMapping.end()) {
		mapping.m_Buffer = NULL;
		mapping.m_Start = 0;
		mapping.m_Size = 0;
		mapping.m_Shift = 0;
		return mapping;
	}

	mapping.m_Buffer = iter->m_Target->m_Buffer;
	mapping.m_Start = iter->m_Start;
	mapping.m_Size = iter->m_Count;
	mapping.m_Shift = iter->m_Shift;
	return mapping;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
MeshMerger::VertexMapping MeshMerger::getMapping(const Tuple<VertexBuffer*, VertexBuffer::MaxBuffers>& vbt) {
	VertexMapping mapping;
	Map<VbTuple, VbMapping>::Iterator iter;

	iter = m_VbMapping.at(vbt);
	if (iter == m_VbMapping.end()) {
		mapping.m_Buffer = NULL;
		mapping.m_Start = 0;
		mapping.m_Size = 0;
		return mapping;
	}

	mapping.m_Buffer = iter->m_Target->m_Buffer;
	mapping.m_Start = iter->m_Start;
	mapping.m_Size = iter->m_Count;
	return mapping;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
MeshMerger::VertexMapping MeshMerger::getMapping(const VertexBuffer* vb) {
	VbTuple vbt;
	for (unsigned int i = 1; i < VertexBuffer::MaxBuffers; i++)
		vbt[i] = NULL;
	vbt[0] = (VertexBuffer*)vb;

	VertexMapping mapping;
	Map<VbTuple, VbMapping>::Iterator iter;

	iter = m_VbMapping.at(vbt);
	if (iter == m_VbMapping.end()) {
		mapping.m_Buffer = NULL;
		mapping.m_Start = 0;
		mapping.m_Size = 0;
		return mapping;
	}

	mapping.m_Buffer = iter->m_Target->m_Buffer;
	mapping.m_Start = iter->m_Start;
	mapping.m_Size = iter->m_Count;
	return mapping;
}

		
/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void MeshMerger::processMesh(Mesh* mesh) {

	// Get type of mesh:
	if (typeid(*mesh) == typeid(IndexMesh)) {
		IndexMesh* imesh = (IndexMesh*)mesh;

		IndexBuffer* ib;
		ib = imesh->getIndexBuffer();
		imesh->detachIndexBuffer();

		VbTuple vbt;
		for (unsigned int i = 0; i < VertexBuffer::MaxBuffers; i++) {
			vbt[i] = imesh->getVertexBuffer(i);
			imesh->detachVertexBuffer(i);
		}

		Map<VbTuple, VbMapping>::Iterator vbm = m_VbMapping.at(vbt);	
		Map<IndexBuffer*, IbMapping>::Iterator ibm = m_IbMapping.at(ib);
		ASSERT(vbm != m_VbMapping.end());
		ASSERT(ibm != m_IbMapping.end());

		imesh->attachIndexBuffer(ibm->m_Target->m_Buffer);
		imesh->attachVertexBuffer(vbm->m_Target->m_Buffer);
		if (imesh->getRange().getSize() == 0)
			imesh->setRange(ibm->m_Start, ibm->m_Count);
		else 
			imesh->setRange(ibm->m_Start + imesh->getRange().getStart(), imesh->getRange().getSize());
	} else if (typeid(*mesh) == typeid(VertexMesh)) {
		VertexMesh* vmesh = (VertexMesh*)mesh;

		VbTuple vbt;
		for (unsigned int i = 0; i < VertexBuffer::MaxBuffers; i++) {
			vbt[i] = vmesh->getVertexBuffer(i);
			vmesh->detachVertexBuffer(i);
		}

		Map<VbTuple, VbMapping>::Iterator vbm = m_VbMapping.at(vbt);	
		ASSERT(vbm != m_VbMapping.end());

		vmesh->attachVertexBuffer(vbm->m_Target->m_Buffer);
		if (vmesh->getRange().getSize() == 0)
			vmesh->setRange(vbm->m_Start, vbm->m_Count);
		else 
			vmesh->setRange(vbm->m_Start + vmesh->getRange().getStart(), vmesh->getRange().getSize());
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void MeshMerger::normalizeFormat(VertexFormat* ) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned int MeshMerger::addVertexBuffers(const VbTuple& vbt) {
	VertexFormat vfmt;
	VertexBuffer* vb;
	VbMapping vbm;
	unsigned int size,i;

	// Check if we already have this tuple
	if (m_VbMapping.contains(vbt)) {
		return m_VbMapping.at(vbt)->m_Start;
	}

	// Get merged vertex format
	size = 0;
	for (i = 0; i < VertexBuffer::MaxBuffers; i++) {
		vb = vbt[i];
		if (vb != NULL) {
			vfmt.append(vb->getFormat());
			size = max(size, vb->getSize());
		}
	}

	if (size == 0)
		return 0;

	// Normalize vertex format
	normalizeFormat(&vfmt);

	// Search Vertex Buffer format
	List<VbInfo>::Iterator iter, end;
	end = m_VbInfo.end();
	for (iter = m_VbInfo.begin(); iter != end; iter++) {
		if (iter->m_Format == vfmt) {
			vbm.m_Target = &(*iter);
			vbm.m_Start = iter->m_Size;
			vbm.m_Count = size;
			iter->m_Size += size;
			break;
		}
	}

	// Create new VertexBuffer Info object if no format was found
	if (iter == end) {
		VbInfo vbi;
		vbi.m_Format = vfmt;
		vbi.m_Size = size;
		vbi.m_Buffer = NULL;
		iter = m_VbInfo.insert(vbi);

		vbm.m_Target = &(*iter);
		vbm.m_Start = 0;
		vbm.m_Count = size;
	}

	// Insert VertexBuffer Mapping
	m_VbMapping.insert(vbt, vbm);

	return vbm.m_Start;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned int MeshMerger::addIndexBuffer(IndexBuffer* ib, unsigned int shift) {
	IndexBuffer::Type ifmt;
	IbMapping ibm;
	unsigned int size;

	if (ib == NULL)
		return 0;

	// Check if we already have this tuple
	if (m_IbMapping.contains(ib)) {
		return m_IbMapping.at(ib)->m_Start;
	}

	ifmt = ib->getType();
	size = ib->getSize();

	if (size == 0)
		return 0;

	// Search Index Buffer format
	List<IbInfo>::Iterator iter, end;
	end = m_IbInfo.end();
	for (iter = m_IbInfo.begin(); iter != end; iter++) {
		if (iter->m_Format == ifmt) {
			ibm.m_Target = &(*iter);
			ibm.m_Start = iter->m_Size;
			ibm.m_Count = size;
			ibm.m_Shift = shift;
			iter->m_Size += size;
			break;
		}
	}

	// Create new IndexBuffer Info object if no format was found
	if (iter == end) {
		IbInfo ibi;
		ibi.m_Format = ifmt;
		ibi.m_Size = size;
		ibi.m_Buffer = NULL;
		iter = m_IbInfo.insert(ibi);

		ibm.m_Target = &(*iter);
		ibm.m_Start = 0;
		ibm.m_Count = size;
		ibm.m_Shift = shift;
	}

	// Insert IndexBuffer Mapping
	m_IbMapping.insert((IndexBuffer*)ib, ibm);

	return ibm.m_Start;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void MeshMerger::mergeVertexBuffers() {
	List<VbInfo>::Iterator vbiiter, vbiend;
	vbiend = m_VbInfo.end();

	// Create index buffers
	for (vbiiter = m_VbInfo.begin(); vbiiter != vbiend; vbiiter++) {
		vbiiter->m_Buffer = VertexBuffer::create(vbiiter->m_Format, vbiiter->m_Size, VertexBuffer::Static);
	}

	// Copy old vertex buffers
	Map<VbTuple, VbMapping>::Iterator iter, end;
	end = m_VbMapping.end();
	for (iter = m_VbMapping.begin(); iter != end; iter++) {
		VbTuple vbt = iter.key();
		VbMapping vbm = iter.item();
		VertexBuffer* target = vbm.m_Target->m_Buffer;
		VertexBuffer* source;
		unsigned int tstride = target->getStride();
		unsigned int sstride;

		unsigned int offset = 0;
		unsigned int i, j;
		for (i = 0; i < VertexBuffer::MaxBuffers; i++) {
			source = vbt[i];
			if (source != NULL) {
				// Lock target and advance to correct position
				void* dst = (void *)(((byte *)target->lock(VertexBuffer::WriteOnly)) + offset + vbm.m_Start*tstride);
				void* src = source->lock(VertexBuffer::ReadOnly);

				sstride = source->getStride();

				// Copy entries
				for (j = 0; j < vbm.m_Count; j++) {
					memcpy(dst, src, sstride);
					dst = (void *)(((byte *)dst) + tstride);
					src = (void *)(((byte *)src) + sstride);
				}

				// Advance to next offset
				offset += sstride;

				source->unlock();
				target->unlock();
			}
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void MeshMerger::mergeIndexBuffers() {
	List<IbInfo>::Iterator ibiiter, ibiend;
	ibiend = m_IbInfo.end();

	// Create vertex buffers
	for (ibiiter = m_IbInfo.begin(); ibiiter != ibiend; ibiiter++) {
		ibiiter->m_Buffer = IndexBuffer::create(ibiiter->m_Format, ibiiter->m_Size, IndexBuffer::Static);
	}

	// Copy old index buffers
	Map<IndexBuffer*, IbMapping>::Iterator iter, end;
	end = m_IbMapping.end();
	for (iter = m_IbMapping.begin(); iter != end; iter++) {
		unsigned int i, shift;
		IbMapping ibm = iter.item();
		IndexBuffer* source = iter.key();
		IndexBuffer* target = ibm.m_Target->m_Buffer;
		shift = ibm.m_Shift;

		// Lock target and advance to correct position
		void* dst;
		void* src;

		switch(source->getType()) {
			case IndexBuffer::Index8:		
				src = source->lock(IndexBuffer::ReadOnly);
				dst = (void *)(((byte *)target->lock(IndexBuffer::WriteOnly)) + ibm.m_Start);

				for (i = 0; i < ibm.m_Count; i++) {
					*((byte *)dst) = (byte)((*((byte *)src)) + shift);
					dst = (void *)((byte *)dst + 1);
					src = (void *)((byte *)src + 1);
				}

				source->unlock();
				target->unlock();
				break;
			case IndexBuffer::Index16:		
				src = source->lock(IndexBuffer::ReadOnly);
				dst = (void *)(((word *)target->lock(IndexBuffer::WriteOnly)) + ibm.m_Start);

				for (i = 0; i < ibm.m_Count; i++) {
					*((word *)dst) = (word)((*((word *)src)) + shift);
					dst = (void *)((word *)dst + 1);
					src = (void *)((word *)src + 1);
				}

				source->unlock();
				target->unlock();
				break;
			case IndexBuffer::Index32:		
				src = source->lock(IndexBuffer::ReadOnly);
				dst = (void *)(((dword *)target->lock(IndexBuffer::WriteOnly)) + ibm.m_Start);

				for (i = 0; i < ibm.m_Count; i++) {
					*((dword *)dst) = (dword)((*((dword *)src)) + shift);
					dst = (void *)((dword *)dst + 1);
					src = (void *)((dword *)src + 1);
				}

				source->unlock();
				target->unlock();
				break;
		}
	}
}


}; };
