/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "graphics3d.optimizer.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.material.h"
#include "graphics3d.shader.h"
#include "graphics3d.pixelprogram.h"
#include "graphics3d.vertexprogram.h"
#include "graphics3d.mode.h"
#include "graphics3d.appearance.h"
#include "graphics3d.extensions.h"
#include "graphics3d.context.h"
#include "graphics3d.indexmesh.h"
#include "graphics3d.vertexmesh.h"
#include "graphics3d.renderstream.h"
#include "graphics3d.meshmerger.h"

#include "object.inl"
#include "object.pointer.inl"
#include "base.string.inl"
#include "math.range.inl"
#include "container.array.inl"
#include "container.map.inl"
#include "container.iterator.inl"
#include "container.tuple.inl"
#include "container.allocator.inl"
#include "graphics3d.optimizer.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.indexmesh.inl"
#include "graphics3d.vertexmesh.inl"
#include "graphics3d.renderstream.inl"
#include "graphics3d.meshmerger.inl"
#include "graphics3d.mesh.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.appearance.inl"
#include "graphics3d.pixelprogram.inl"
#include "graphics3d.vertexprogram.inl"

#include <typeinfo.h>


namespace magnum {namespace graphics3d {
using namespace magnum::container;
using namespace magnum::math;


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Optimizer::Optimizer() {
	m_Strategy = SortShader;
	clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Optimizer::~Optimizer() {
	clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Optimizer::clear() {
	unsigned int i;

	m_CurrentMaterial = NULL;
	m_CurrentShader = NULL;
	m_CurrentMode = NULL;
	m_CurrentIndexBuffer = NULL;
	for (i = 0; i < Attributes::MaxAttributes; i++)
		m_CurrentVertexBuffers[i] = NULL;
	for (i = 0; i < Texture::MaxTextures; i++)
		m_CurrentTextures[i] = NULL;

	m_PreselectedMaterial = NULL;
	m_PreselectedShader = NULL;
	m_PreselectedMode = NULL;
	m_PreselectedIndexBuffer = NULL;
	for (i = 0; i < Attributes::MaxAttributes; i++)
		m_PreselectedVertexBuffers[i] = NULL;
	for (i = 0; i < Texture::MaxTextures; i++)
		m_PreselectedTextures[i] = NULL;

	m_Jobs.clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Optimizer::renderMesh(Mesh* mesh) {
	Job job;
	unsigned int i, j, attribs;
	VertexBuffer* vb;
	Appearance* app;

	// Get type of mesh:
	if (typeid(*mesh) == typeid(IndexMesh)) {
		IndexMesh* imesh = (IndexMesh*)mesh;
		job.m_Start = imesh->getRange().getStart();
		job.m_Count = imesh->getRange().getSize();
		if (job.m_Count == 0)
			job.m_Count = imesh->getIndexBuffer()->getSize();
		job.m_Mesh = NULL;
		job.m_IndexBuffer = imesh->getIndexBuffer();
		for (i = 0; i < VertexBuffer::MaxBuffers; i++) {
			vb = imesh->getVertexBuffer(i);
			if (vb == NULL)
				continue;
			attribs = vb->getAttributes();
			for (j = 0; j < Attributes::MaxAttributes; j++) {
				if ((attribs & (1 << j)) == 0)
					continue;
				job.m_VertexBuffers[j] = vb;
			}
		}
	} else if (typeid(*mesh) == typeid(VertexMesh)) {
		VertexMesh* vmesh = (VertexMesh*)mesh;
		job.m_Start = vmesh->getRange().getStart();
		job.m_Count = vmesh->getRange().getSize();
		job.m_Mesh = NULL;
		job.m_IndexBuffer = NULL;
		for (i = 0; i < VertexBuffer::MaxBuffers; i++) {
			vb = vmesh->getVertexBuffer(i);
			if (vb == NULL)
				continue;
			attribs = vb->getAttributes();
			for (j = 0; j < Attributes::MaxAttributes; j++) {
				if ((attribs & (1 << j)) == 0)
					continue;
				job.m_VertexBuffers[j] = vb;
			}
		}
	} else {
		job.m_Start = 0;
		job.m_Count = 0;
		job.m_Mesh = mesh;
		job.m_IndexBuffer = NULL;
		for (i = 0; i < Attributes::MaxAttributes; i++) 
			job.m_VertexBuffers[i] = NULL;
	}

	job.m_Primitive = mesh->getPrimitive();
	app = mesh->getAppearance();
	if (app != NULL) {
		job.m_Material = app->getMaterial();
		job.m_Shader = app->getShader();
		for (i = 0; i < Texture::MaxTextures; i++)
			job.m_Textures[i] = app->getTexture(i);
	} else {
		job.m_Material = NULL;
		job.m_Shader = NULL;
		for (i = 0; i < Texture::MaxTextures; i++)
			job.m_Textures[i] = NULL;
	}

	job.m_Mode = m_CurrentMode;

	if (job.m_Material == NULL)
		job.m_Material = m_CurrentMaterial;
	else
		m_CurrentMaterial = job.m_Material;

	if (job.m_Shader == NULL)
		job.m_Shader = m_CurrentShader;
	else
		m_CurrentShader = job.m_Shader;

	if (job.m_IndexBuffer != NULL)
		m_CurrentIndexBuffer = job.m_IndexBuffer;

	for (i = 0; i < Attributes::MaxAttributes; i++) {
		if (job.m_VertexBuffers[i] != NULL)
			m_CurrentVertexBuffers[i] = job.m_VertexBuffers[i];
	}

	for (i = 0; i < Texture::MaxTextures; i++) {
		if (job.m_Textures[i] != NULL)
			m_CurrentTextures[i] = job.m_Textures[i];
	}

	m_Jobs.insert(job);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Optimizer::renderVertices(Mesh::Primitive p, unsigned int start, unsigned int size) {
	Job job;
	unsigned int i;
	unsigned int attribs;

	job.m_Primitive = p;
	job.m_Start = start;
	job.m_Count = size;
	job.m_Mesh = NULL;
	job.m_IndexBuffer = NULL;
	job.m_Shader = m_CurrentShader;
	job.m_Mode = m_CurrentMode;
	job.m_Material = m_CurrentMaterial;

	if (m_CurrentShader != NULL)
		attribs = m_CurrentShader->getAttributes();
	else
		attribs = 0xffffffff;

	for (i = 0; i < Attributes::MaxAttributes; i++) {
		if (attribs & (1 << i))
			job.m_VertexBuffers[i] = m_CurrentVertexBuffers[i];
		else
			job.m_VertexBuffers[i] = NULL;
	}

	for (i = 0; i < Texture::MaxTextures; i++) {
		job.m_Textures[i] = m_CurrentTextures[i];
	}

	m_Jobs.insert(job);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Optimizer::renderIndices(Mesh::Primitive p, unsigned int start, unsigned int size) {
	Job job;
	unsigned int i;
	unsigned int attribs;

	job.m_Primitive = p;
	job.m_Start = start;
	job.m_Count = size;
	job.m_Mesh = NULL;
	job.m_IndexBuffer = m_CurrentIndexBuffer;
	job.m_Shader = m_CurrentShader;
	job.m_Mode = m_CurrentMode;
	job.m_Material = m_CurrentMaterial;

	if (m_CurrentShader != NULL)
		attribs = m_CurrentShader->getAttributes();
	else
		attribs = 0xffffffff;

	for (i = 0; i < Attributes::MaxAttributes; i++) {
		if (attribs & (1 << i))
			job.m_VertexBuffers[i] = m_CurrentVertexBuffers[i];
		else
			job.m_VertexBuffers[i] = NULL;
	}

	for (i = 0; i < Texture::MaxTextures; i++) {
		job.m_Textures[i] = m_CurrentTextures[i];
	}

	m_Jobs.insert(job);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Optimizer::finish() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
RenderStream* Optimizer::optimize() {
	// Split up index buffers that are used for more than one mesh
	splitBuffers();

	// Now sort jobs
	qsort(m_Jobs.getArray(), m_Jobs.getSize(), sizeof(Job), cmpJobs);

	// Merge buffers
	mergeBuffers();

	// Merge jobs
	mergeJobs();

	return createStream();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
RenderStream* Optimizer::createStream() {
	// Loop through all jobs again
	unsigned int i, j;
    const Material* curMaterial = m_PreselectedMaterial;
    const Shader* curShader = m_PreselectedShader;
    const Mode* curMode = m_PreselectedMode;
    const IndexBuffer* curIndexBuffer = m_PreselectedIndexBuffer;
    const VertexBuffer* curVertexBuffers[Attributes::MaxAttributes];
	const Texture* curTextures[Texture::MaxTextures];
	unsigned int curStart = 0xffffffff, curEnd = 0xffffffff;
	for (i = 0; i < Texture::MaxTextures; i++)
		curTextures[i] = m_PreselectedTextures[i];
	for (i = 0; i < Attributes::MaxAttributes; i++)
		curVertexBuffers[i] = m_PreselectedVertexBuffers[i];

	m_Compiler.clear();

	for (i = 0; i < m_Jobs.getSize(); i++) {
		if (m_Jobs[i].m_Mesh != NULL) {
			continue;
		}
		if (m_Jobs[i].m_Mode != curMode) {
			curMode = m_Jobs[i].m_Mode;
			if (curMode != NULL)
				m_Compiler.selectMode(curMode);
			else
				m_Compiler.deselectMode();
		}
		if (m_Jobs[i].m_Shader != curShader) {
			curShader = m_Jobs[i].m_Shader;
			if (curShader != NULL)
				m_Compiler.selectShader(curShader);
			else
				m_Compiler.deselectShader();
		}
		if (m_Jobs[i].m_Material != curMaterial) {
			curMaterial = m_Jobs[i].m_Material;
			if (curMaterial != NULL)
				m_Compiler.selectMaterial(curMaterial);
		}
		if (m_Jobs[i].m_IndexBuffer != curIndexBuffer) {
			curIndexBuffer = m_Jobs[i].m_IndexBuffer;
			if (curIndexBuffer != NULL)
				m_Compiler.selectIndexBuffer(curIndexBuffer);
		}

		Map<const VertexBuffer*, unsigned int> vbAttribs;
		Map<const VertexBuffer*, unsigned int>::Iterator iter;
		for (j = 0; j < Attributes::MaxAttributes; j++) {
			const VertexBuffer* vb = m_Jobs[i].m_VertexBuffers[j];
			if (vb == NULL) 
				continue;
			if (vb != curVertexBuffers[j]) {
				curVertexBuffers[j] = vb;
				iter = vbAttribs.at(vb);
				if (iter == vbAttribs.end()) {
					vbAttribs.insert(vb, 1 << j);
				} else {
					*iter = *iter | (1 << j);
				}
			}
		}
		for (iter = vbAttribs.begin(); iter != vbAttribs.end(); iter++) {
			m_Compiler.selectVertexBuffer(iter.key(), iter.item());
		}

		for (j = 0; j < Texture::MaxTextures; j++) {
			if (m_Jobs[i].m_Textures[j] != curTextures[j]) {
				curTextures[j] = m_Jobs[i].m_Textures[j];
				if (curTextures[j] != NULL)
					m_Compiler.selectTexture(curTextures[j], j);
			}
		}

		if (curIndexBuffer != NULL) {
			m_Compiler.renderIndices(m_Jobs[i].m_Primitive, m_Jobs[i].m_Start, m_Jobs[i].m_Count);
		} else {
			m_Compiler.renderVertices(m_Jobs[i].m_Primitive, m_Jobs[i].m_Start, m_Jobs[i].m_Count);
		}
	}

	m_Compiler.finish();

	return m_Compiler.compile();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Optimizer::mergeJobs() {
	// Loop through all jobs again
    const Material* curMaterial = NULL;
    const Shader* curShader = NULL;
    const Mode* curMode = NULL;
    const IndexBuffer* curIndexBuffer = NULL;
    const VertexBuffer* curVertexBuffers[Attributes::MaxAttributes];
	const Texture* curTextures[Texture::MaxTextures];
	unsigned int curStart = 0xffffffff, curEnd = 0xffffffff;
	Mesh::Primitive curPrimitive = Mesh::Unknown;
	memset(curVertexBuffers, 0, sizeof(curVertexBuffers));
	memset(curTextures, 0, sizeof(curTextures));

	unsigned int i, j;
	bool equal;
	for (i = 0; i < m_Jobs.getSize(); i++) {
		if (m_Jobs[i].m_Mesh != NULL) {
			continue;
		}

		equal = true;
		if (m_Jobs[i].m_Primitive != curPrimitive) {
			curPrimitive = m_Jobs[i].m_Primitive;
			equal = false;
		}
		if (m_Jobs[i].m_Mode != curMode) {
			curMode = m_Jobs[i].m_Mode;
			equal = false;
		}
		if (m_Jobs[i].m_Shader != curShader) {
			curShader = m_Jobs[i].m_Shader;
			equal = false;
		}
		if (m_Jobs[i].m_Material != curMaterial) {
			curMaterial = m_Jobs[i].m_Material;
			equal = false;
		}
		if (m_Jobs[i].m_IndexBuffer != curIndexBuffer) {
			curIndexBuffer = m_Jobs[i].m_IndexBuffer;
			equal = false;
		}
		for (j = 0; j < Attributes::MaxAttributes; j++) {
			if (m_Jobs[i].m_VertexBuffers[j] != curVertexBuffers[j]) {
				curVertexBuffers[j] = m_Jobs[i].m_VertexBuffers[j];
				equal = false;
			}
		}
		for (j = 0; j < Texture::MaxTextures; j++) {
			if (m_Jobs[i].m_Textures[j] != curTextures[j]) {
				curTextures[j] = m_Jobs[i].m_Textures[j];
				equal = false;
			}
		}

		// Currently we cannot handle Triangle Strips
		if (curPrimitive == Mesh::TriangleStrip)
			equal = false;

		// Check if we can merge two subsequent jobs...
		if ((m_Jobs[i].m_Start == curEnd) && equal) {
			m_Jobs[i-1].m_Count += m_Jobs[i].m_Count;
			curEnd += m_Jobs[i].m_Count;
			m_Jobs.remove(i);
			i--;
			continue;
		}

		curStart = m_Jobs[i].m_Start;
		curEnd = m_Jobs[i].m_Start + m_Jobs[i].m_Count;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Optimizer::mergeBuffers() {
	MeshMerger merger;
	Tuple<VertexBuffer*, VertexBuffer::MaxBuffers> vbt;
	IndexBuffer* ib;

	// Collect buffer infos
	unsigned int i, j;
	for (i = 0; i < m_Jobs.getSize(); i++) {
		if (m_Jobs[i].m_Mesh != NULL)
			continue;

		vbt = collectVertexBuffers(m_Jobs[i].m_VertexBuffers);
		ib = m_Jobs[i].m_IndexBuffer;
		merger.add(vbt, ib);
	}

	// merge buffers
	merger.finish();

	// Walk through all jobs and replace buffers
	for (i = 0; i < m_Jobs.getSize(); i++) {
		if (m_Jobs[i].m_Mesh != NULL)
			continue;

		vbt = collectVertexBuffers(m_Jobs[i].m_VertexBuffers);
		ib = m_Jobs[i].m_IndexBuffer;

		MeshMerger::VertexMapping vmap = merger.getMapping(vbt);
		for (j = 0; j < Attributes::MaxAttributes; j++) {
			if (m_Jobs[i].m_VertexBuffers[j] != NULL) 
				m_Jobs[i].m_VertexBuffers[j] = vmap.m_Buffer;
		}

		MeshMerger::IndexMapping imap = merger.getMapping(ib);
		if (m_Jobs[i].m_IndexBuffer != NULL) {
			m_Jobs[i].m_IndexBuffer = imap.m_Buffer;
			m_Jobs[i].m_Start += imap.m_Start;
		} else {
			m_Jobs[i].m_Start += vmap.m_Start;
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Optimizer::splitBuffers() {
	unsigned int i;

	// Walk through all jobs and replace buffers
	for (i = 0; i < m_Jobs.getSize(); i++) {
		if (m_Jobs[i].m_Mesh != NULL)
			continue;

		if (m_Jobs[i].m_IndexBuffer == NULL)
			continue;
		
		IndexBuffer* src = m_Jobs[i].m_IndexBuffer;
		if (m_Jobs[i].m_Count == src->getSize())
			continue;
			
		unsigned int size = m_Jobs[i].m_Count;
		unsigned int start = m_Jobs[i].m_Start;
		IndexBuffer* dst = IndexBuffer::create(src->getType(), size, IndexBuffer::System);
		ASSERT(dst != NULL);
		if (dst == NULL)
			continue;
		
		// Copy index buffer
		void* srcdata = src->lock(IndexBuffer::ReadOnly);
		srcdata = (byte *)srcdata + start * src->getType();
		void* dstdata = dst->lock(IndexBuffer::Discard);
		memcpy(dstdata, srcdata, size*src->getType());
		dst->unlock();
		src->unlock();

		m_Jobs[i].m_IndexBuffer = dst;
		m_Jobs[i].m_Start = 0;
		dst->release();
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
int Optimizer::cmpJobs(const void* a,const void* b) {
	unsigned int i;
	Job* ja = (Job*)a;
	Job* jb = (Job*)b;

	// check mode
	if (ja->m_Mode != jb->m_Mode)
		return (int)((intptr_t)(void *)ja->m_Mode - (intptr_t)(void *)jb->m_Mode);

	// check shader
	if (ja->m_Shader != jb->m_Shader)
		return (int)((intptr_t)(void *)ja->m_Shader - (intptr_t)(void *)jb->m_Shader);

	// check textures
	for (i = 0; i < Texture::MaxTextures; i++) {
		if (ja->m_Textures[i] == jb->m_Textures[i])
			continue;
		return (int)((intptr_t)(void*)ja->m_Textures[i] - (intptr_t)(void*)jb->m_Textures[i]);
	}

	// check material
	if (ja->m_Material != jb->m_Material)
		return (int)((intptr_t)(void *)ja->m_Material - (intptr_t)(void *)jb->m_Material);

	// check index buffer
	if (ja->m_IndexBuffer != jb->m_IndexBuffer)
		return (int)((intptr_t)(void *)ja->m_IndexBuffer - (intptr_t)(void *)jb->m_IndexBuffer);
	
	// check vertex buffers
	for (i = 0; i < Attributes::MaxAttributes; i++) {
		if (ja->m_VertexBuffers[i] == jb->m_VertexBuffers[i])
			continue;
		return (int)((intptr_t)(void*)ja->m_VertexBuffers[i] - (intptr_t)(void*)jb->m_VertexBuffers[i]);
	}

	// check ranges
	if (ja->m_Start != jb->m_Start)
		return ja->m_Start - jb->m_Start;

	// check primitives
	if (ja->m_Primitive != jb->m_Primitive)
		return ja->m_Primitive - jb->m_Primitive;

	return 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Tuple<VertexBuffer*,VertexBuffer::MaxBuffers> Optimizer::collectVertexBuffers(const Pointer<VertexBuffer>* vb) {
	Tuple<VertexBuffer*,VertexBuffer::MaxBuffers> vbt;
	unsigned int i,j;

	for (j = 0; j < VertexBuffer::MaxBuffers; j++)
		vbt[j] = NULL;

	for (j = 0; j < Attributes::MaxAttributes; j++) {
		if (vb[j] == NULL) 
			continue;
		if (vbt.contains(vb[j]))
			continue;
		for (i = 0; i < VertexBuffer::MaxBuffers; i++) {
			if (vbt[i] == NULL) {
				vbt[i] = vb[j];
				break;
			}
		}
	}

	return vbt;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned int Optimizer::collectAttributes(const Tuple<VertexBuffer*,Attributes::MaxAttributes>& vbt) {
	unsigned int i, attribs = 0;

	for (i = 0; i < Attributes::MaxAttributes; i++) {
		if(vbt[i] != NULL) {
			attribs |= (1 << i);
		}
	}

	return attribs;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned int Optimizer::collectAttributes(const Tuple<VertexBuffer*,Attributes::MaxAttributes>& vbt,const VertexBuffer* vb) {
	unsigned int i, attribs = 0;

	for (i = 0; i < Attributes::MaxAttributes; i++) {
		if(vbt[i] == vb) {
			attribs |= (1 << i);
		}
	}

	return attribs;
}

}; };
