/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "io.exception.h"
#include "io.inputstream.h"
#include "io.sysfile.h"
#include "container.heap.h"
#include "graphics3d.context.h"
#include "graphics3d.shader.h"
#include "graphics3d.material.h"
#include "graphics3d.texture.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.indexmesh.h"
#include "graphics3d.meshgroup.h"
#include "graphics3d.framemodel.h"
#include "graphics3d.pixelprogram.h"
#include "graphics3d.vertexprogram.h"
#include "graphics3d.renderstream.h"
#include "graphics3d.appearance.h"
#include "graphics3d.texturemanager.h"
#include "graphics3d.shadermanager.h"
#include "graphics3d.translator.md2.h"

#include "object.inl"
#include "object.pointer.inl"
#include "object.exception.inl"
#include "base.string.inl"
#include "io.inputstream.inl"
#include "io.translator.inl"
#include "container.heap.inl"
#include "container.map.inl"
#include "container.array.inl"
#include "container.iterator.inl"
#include "container.pair.inl"
#include "container.chain.inl"
#include "container.allocator.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.material.inl"
#include "graphics3d.texture.inl"
#include "graphics3d.context.inl"
#include "graphics3d.meshgroup.inl"
#include "graphics3d.mesh.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.indexmesh.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.model.inl"
#include "graphics3d.appearance.inl"
#include "graphics3d.framemodel.inl"
#include "graphics3d.translator.inl"
#include "graphics3d.translator.md2.inl"

#define MD2_MAGIC		0x32504449


namespace magnum { namespace graphics3d {
using magnum::container::Heap;

static float vertexnormals[][3] = {
#include "graphics3d.translator.md2normals.h"
};


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
TranslatorMD2::TranslatorMD2() {
	m_Model = NULL;
	m_Mesh = NULL;
	m_MeshGroup = NULL;
	m_Triangles = NULL;
	m_Vertices = NULL;
	m_TexCoords = NULL;
	m_In = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Model* TranslatorMD2::onDecode(InputStream* in,long size) {
	if (size == Autosize)
		size = in->available();

	m_Frames.clear();

	m_Model = NULL;
	m_Mesh = NULL;
	m_MeshGroup = NULL;
	m_Triangles = NULL;
	m_Vertices = NULL;
	m_TexCoords = NULL;
	m_In = in;

	// Read in Header
	Header head;

	if (in->read(&head, sizeof(head)) != sizeof(head))
		throw magnum::io::TranslatorFormatException();
	if (head.m_Magic != Magic)
		throw magnum::io::TranslatorFormatException();

	int position = sizeof(head);

	m_TriangleCount = head.m_NumTriangles;
	m_FrameSize = head.m_FrameSize;

	int offsetSkins = head.m_OffsetSkins; 
	int offsetTexCoords = head.m_OffsetTexCoords;
	int offsetTriangles = head.m_OffsetTriangles;
	int offsetFrames = head.m_OffsetFrames;
	int offsetGlCommands = head.m_OffsetGlCommands;
	int skip;

	// Get next chunk to read
	Heap<int> ofHeap;
	ofHeap.insert(-offsetSkins);
	ofHeap.insert(-offsetTexCoords);
	ofHeap.insert(-offsetTriangles);
	ofHeap.insert(-offsetFrames);
	ofHeap.insert(-offsetGlCommands);
	for (unsigned int i= 0; i < 5; i++) {
		int nextPos = -ofHeap.remove();
		if (nextPos < position)
			throw magnum::io::TranslatorFormatException();

		m_In->skip(nextPos - position);
		position += nextPos - position;

		if (nextPos == offsetSkins) {
			skip = readSkins(head.m_NumSkins);	
		} else if (nextPos == offsetTexCoords) {
			skip = readTexCoords(head.m_NumTexCoords);
		} else if (nextPos == offsetTriangles) {
			skip = readTriangles(head.m_NumTriangles);
		} else if (nextPos == offsetFrames) {
			skip = readFrames(head.m_NumFrames);
		} else if (nextPos == offsetGlCommands) {
			skip = readGlCommands(head.m_OffsetGlCommands);
		}

		position += skip;
	}

	// CReate Model etc
	m_Model = new FrameModel();
	m_MeshGroup = new MeshGroup();
	m_Mesh = new IndexMesh();
	m_Mesh->setPrimitive(Mesh::TriangleList);

	m_MeshGroup->insert(m_Mesh);
	m_Mesh->release();
	m_Model->insertGroup(m_MeshGroup);
	m_MeshGroup->release();

	// Process model data
	createIndexMapping();
	createTextureCoords();
	createFrames();
	createIndices();

	// finish Model
	m_Model->finish();

	// Release memory
	m_VertexSource.clear();
	m_TextureSource.clear();
	m_VertexMapping.clear();
	for (i = 0; i < m_Frames.getSize(); i++)
		delete[] (byte*)m_Frames[i];
	m_Frames.clear();
	if (m_TexCoords != NULL) {
		delete[] m_TexCoords;
		m_TexCoords = NULL;
	}
	if (m_Triangles != NULL) {
		delete[] m_Triangles;
		m_Triangles = NULL;
	}
	if (m_Vertices != NULL) {
		delete[] m_Vertices;
		m_Vertices = NULL;
	}

	return m_Model;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned int TranslatorMD2::readSkins(unsigned int count) {
	char skin[64];
	unsigned int read = 0;

	for (unsigned int i = 0; i < count; i++) {
		read += m_In->read(skin, sizeof(skin));
	}

	return read;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned int TranslatorMD2::readTexCoords(unsigned int count) {
	long read;

	m_TexCoords = new TextureCoordinate[count];
	read = m_In->read(m_TexCoords, sizeof(TextureCoordinate)*count);

	return (int)read;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned int TranslatorMD2::readTriangles(unsigned int count) {
	long read;

	m_Triangles = new Triangle[count];
	read = m_In->read(m_Triangles, sizeof(Triangle)*count);

	return (int)read;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned int TranslatorMD2::readFrames(unsigned int count) {
	long read = 0;

	for (unsigned int i = 0; i < count; i++) {
		Frame* frame = (Frame *)new byte[m_FrameSize];
		read += m_In->read(frame, m_FrameSize);
		m_Frames.insert(frame);
	}

	return (int)read;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned int TranslatorMD2::readGlCommands(unsigned int count) {
	return 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorMD2::createIndexMapping() {
	m_VertexSource.clear();
	m_TextureSource.clear();
	m_VertexMapping.clear();

	unsigned int vertex = 0;

	for (unsigned int i = 0; i < m_TriangleCount; i++) {
		IPair ip;
		for (unsigned int j = 0; j < 3; j++) {
			ip = IPair(m_Triangles[i].m_VertexIndices[j],m_Triangles[i].m_TextureIndices[j]);
			if (m_VertexMapping.contains(ip))
				continue;
			m_VertexMapping.insert(ip, vertex);
			m_VertexSource.insert(m_Triangles[i].m_VertexIndices[j]);
			m_TextureSource.insert(m_Triangles[i].m_TextureIndices[j]);
			vertex++;
		}
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorMD2::createTextureCoords() {
	unsigned int count = m_TextureSource.getSize();

	// Allocate Vertex Buffer
	VertexFormat fmt;
	VertexBuffer* vb;
	fmt.set(0, Types::Float2, Attributes::TextureAll);
	vb = VertexBuffer::create(fmt, count, VertexBuffer::Static);
	m_Mesh->attachVertexBuffer(vb, 0);
	vb->release();

	Vector2<float>* texcords = (Vector2<float>*)vb->lock(VertexBuffer::Discard);
	unsigned int texsrc;
	float u,v;

	for (unsigned int i = 0; i < count; i++) {
		texsrc = m_TextureSource[i];
		u = (float)m_TexCoords[texsrc].s / (float)m_SkinWidth;
		v = (float)m_TexCoords[texsrc].t / (float)m_SkinHeight;
		texcords[i].x = u;
		texcords[i].y = v;
	}

	vb->unlock();
}



/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorMD2::createFrames() {
	unsigned int vc = m_VertexSource.getSize();
	unsigned int fc = m_Frames.getSize();

	VertexFormat fmt;
	fmt.set(0, Types::Float3, Attributes::Vertex);
	fmt.set(1, Types::Float3, Attributes::Normal);
	Frame* frame;
	VertexBuffer* vb;
	unsigned int idx;

	for (unsigned int f = 0; f < fc; f++) {
		frame = (Frame*)m_Frames[f];
		vb = VertexBuffer::create(fmt, vc, VertexBuffer::Static);
		vb->setName(frame->m_Name);
		m_Model->insertFrame(m_MeshGroup, vb, f);
		vb->release();

		void* buf = vb->lock(VertexBuffer::Discard);
		Vector3<float>* vptr = (Vector3<float>*)buf;
		Vector3<float>* nptr = vptr + 1;
		for (unsigned int i = 0; i < vc; i++) {
			idx = m_VertexSource[i];
			vptr->x = (float)frame->vertices[idx].m_Vertex[0] * frame->m_Scale[0] + frame->m_Translate[0];
			vptr->y = (float)frame->vertices[idx].m_Vertex[1] * frame->m_Scale[1] + frame->m_Translate[1];
			vptr->z = (float)frame->vertices[idx].m_Vertex[2] * frame->m_Scale[2] + frame->m_Translate[2];
			nptr->x = vertexnormals[frame->vertices[idx].m_LightNormalIndex][0];
			nptr->y = vertexnormals[frame->vertices[idx].m_LightNormalIndex][1];
			nptr->z = vertexnormals[frame->vertices[idx].m_LightNormalIndex][2];
			vptr += 2;
			nptr += 2;
		}
		vb->unlock();
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void TranslatorMD2::createIndices() {
	unsigned int tc = m_TriangleCount;
	IPair ip;

	IndexBuffer* ib = IndexBuffer::create(IndexBuffer::Index16, tc*3, IndexBuffer::Static);
	m_Mesh->attachIndexBuffer(ib);
	ib->release();

	word *indices = (word*)ib->lock(IndexBuffer::Discard);
	for (unsigned int i = 0; i < m_TriangleCount; i++) {
		ip = IPair(m_Triangles[i].m_VertexIndices[0],m_Triangles[i].m_TextureIndices[0]);
		indices[i*3] = m_VertexMapping[ip];
		ip = IPair(m_Triangles[i].m_VertexIndices[2],m_Triangles[i].m_TextureIndices[2]);
		indices[i*3 + 1] = m_VertexMapping[ip];
		ip = IPair(m_Triangles[i].m_VertexIndices[1],m_Triangles[i].m_TextureIndices[1]);
		indices[i*3 + 2] = m_VertexMapping[ip];
	}
	ib->unlock();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
long TranslatorMD2::onEncode(OutputStream* ,const Model* ) {
	return 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool TranslatorMD2::onIdentify(File* in) {
	dword id;

	try {
		unsigned long oldpos;
		oldpos = in->getPosition();	
		in->read(&id, sizeof(id));
		in->setPosition(oldpos, File::Absolute);
	} catch (...) {
		return false;
	}

	if (id != MD2_MAGIC)
		return false;

	return true;
}


}; };
