/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "graphics3d.vertexmesh.h"
#include "graphics3d.context.h"
#include "graphics3d.material.h"
#include "graphics3d.indexbuffer.h"
#include "graphics3d.vertexbuffer.h"
#include "graphics3d.shader.h"
#include "graphics3d.pixelprogram.h"
#include "graphics3d.vertexprogram.h"
#include "graphics3d.appearance.h"

#include "object.inl"
#include "base.string.inl"
#include "system.thread.inl"
#include "math.vector.inl"
#include "math.vector3.inl"
#include "math.matrix.inl"
#include "math.range.inl"
#include "geometry.trianglemesh.inl"
#include "geometry.bounds.inl"
#include "geometry.sphere.inl"
#include "geometry.box.inl"
#include "geometry.alignedbox.inl"
#include "container.array.inl"
#include "graphics3d.mesh.inl"
#include "graphics3d.vertexmesh.inl"
#include "graphics3d.context.inl"
#include "graphics3d.material.inl"
#include "graphics3d.texture.inl"
#include "graphics3d.indexbuffer.inl"
#include "graphics3d.vertexbuffer.inl"
#include "graphics3d.vertexformat.inl"
#include "graphics3d.shader.inl"
#include "graphics3d.pixelprogram.inl"
#include "graphics3d.vertexprogram.inl"
#include "graphics3d.resource.inl"
#include "graphics3d.appearance.inl"


namespace magnum { namespace graphics3d {
using magnum::geometry::Bounds;
using magnum::geometry::AlignedBox3;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
VertexMesh::VertexMesh() {
	m_Range.setStart(0);
	m_Range.setEnd(0);
	m_BufferCount = 0;
	m_Attributes = 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void VertexMesh::attachVertexBuffer(VertexBuffer* vb,unsigned int i) {
	ASSERT(i < VertexBuffer::MaxBuffers);
	ASSERT(vb != NULL);

	m_VertexBuffer[i] = vb;
	updateBufferInfo();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void VertexMesh::detachVertexBuffer(unsigned int i) {
	ASSERT(i < VertexBuffer::MaxBuffers);

	m_VertexBuffer[i] = NULL;
	updateBufferInfo();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void VertexMesh::releaseVertexBuffers() {
	for (int i = 0; i < VertexBuffer::MaxBuffers; i++) {
		m_VertexBuffer[i] = NULL;
	}
	m_BufferCount = 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void VertexMesh::onRender() const {
	selectVertexBuffers();

	if (getAppearance() != NULL)
		getAppearance()->select();

	if (m_Range.getSize()) 
		getContext()->renderVertices(getPrimitive(), m_Range.getStart(), m_Range.getSize());
	else
		getContext()->renderVertices(getPrimitive(), m_Range.getSize());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool VertexMesh::onBound(Bounds3<float>* bounds) const {
	ASSERT(bound != NULL);
	unsigned int i;

	// Search appropriate vertex buffer
	VertexBuffer* vb = NULL;
	for (i = 0; i < VertexBuffer::MaxBuffers; i++) {
		if (getVertexBuffer(i)->getAttributes() & Attributes::Vertex) {
			vb = getVertexBuffer(i);
			break;
		}
	}
	if (i == VertexBuffer::MaxBuffers)
		return false;

	// Find correct attribute and offset
	unsigned int offset;
	unsigned int type;
	unsigned int count;
	for (i = 0; i < VertexFormat::MaxComponents; i++) {
		if (vb->getAttribute(i) & Attributes::Vertex) {
			offset = vb->getOffset(i);
			type = vb->getType(i);
			count = vb->getCount(i);
			break;
		}
	}
	if (i == VertexFormat::MaxComponents)
		return false;
	
	// Get offset, stride etc	
	unsigned int numVertices = vb->getSize();
	unsigned int incVertices = vb->getStride();
	void* vertices = vb->lock(VertexBuffer::ReadOnly);
	if (vertices == NULL)
		return NULL;

	// Adjust buffer for offset and range	
	vertices = (void*)((byte *)vertices + offset);
	if (m_Range.getSize() > 0) {
		vertices = (void*)((byte *)vertices + m_Range.getStart()*incVertices);
		numVertices = m_Range.getSize();	
	}
	
	switch(type) {
		case Types::Byte3:
		case Types::Byte4: {
			AlignedBox3<byte> box;
			box.bound((Vector<byte,3>*)vertices, incVertices, numVertices);
			bounds->setBounds(AlignedBox3<float>(box.getMinimum(), box.getMaximum()));
			break;		
		}
		case Types::Integer3:
		case Types::Integer4: {
			AlignedBox3<int> box;
			box.bound((Vector<int,3>*)vertices, incVertices, numVertices);
			bounds->setBounds(AlignedBox3<float>(box.getMinimum(), box.getMaximum()));
			break;		
		}
		case Types::Float3:
		case Types::Float4: {
			AlignedBox3<float> box;
			box.bound((Vector<float,3>*)vertices, incVertices, numVertices);
			bounds->setBounds(AlignedBox3<float>(box.getMinimum(), box.getMaximum()));
			break;		
		}
		case Types::Double3:
		case Types::Double4: {
			AlignedBox3<double> box;
			box.bound((Vector<double,3>*)vertices, incVertices, numVertices);
			bounds->setBounds(AlignedBox3<float>(box.getMinimum(), box.getMaximum()));
			break;		
		}
		
		default:
			vb->unlock();
			return false;
	}
	
	vb->unlock();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
TriangleMesh3<float>* VertexMesh::onTriangulate() const {
	unsigned int i;
	float x,y,z;

	// Search appropriate vertex buffer
	VertexBuffer* vb = NULL;
	for (i = 0; i < VertexBuffer::MaxBuffers; i++) {
		if (getVertexBuffer(i)->getAttributes() & Attributes::Vertex) {
			vb = getVertexBuffer(i);
			break;
		}
	}
	if (i == VertexBuffer::MaxBuffers)
		return NULL;

	// Find correct attribute and offset
	unsigned int offset;
	unsigned int type;
	unsigned int count;
	for (i = 0; i < VertexFormat::MaxComponents; i++) {
		if (vb->getAttribute(i) & Attributes::Vertex) {
			offset = vb->getOffset(i);
			type = (vb->getType(i) >> Types::TypeShift) & Types::TypeMask;
			count = vb->getCount(i);
			break;
		}
	}
	if (i == VertexFormat::MaxComponents)
		return NULL;

	// Copy vertices
	unsigned int numVertices = vb->getSize();
	unsigned int incVertices = vb->getStride();
	void* vertices = vb->lock(VertexBuffer::ReadOnly);
	if (vertices == NULL)
		return NULL;
		
	// Adjust buffer for offset and range	
	vertices = (void*)((byte *)vertices + offset);
	if (m_Range.getSize() > 0) {
		vertices = (void*)((byte *)vertices + m_Range.getStart()*incVertices);
		numVertices = m_Range.getSize();	
	}

	// Create Triangle mesh
	TriangleMesh3<float>* mesh = new TriangleMesh3<float>();

#define copyVertex(T) \
		x = (float)((T*)vertices)[0]; \
		y = (count >= 2) ? (float)((T*)vertices)[1] : 0.0f; \
		z = (count >= 3) ? (float)((T*)vertices)[2] : 0.0f; \
		vertices = (void*)((byte *)vertices + incVertices); \
		mesh->setVertex(i, Vector3<float>(x,y,z));
#define copyVertices(T) \
		mesh->setVertices(numVertices); \
		for (i = 0; i < numVertices; i++) { \
			copyVertex(T); \
		}
	switch(type) {
		case ((Types::Byte >> Types::TypeShift) & Types::TypeMask):
			copyVertices(int8);
			break;
		case ((Types::UByte >> Types::TypeShift) & Types::TypeMask):
			copyVertices(uint8);
			break;
		case ((Types::Short >> Types::TypeShift) & Types::TypeMask):
			copyVertices(int16);
			break;
		case ((Types::UShort >> Types::TypeShift) & Types::TypeMask):
			copyVertices(uint16);
			break;
		case ((Types::Int >> Types::TypeShift) & Types::TypeMask):
			copyVertices(int32);
			break;
		case ((Types::UInt >> Types::TypeShift) & Types::TypeMask):
			copyVertices(uint32);
			break;
		case ((Types::Float >> Types::TypeShift) & Types::TypeMask):
			copyVertices(float);
			break;
		case ((Types::Double >> Types::TypeShift) & Types::TypeMask):
			copyVertices(double);
			break;
		default:
			break;
	}
	vb->unlock();

	// Create Triangles
	switch(getPrimitive()) {
		case TriangleList:
			for (i = 0; i < numVertices/3; i++) {
				mesh->setTriangle(i, i*3, i*3+1, i*3+2);
			}
			break;
		case TriangleStrip:
			for (i = 0; i < numVertices - 2; i++) {
				if (i & 1)
					mesh->setTriangle(i, i, i+1, i+2);
				else
					mesh->setTriangle(i, i+1, i, i+2);
			}
			break;
		case TriangleFan:
			for (i = 0; i < numVertices - 2; i++) {
				mesh->setTriangle(i, 0, i+1, i+2);
			}
			break;
		case QuadStrip:
		case QuadList:
		default:
			break;
	}

	return mesh;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void VertexMesh::updateBufferInfo() {
	m_VertexCount = 0;
	m_BufferCount = 0;
	m_Attributes = 0;
	for (int i = 0; i < VertexBuffer::MaxBuffers; i++) {
		if (m_VertexBuffer[i] == NULL)
			continue;

		m_BufferCount = i+1;
		m_Attributes |= m_VertexBuffer[i]->getAttributes();

		if (m_VertexCount == 0)
			m_VertexCount = m_VertexBuffer[i]->getSize();
		else
			m_VertexCount = min(m_VertexCount, m_VertexBuffer[i]->getSize());
	}
}


}; };
