/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "gui.container.h"
#include "gui.window.h"
#include "gui.context.h"
#include "gui.bitmap.h"
#include "container.chain.h"
#include "graphics.dimension.h"
#include "graphics.rectangle.h"
#include "application.message.h"

#include "base.string.inl"
#include "object.class.inl"
#include "object.inl"
#include "system.thread.inl"
#include "system.signal.inl"
#include "application.looper.inl"
#include "gui.view.inl"
#include "gui.container.inl"
#include "gui.window.inl"
#include "gui.context.inl"
#include "graphics.dimension.inl"
#include "graphics.rectangle.inl"
#include "graphics.point.inl"
#include "graphics.transform.inl"
#include "container.chain.inl"
#include "container.iterator.inl"


namespace magnum { namespace gui {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Container::Container(const Rectangle<float>& frame) : View(frame) {
	m_OldFrame = m_Frame;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Container::~Container() {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Container::insert(View* view) {
	if (view->getParent() != NULL)
		return false;
	if (view->getWindow() != NULL)
		return false;
	if (view == this)
		return false;

	m_Views.insert(view);
	view->m_Parent = this;
	view->updateTransform();

	if (getWindow() != NULL) {
		view->m_Window = getWindow();
		if (getWindow()->isAlive())
			view->onAttachWindow();
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Container::remove(View *view) {
	if (view->getParent() != this)
		return false;
	if (view->getWindow() != getWindow())
		return false;
	if (view == this)
		return false;

	if (getWindow() != NULL) {
		if (view->hasFocus())
			view->releaseFocus();
		if (getWindow()->isAlive())
			view->onDetachWindow();
		view->m_Window = NULL;
	}

	m_Views.remove(view);
	view->m_Parent = NULL;
	view->updateTransform();

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Container::clear() {
	View *view;

	while(m_Views.begin() != m_Views.end()) {
		view = &(*m_Views.begin());

		if (getWindow() != NULL) {
			if (view->hasFocus())
				view->releaseFocus();
			if (getWindow()->isAlive())
				view->onDetachWindow();
			view->m_Window = NULL;
		}

		m_Views.remove(view);
		view->m_Parent = NULL;
		view->updateTransform();
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Dimension<float> Container::getPreferredSize() const {
	Chain<View>::Iterator iter = m_Views.begin();
	View* view;
	Rectangle<float> rect = m_Frame;
	Rectangle<float> vrect;

	while (iter != m_Views.end()) {
		view = &(*iter);
		vrect.setOrigin(view->getFrame().getOrigin());
		vrect.setSize(view->getPreferredSize());
		rect |= vrect;
		iter++;
	}

	return rect.getSize();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Container::setPreferredSize(void) {
	Chain<View>::Iterator iter = m_Views.begin();
	View* view;
	Rectangle<float> rect = m_Frame;

	while (iter != m_Views.end()) {
		view = &(*iter);
		view->setPreferredSize();
		rect |= view->getFrame();
		iter++;
	}

	setFrame(rect);
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Container::onPaint(Context* context, const Rectangle<float>& rect) {
#if defined(__FORWINDOWS__)
	return true;
#else
	Rectangle<float> localRect = convertFromParent(rect);
	Chain<View>::Iterator iter = m_Views.begin();
	View* view;
//	paintBeforeChildren();

	// Draw all children
	while (iter != m_Views.end()) {
		view = &(*iter);
		if (localRect.intersects(view->getFrame())) {
			view->onPaint(context, localRect);
		}
		iter++;
	}

	// Call the paint function of View for initialisation of the Context
	View::onPaint(context, rect);

//	paintAfterChildren();

	return true;
#endif	
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Container::onResize(const Dimension<float>& ) {
	Chain<View>::Iterator iter = m_Views.begin();
	View* view;
	float dw, dh;

	dw = getFrame().getWidth() - m_OldFrame.getWidth();
	dh = getFrame().getHeight() - m_OldFrame.getHeight();

	if ((dw == 0) && (dh == 0)) {
		m_OldFrame = m_Frame;
		return true;
	}

	// Now resize all childs (if needed)
	while (iter != m_Views.end()) {
		view = &(*iter);
		float x,y,w,h;

		if (view->m_Alignement == 0) {
			iter++;
			continue;
		}

		if ((view->m_Alignement & Container::FollowLeftRight) == Container::FollowLeftRight) {
			x = view->getFrame().getX();
			w = view->getFrame().getWidth() + dw;
		} else if (view->m_Alignement & Container::FollowLeft) {
			x = view->getFrame().getX();
			w = view->getFrame().getWidth();
		} else if (view->m_Alignement & Container::FollowRight) {
			x = view->getFrame().getX() + dw;
			w = view->getFrame().getWidth();
		} else {
			x = view->getFrame().getX();
			w = view->getFrame().getWidth();
		}

		if ((view->m_Alignement & Container::FollowTopBottom) == Container::FollowTopBottom) {
			y = view->getFrame().getY();
			h = view->getFrame().getHeight() + dh;
		} else if (view->m_Alignement & Container::FollowTop) {
			y = view->getFrame().getY();
			h = view->getFrame().getHeight();
		} else if (view->m_Alignement & Container::FollowBottom) {
			y = view->getFrame().getY() + dh;
			h = view->getFrame().getHeight();
		} else {
			y = view->getFrame().getY();
			h = view->getFrame().getHeight();
		}

		view->setFrame(Rectangle<float>(x, y, w, h));
		iter++;
	}

	m_OldFrame = m_Frame;

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Container::onActivateWindow() {
	// attach all children to window
	Chain<View>::Iterator iter = m_Views.begin();

	while (iter != m_Views.end()) {
		iter->onActivateWindow();
		iter++;
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Container::onDeactivateWindow() {
	// attach all children to window
	Chain<View>::Iterator iter = m_Views.begin();

	while (iter != m_Views.end()) {
		iter->onDeactivateWindow();
		iter++;
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Container::onAttachWindow() {
	View::onAttachWindow();
	
	Chain<View>::Iterator iter = m_Views.begin();

	while (iter != m_Views.end()) {
		iter->m_Window = getWindow();
		iter->onAttachWindow();
		iter++;
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Container::onDetachWindow() {
	Chain<View>::Iterator iter = m_Views.begin();

	while (iter != m_Views.end()) {
		iter->onDetachWindow();
		iter->m_Window = NULL;
		iter++;
	}

	View::onDetachWindow();
	
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Container::updateTransform() {
	Chain<View>::Iterator iter = m_Views.begin();

	View::updateTransform();

	while (iter != m_Views.end()) {
		iter->updateTransform();
		iter++;
	}
}


}; };
