/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "gui.view.h"
#include "gui.container.h"
#include "gui.window.h"
#include "gui.context.h"
#include "gui.bitmap.h"
#include "graphics.dimension.h"
#include "graphics.rectangle.h"
#include "container.chain.h"
#include "system.thread.h"
#include "application.message.h"
#include "application.keyboard.h"

#include "gui.view.inl"
#include "gui.window.inl"
#include "gui.context.inl"
#include "graphics.point.inl"
#include "graphics.dimension.inl"
#include "graphics.rectangle.inl"
#include "graphics.colour.inl"
#include "graphics.transform.inl"
#include "container.chain.inl"
#include "system.thread.inl"
#include "system.signal.inl"
#include "system.inl"
#include "base.string.inl"
#include "object.inl"
#include "object.class.inl"
#include "application.looper.inl"
#include "application.message.inl"
#include "application.handler.inl"
#include "container.iterator.inl"
#include "container.chain.inl"



namespace magnum { namespace gui {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
View::View(const Rectangle<float>& frame) {
	m_Window = NULL;
	m_Parent = NULL;
	m_Frame = frame;
	m_Transform.setIdentity();
	m_Bounds.setOrigin(Point<float>(0,0));
	m_Bounds.setSize(Dimension<float>(m_Frame.getWidth(),m_Frame.getHeight()));
	updateTransform();
#if defined(__FORWINDOWS__)	
	m_WindowHandle = NULL;
	m_DeviceHandle = NULL;
	m_MouseTracking = false;
#endif	
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
View::~View() {
	if (m_Window != NULL) {
		m_Window->remove(this);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Dimension<float> View::getPreferredSize() const {
	return m_Frame.getSize();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool View::setPreferredSize(void) {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void View::setBounds(const Rectangle<float>& bounds) {
	m_Bounds = bounds;
	m_Transform.setTranslation(m_Bounds.getOrigin());
	m_Transform.setScale(m_Frame.getWidth() / m_Bounds.getWidth(), m_Frame.getHeight() / m_Bounds.getHeight());
	updateTransform();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void View::setScale(const Point<float>& Scale) {
	m_Transform.setScale(Scale);
	m_Bounds = convertFromParent(m_Frame);
	updateTransform();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void View::setOrigin(const Point<float>& origin) {
	m_Transform.setTranslation(origin);
	m_Bounds = convertFromParent(m_Frame);
	updateTransform();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool View::invalidate(const Rectangle<float>& rect) {
	if (m_Window == NULL)
		return false;
		
#if defined(__FORWINDOWS__)
	RECT r;
	r.left = (int)rect.getX();
	r.top = (int)rect.getY(),
	r.right = (int)rect.getX() + (int)rect.getWidth() - 1;
	r.bottom = (int)rect.getY() + (int)rect.getHeight() - 1;
	InvalidateRect(m_WindowHandle, &r, TRUE);
	return true;
#else
	Message msg(Window::MSG_INVALIDATE_VIEW);
	msg.addPointer("view",this);
	msg.addRectangle("area",rect);
	m_Window->postMessage(msg);	
	return true;
#endif	
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool View::invalidate(void) {
	if (m_Window == NULL)
		return false;

#if defined(__FORWINDOWS__)
	InvalidateRect(m_WindowHandle, NULL, TRUE);
	return true;
#else
	Message msg(Window::MSG_INVALIDATE_VIEW);
	msg.addPointer("view",this);
	msg.addRectangle("area",m_Frame);
	m_Window->postMessage(msg);	
	return true;
#endif
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool View::onGetFocus() {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool View::onReleaseFocus() {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool View::onPaint(Context* context, const Rectangle<float>& rect) {
#if defined(__FORWINDOWS__)
	return true;
#else
	Rectangle<float> clipRect;

	// Get clipping rectangle
	if (m_Parent != NULL) {
		clipRect = m_Parent->convertToWindow(rect) & (((View*)m_Parent)->m_AbsTransform * m_Frame);
	} else {
		clipRect = rect & m_Frame;
	}	

	// Initialize graphic context etc
	// Set new tranformation matrix
	context->setTransform(m_AbsTransform);
	context->setClipping(clipRect);

	return true;
#endif	
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool View::onResize(const Dimension<float>& ) {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool View::onKeyDown(Keyboard::Key ) {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool View::onKeyUp(Keyboard::Key ) {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool View::onChar(Character::Code ,Character::Modifier ) {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool View::onMouseMove(const Point<float>& ,Mouse::Button ) {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool View::onMouseButton(const Point<float>& ,Mouse::Button ) {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool View::onMouseEnter(const Point<float>& ,Mouse::Button ) {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool View::onMouseLeave() {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool View::onMouseWheel(float ,Mouse::Button ) {
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool View::onActivateWindow() {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool View::onDeactivateWindow() {
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void View::updateFrame() {
	m_Bounds = convertFromParent(m_Frame);

	if (m_Parent != NULL) {
		m_AbsTransform = ((View*)m_Parent)->m_AbsTransform * (m_Transform + m_Frame.getOrigin());
	} else {
		m_AbsTransform = m_Transform + m_Frame.getOrigin();
	}	

	if (m_Window == NULL) {
		onResize(m_Frame.getSize());
	} else if (!m_Window->isAlive()) {
		onResize(m_Frame.getSize());
#if defined(__FORWINDOWS__)
	} else {
		RECT rect;
  		rect.left = (int)m_Frame.getX();
  		rect.top = (int)m_Frame.getY();
  		rect.right = (int)m_Frame.getX() + (int)m_Frame.getWidth();
  		rect.bottom = (int)m_Frame.getY() + (int)m_Frame.getHeight();
		SetWindowPos(m_WindowHandle, NULL, rect.left, rect.top, rect.right - rect.left, rect.bottom - rect.top, SWP_NOZORDER | SWP_NOCOPYBITS);
	}
#else		
	} else if (Thread::getCurrent() == m_Window->getThread()) {
		onResize(m_Frame.getSize());
	} else {
		Message msg(Window::MSG_RESIZE_VIEW);
		msg.addPointer("view",this);
		m_Window->postMessage(msg);
	}
#endif	
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void View::updateTransform() {
	if (m_Parent != NULL) {
		m_AbsTransform = ((View*)m_Parent)->m_AbsTransform * (m_Transform + m_Frame.getOrigin());
	} else {
		m_AbsTransform = m_Transform + m_Frame.getOrigin();
	}	
}


#if defined(__FORX11__)
/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool View::dispatchEvent(const XEvent& event) {
	return true;
}
#endif

}; };
