/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "gui3d.container.h"
#include "gui3d.context.h"

#include "gui3d.container.inl"
#include "gui3d.context.inl"


namespace magnum { namespace gui3d {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Container::insert(View* view) {
	if (view->getParent() != NULL)
		return false;
	if (view->getRoot() != NULL)
		return false;
	if (view == this)
		return false;

	m_Views.insert(view);
	view->m_Parent = this;
	view->updateTransform();

	if (getRoot() != NULL) {
		view->m_Root = getRoot();
		view->m_Context = getContext();
		view->onAttach();
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Container::remove(View *view) {
	if (view->getParent() != this)
		return false;
	if (view->getRoot() != getRoot())
		return false;
	if (view == this)
		return false;

	if (getRoot() != NULL) {
		if (view->hasFocus())
			view->releaseFocus();
		view->onDetach();
	}

	m_Views.remove(view);
	view->m_Parent = NULL;
	view->m_Root = NULL;
	view->m_Context = NULL;
	view->updateTransform();

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Container::clear() {
	View *view;

	while(m_Views.begin() != m_Views.end()) {
		view = &(*m_Views.begin());

		if (getRoot() != NULL) {
			if (view->hasFocus())
				view->releaseFocus();
			view->onDetach();
		}

		m_Views.remove(view);
		view->m_Parent = NULL;
		view->m_Root = NULL;
		view->m_Context = NULL;
		view->updateTransform();
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Container::onEnable() {
	Chain<View>::Iterator iter = m_Views.begin();

	while (iter != m_Views.end()) {
		iter->enable();
		iter++;
	}
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Container::onDisable() {
	Chain<View>::Iterator iter = m_Views.begin();

	while (iter != m_Views.end()) {
		iter->disable();
		iter++;
	}
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Dimension<float> Container::onGetPreferredSize() const {
	Chain<View>::Iterator iter = m_Views.begin();
	View* view;
	Rectangle<float> rect = m_Frame;
	Rectangle<float> vrect;

	while (iter != m_Views.end()) {
		view = &(*iter);
		vrect.setOrigin(view->getFrame().getOrigin());
		vrect.setSize(view->onGetPreferredSize());
		rect |= vrect;
		iter++;
	}

	return rect.getSize();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Container::onSetPreferredSize(void) {
	Chain<View>::Iterator iter = m_Views.begin();
	View* view;
	Rectangle<float> rect = m_Frame;

	while (iter != m_Views.end()) {
		view = &(*iter);
		view->onSetPreferredSize();
		rect |= view->getFrame();
		iter++;
	}

	setFrame(rect);
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Container::onPaint() {
	Chain<View>::Iterator iter = m_Views.begin();

	while (iter != m_Views.end()) {
		getContext()->setTransform(iter->m_AbsTransform);
		iter->onPaint();
		iter++;
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Container::onResize(const Dimension<float>& ) {
	Chain<View>::Iterator iter = m_Views.begin();
	View* view;
	float dw, dh;

	dw = getFrame().getWidth() - m_OldFrame.getWidth();
	dh = getFrame().getHeight() - m_OldFrame.getHeight();

	if ((dw == 0) && (dh == 0)) {
		m_OldFrame = m_Frame;
		return true;
	}

	// Now resize all childs (if needed)
	while (iter != m_Views.end()) {
		view = &(*iter);
		float x,y,w,h;

		if (view->m_Alignement == 0) {
			iter++;
			continue;
		}

		if ((view->m_Alignement & Container::FollowLeftRight) == Container::FollowLeftRight) {
			x = view->getFrame().getX();
			w = view->getFrame().getWidth() + dw;
		} else if (view->m_Alignement & Container::FollowLeft) {
			x = view->getFrame().getX();
			w = view->getFrame().getWidth();
		} else if (view->m_Alignement & Container::FollowRight) {
			x = view->getFrame().getX() + dw;
			w = view->getFrame().getWidth();
		} else {
			x = view->getFrame().getX();
			w = view->getFrame().getWidth();
		}

		if ((view->m_Alignement & Container::FollowTopBottom) == Container::FollowTopBottom) {
			y = view->getFrame().getY();
			h = view->getFrame().getHeight() + dh;
		} else if (view->m_Alignement & Container::FollowTop) {
			y = view->getFrame().getY();
			h = view->getFrame().getHeight();
		} else if (view->m_Alignement & Container::FollowBottom) {
			y = view->getFrame().getY() + dh;
			h = view->getFrame().getHeight();
		} else {
			y = view->getFrame().getY();
			h = view->getFrame().getHeight();
		}

		view->setFrame(Rectangle<float>(x, y, w, h));
		iter++;
	}

	m_OldFrame = m_Frame;

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Container::onMouseMove(const Point<float>& pos, Mouse::Button button) {
	Chain<View>::Iterator iter = m_Views.begin();
	View* view;

	for (;iter != m_Views.end(); ++iter) {
		view = &(*iter);
		if (view->getFrame().isInside(pos)) {
			if (!view->isEnabled())
				continue;
			if (view->onMouseMove(view->convertFromParent(pos), button))
				return true;
		}
	}

	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Container::onMouseButton(const Point<float>& pos, Mouse::Button button) {
	Chain<View>::Iterator iter = m_Views.begin();
	View* view;

	for (;iter != m_Views.end(); ++iter) {
		view = &(*iter);
		if (view->getFrame().isInside(pos)) {
			if (!view->isEnabled())
				continue;
			if (view->onMouseButton(view->convertFromParent(pos), button))
				return true;
		}
	}

	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Container::onMouseEnter(const Point<float>& pos, Mouse::Button button) {
	Chain<View>::Iterator iter = m_Views.begin();
	View* view;

	for (;iter != m_Views.end(); ++iter) {
		view = &(*iter);
		if (view->getFrame().isInside(pos)) {
			if (!view->isEnabled())
				continue;
			if (view->onMouseEnter(view->convertFromParent(pos), button))
				return true;
		}
	}

	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Container::onActivate() {
	// attach all children to window
	Chain<View>::Iterator iter = m_Views.begin();

	while (iter != m_Views.end()) {
		iter->onActivate();
		iter++;
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Container::onDeactivate() {
	// attach all children to window
	Chain<View>::Iterator iter = m_Views.begin();

	while (iter != m_Views.end()) {
		iter->onDeactivate();
		iter++;
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Container::onAttach() {
	View::onAttach();
	
	Chain<View>::Iterator iter = m_Views.begin();

	while (iter != m_Views.end()) {
		iter->m_Root = getRoot();
		iter->m_Context = getContext();
		iter->onAttach();
		iter++;
	}

	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool Container::onDetach() {
	Chain<View>::Iterator iter = m_Views.begin();

	while (iter != m_Views.end()) {
		iter->onDetach();
		iter->m_Root = NULL;
		iter->m_Context = NULL;
		iter++;
	}

	View::onDetach();
	
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Container::updateTransform() {
	Chain<View>::Iterator iter = m_Views.begin();

	View::updateTransform();

	while (iter != m_Views.end()) {
		iter->updateTransform();
		iter++;
	}
}


}; };
