/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "base.assert.h"
#include "io.sysfile.h"
#include "io.exception.h"

#include "system.timestamp.inl"
#include "io.sysfile.inl"
#include "object.exception.inl"

namespace magnum { namespace io {
using magnum::types::uint64;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
SystemFile::SystemFile() : File() {
	m_FileHandle = NULL;
	m_Writeable = m_Readable = false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
SystemFile::~SystemFile() {
	if (m_FileHandle ==NULL )
		return;

	if (fclose(m_FileHandle) != 0)
		throw IOException();		

	m_FileHandle = NULL;
	m_Writeable = m_Readable = false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned long SystemFile::write(const void *buf, unsigned long size) {
	ASSERT((m_FileHandle!=NULL) && (buf!=NULL));
	if (!m_Writeable)
		return 0;

	size_t written;
	written = fwrite(buf, size, 1, m_FileHandle);
	return written == 1 ? size : 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned long SystemFile::read(void *buf, unsigned long size) {
	ASSERT((m_FileHandle!=NULL) && (buf!=NULL));

	if (!m_Readable)
		return 0;

	size_t read;
	read = fread(buf, size, 1, m_FileHandle);
	return read == 1 ? size : 0;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool SystemFile::flush() {
	ASSERT(m_FileHandle!=NULL);

	return (fflush(m_FileHandle) == 0);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned long SystemFile::setPosition(long where, POSITION_MODE mode) {
	ASSERT(m_FileHandle!=NULL);

	int method, val;
	switch (mode) {
		case SetAbs:
			method = SEEK_SET;
			break;
		case SetCur:
			method = SEEK_CUR;
			break;
		case SetEnd:
			method = SEEK_END;
			break;
		default:
			throw IllegalPositionModeException;
	}
	val = fseek(m_FileHandle, where, method);
	if (val != 0) 
		throw FileSeekException();
	return ftell(m_FileHandle);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool SystemFile::rewind() {
	ASSERT(m_FileHandle!=NULL);

	if (fseek(m_FileHandle, 0, SEEK_SET) != 0)
		throw FileSeekException();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned long SystemFile::getPosition() {
	ASSERT(m_FileHandle!=NULL);

	return ftell(m_FileHandle);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned long SystemFile::getSize() {
	ASSERT(m_FileHandle!=NULL);

	unsigned long cur, size;
	cur = ftell(m_FileHandle);
	fseek(m_FileHandle, 0, SEEK_END);
	size = ftell(m_FileHandle);
	fseek(m_FileHandle, cur, SEEK_SET);
	return size;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned long SystemFile::setSize(unsigned long newSize) {
	ASSERT(m_FileHandle!=NULL);

	if (!m_Writeable)	
		return 0;

	return (unsigned long)-1;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool SystemFile::isWriteable() {
	ASSERT(m_FileHandle!=NULL);
	return m_Writeable;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool SystemFile::isReadable() {
	ASSERT(m_FileHandle!=NULL);
	return m_Readable;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Timestamp SystemFile::getTimestamp() {
	ASSERT(m_FileHandle!=NULL);

	return Timestamp(0);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
SystemFile* SystemFile::open(const char *fname, OPEN_FLAGS flags) {
	const char* sysFlags;

	if (flags & WriteOnly)
		sysFlags = "r+b";
	else if (flags & ReadOnly)
		sysFlags = "rb";

	FILE* handle = fopen(fname, sysFlags);
	if (handle == NULL) 
		throw FileOpenException();

	SystemFile* file = new SystemFile();
	file->m_FileHandle = handle;
	file->m_Readable = false;
	file->m_Writeable = false;

	if (flags & ReadOnly)
		file->m_Readable = true;
	if (flags & WriteOnly)
		file->m_Writeable = true;

	return file;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
SystemFile* SystemFile::create(const char *fname) {
	FILE* handle = fopen(fname, "w+b");
	if (handle == NULL) 
		throw FileOpenException();

	SystemFile* file =new SystemFile();
	file->m_FileHandle = handle;
	file->m_Readable = true;
	file->m_Writeable = true;
	return file;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool SystemFile::remove(const char *fname) {
	return false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void SystemFile::throwException() {
	throw IOException();
}

}; };
