/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "base.assert.h"
#include "io.sysfile.h"
#include "io.exception.h"

#include "object.inl"
#include "object.exception.inl"
#include "base.string.inl"
#include "system.timestamp.inl"
#include "io.file.inl"
#include "io.sysfile.inl"
#include "io.inputstream.inl"
#include "io.outputstream.inl"

#ifndef __FORWINDOWS__
#error "This file is only for Windows"
#endif


namespace magnum { namespace io {
using magnum::types::uint64;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
SystemFile::SystemFile() : File() {
	m_FileHandle = NULL;
	m_Writeable = m_Readable = false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
SystemFile::~SystemFile() {
	if (m_FileHandle ==NULL )
		return;

	if (CloseHandle(m_FileHandle) == 0)
		throwException();

	m_FileHandle = NULL;
	m_Writeable = m_Readable = false;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned long SystemFile::write(const void *buf, unsigned long size) {
	ASSERT((m_FileHandle!=NULL) && (buf!=NULL));
	if (!m_Writeable)
		return 0;

	DWORD written;
	WriteFile(m_FileHandle, buf, size, &written, NULL);
	return written;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned long SystemFile::read(void *buf, unsigned long size) {
	ASSERT((m_FileHandle!=NULL) && (buf!=NULL));

	if (!m_Readable)
		return 0;

	DWORD read;
	ReadFile(m_FileHandle, buf, size, &read, NULL);
	return read;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool SystemFile::flush() {
	ASSERT(m_FileHandle!=NULL);

	return (FlushFileBuffers(m_FileHandle) == TRUE);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned long SystemFile::setPosition(long where, PositionMode mode) {
	ASSERT(m_FileHandle!=NULL);

	DWORD method, ret;
	switch (mode) {
		case Absolute:
			method = FILE_BEGIN;
			break;
		case Current:
			method = FILE_CURRENT;
			break;
		case End:
			method = FILE_END;
			break;
		default:
			throw IllegalPositionModeException();
	}
	ret = SetFilePointer(m_FileHandle, where, NULL, method);
	if (ret == 0xffffffff)
		throwException();
	return ret;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool SystemFile::rewind() {
	ASSERT(m_FileHandle!=NULL);

	if (SetFilePointer(m_FileHandle, 0, NULL, FILE_BEGIN) != 0)
		throwException();
	return true;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned long SystemFile::getPosition() const {
	ASSERT(m_FileHandle!=NULL);

	return SetFilePointer(m_FileHandle, 0, NULL, FILE_CURRENT);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned long SystemFile::getSize() const {
	ASSERT(m_FileHandle!=NULL);

	unsigned long cur, size;
	cur = SetFilePointer(m_FileHandle, 0, NULL, FILE_CURRENT);
	size = SetFilePointer(m_FileHandle, 0, NULL, FILE_END);
	SetFilePointer(m_FileHandle, cur, NULL, FILE_BEGIN);
	return size;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
unsigned long SystemFile::setSize(unsigned long newSize) {
	ASSERT(m_FileHandle!=NULL);

	if (!m_Writeable)	
		return 0;

	unsigned long cur;
	cur = SetFilePointer(m_FileHandle, 0, NULL, FILE_CURRENT);
	SetFilePointer(m_FileHandle, newSize, NULL, FILE_BEGIN);
	SetEndOfFile(m_FileHandle);
	SetFilePointer(m_FileHandle, cur, NULL, FILE_BEGIN);
	return newSize;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool SystemFile::isWriteable() const {
	ASSERT(m_FileHandle!=NULL);
	return m_Writeable;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool SystemFile::isReadable() const {
	ASSERT(m_FileHandle!=NULL);
	return m_Readable;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Timestamp SystemFile::getCreationTime() const {
	ASSERT(m_FileHandle!=NULL);

	FILETIME ftime;
	GetFileTime(m_FileHandle,&ftime,NULL,NULL);
	return Timestamp((((uint64)ftime.dwHighDateTime) << 32) + ftime.dwLowDateTime);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Timestamp SystemFile::getModificationTime() const {
	ASSERT(m_FileHandle!=NULL);

	FILETIME ftime;
	GetFileTime(m_FileHandle,NULL,NULL,&ftime);
	return Timestamp((((uint64)ftime.dwHighDateTime) << 32) + ftime.dwLowDateTime);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Timestamp SystemFile::getAccessTime() const {
	ASSERT(m_FileHandle!=NULL);

	FILETIME ftime;
	GetFileTime(m_FileHandle,NULL,&ftime,NULL);
	return Timestamp((((uint64)ftime.dwHighDateTime) << 32) + ftime.dwLowDateTime);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
SystemFile* SystemFile::open(const char *fname, OpenFlags flags) {
	DWORD sysFlags = 0;

	if (flags & WriteOnly)
		sysFlags |= GENERIC_WRITE;
	if (flags & ReadOnly)
		sysFlags |= GENERIC_READ;

	HANDLE handle = CreateFile(fname, sysFlags, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (handle == INVALID_HANDLE_VALUE) 
		throwException();

	SystemFile* file = new SystemFile();
	file->m_FileHandle = handle;
	file->m_Readable = false;
	file->m_Writeable = false;

	if (flags & ReadOnly)
		file->m_Readable = true;
	if (flags & WriteOnly)
		file->m_Writeable = true;

	return file;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
SystemFile* SystemFile::create(const char *fname) {
	HANDLE handle = CreateFile(fname, GENERIC_READ | GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (handle == INVALID_HANDLE_VALUE)
		throwException();

	SystemFile* file =new SystemFile();
	file->m_FileHandle = handle;
	file->m_Readable = true;
	file->m_Writeable = true;
	return file;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
bool SystemFile::remove(const char *fname) {
	return (DeleteFile(fname)==0);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void SystemFile::throwException() {
	DWORD err = GetLastError();
	switch(err) {
		case ERROR_READ_FAULT:
			throw FileReadException("The system cannot read from the specified device.");
		case ERROR_WRITE_FAULT:
			throw FileWriteException("The system cannot write to the specified device.");
		case ERROR_WRITE_PROTECT:
			throw FileWriteException("The media is write protected.");
		case ERROR_WRONG_DISK:
			throw IOException("The wrong diskette is in the drive.");
		case ERROR_UNRECOGNIZED_VOLUME:
			throw IOException("The volume does not contain a recognized file system.");
		case ERROR_UNRECOGNIZED_MEDIA:
			throw IOException("The disk media is not recognized. It may not be formatted.");
		case ERROR_TOO_MANY_OPEN_FILES:
			throw FileOpenException("The system cannot open the file.");
		case ERROR_OPEN_FAILED:
			throw FileOpenException("The system cannot open the device or file specified.");
		case ERROR_SHARING_VIOLATION:
			throw IOException("The process cannot access the file because it is being used by another process.");
		case ERROR_SEEK:
			throw IOException("The drive cannot locate a specific area or track on the disk.");
		case ERROR_SEEK_ON_DEVICE:
			throw FileSeekException("The file pointer cannot be set on the specified device or file.");
		case ERROR_FILE_NOT_FOUND:
			throw FileNotFoundException("The system cannot find the file specified.");
		case ERROR_FILE_EXISTS:
			throw FileExistsException("The file exists.");
		case ERROR_FILE_INVALID:
			throw IOException("The volume for a file has been externally altered so that the opened file is no longer valid.");
		case ERROR_DISK_FULL:
			throw DiskFullException("There is not enough space on the disk.");
		default:
			throw IOException();
	}
}

}; };
