/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "scene3d.enumerator.h"
#include "scene3d.universe.h"
#include "scene3d.locale.h"
#include "scene3d.root.h"
#include "scene3d.link.h"
#include "scene3d.sharedgroup.h"
#include "scene3d.transformgroup.h"

#include "object.pointer.inl"
#include "math.vector3.inl"
#include "container.iterator.inl"
#include "container.stack.inl"
#include "container.chain.inl"
#include "geometry.transform.inl"
#include "scene3d.enumerator.inl"
#include "scene3d.root.inl"
#include "scene3d.node.inl"
#include "scene3d.locale.inl"
#include "scene3d.link.inl"
#include "scene3d.universe.inl"
#include "scene3d.nodepath.inl"
#include "scene3d.reference.inl"
#include "scene3d.transformgroup.inl"


namespace magnum { namespace scene3d {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Enumerator::Enumerator(unsigned int t) {
	m_Universe = NULL;
	m_Locale = NULL;
	m_Types = (Node::Type)t;
	m_BasePosition = Vector3<double>(0.0, 0.0, 0.0);
	m_BaseTransform.identity();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Enumerator::Enumerator(const Universe* u,unsigned int t) {
	m_Types = (Node::Type)t;
	m_BasePosition = Vector3<double>(0.0, 0.0, 0.0);
	m_BaseTransform.identity();
	attachUniverse(u);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Enumerator::Enumerator(const Locale* l,unsigned int t) {
	m_Types = (Node::Type)t;
	m_BasePosition = Vector3<double>(0.0, 0.0, 0.0);
	m_BaseTransform.identity();
	attachLocale(l);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Enumerator::begin() {
	Transform3<float> t;

	if (m_Locale == NULL) {
		if (m_Universe == NULL)
			goto foundNone;
		m_LocaleIterator = m_Universe->begin();
		if (m_LocaleIterator == m_Universe->end())
			goto foundNone;
		m_CurrentLocale = m_LocaleIterator;
	} else {
		m_CurrentLocale = (Locale*)m_Locale;
	}

	// FInd first non-empty locale
	for (;;) {
		m_RootIterator = m_CurrentLocale->begin();
		if (m_RootIterator != m_CurrentLocale->end())
			break;

		// Next Locale
		if (m_Universe != NULL) {
			m_LocaleIterator++;
			if (m_LocaleIterator == m_Universe->end())
				goto foundNone;
			m_CurrentLocale = m_LocaleIterator;
		} else {
			goto foundNone;
		}
	}
	m_CurrentRoot = m_RootIterator;

	// We found a node
	m_CurrentNode = m_CurrentRoot;

	// Calculate transformation
	t = m_BaseTransform;
	t.translate((Vector3<float>)(m_BasePosition + m_CurrentLocale->getOrigin()));
	m_Path.push(m_CurrentNode, t);

	// Has node the correct type?
	if (!(m_CurrentNode->getType() & m_Types)) {
		next();
		return;
	}

	return;


foundNone:
	m_CurrentLocale = NULL;
	m_CurrentRoot = NULL;
	m_CurrentNode = NULL;
	return;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Enumerator::next() {
	ASSERT(m_CurrentNode != NULL);
	Transform3<float> t;
	Node* current = m_CurrentNode;

	for(;;) {
		// Try one down
		if (current->getType() & Node::GroupType) {
			if (((Group*)current)->m_ChildTypes & m_Types) {
				current = ((Group*)current)->m_First;
				ASSERT(current != NULL);
				goto found;
			}
		} 

 		//Check for link
		else if (current->getType() & Node::LinkType) {
			if (((Link*)current)->getTarget() != NULL) {
				current = ((Link*)current)->getTarget();
				goto found;
			}
		}

		// Try next sibling
		m_Path.pop();
		if (current->m_Next != NULL) {
			current = current->m_Next;
			goto found;
		}

		// Try parents
		for(;;) {
			if (m_Path.isEmpty()) {
				goto foundNull;
			}
			current = m_Path.pop();
			if (current->m_Next != NULL) {
				current = current->m_Next;
				goto found;
			}
		}

foundNull:
		// Did we find another node in current tree
		if (m_CurrentLocale == NULL) {
			m_CurrentNode = NULL;
			return;
		}

		m_RootIterator++;
		while (m_RootIterator == m_CurrentLocale->end()) {
			if (m_Locale != NULL) {
				m_CurrentNode = NULL;
				return;
			}

			m_LocaleIterator++;
			if (m_LocaleIterator == m_Universe->end()) {
				m_CurrentNode = NULL;
				return;
			}

			m_CurrentLocale = m_LocaleIterator;
			m_RootIterator = m_CurrentLocale->begin();
		}
		m_CurrentRoot = m_RootIterator;
		current = m_CurrentRoot;

		// Get Transformation
		t = m_BaseTransform;
		t.translate((Vector3<float>)(m_BasePosition + m_CurrentLocale->getOrigin()));

		// Push new root onto stack
		m_Path.push(current, t);
		goto foundRoot;

found:
		// check for transformation and push node onto stack
		if (current->getType() & Node::TransformType) {
			m_Path.push(current, m_Path.peek().getTransform() * ((TransformGroup*)current)->getTransform());
		} else {
			m_Path.push(current, m_Path.peek().getTransform());
		}

foundRoot:
		// Is this node of correct type?
		if (current->getType() & m_Types) {
			m_CurrentNode = current;
			return;
		}
	}
}


}; };
