/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "scene3d.group.h"
#include "scene3d.locale.h"
#include "scene3d.universe.h"

#include "scene3d.node.inl"
#include "scene3d.group.inl"
#include "scene3d.locale.inl"


namespace magnum { namespace scene3d {

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Group::insert(Node* node) {
	ASSERT(node->m_Parent == NULL);
	if (m_First == NULL) {
		m_First = m_Last = node;
		node->m_Next = node->m_Prev = NULL;
	} else {
		node->m_Next = NULL;
		node->m_Prev = m_Last;
		m_Last->m_Next = node;
		m_Last = node;
	}
	node->m_Parent = this;
	node->refer();

	// Enable / Disable node
	if (!isEnabled() && node->isEnabled()) 
		node->disable();

	// Update types
	m_ChildTypes |= getAllTypes(node);

	//propagate changes to parents...
	Group* g = m_Parent;
	while (g) {
		if ((g->m_ChildTypes & m_ChildTypes) == m_ChildTypes)
			break;
		g->m_ChildTypes |= m_ChildTypes;
		g = g->m_Parent;
	}

	// Attach node to locale
	if (m_Locale != NULL)
		node->onAttachLocale(m_Locale);

	// Inform about insertion		
	onInsert(node);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Group::remove(Node* node) {
	ASSERT(node->m_Parent == this);
	
	// Inform about removal
	onRemove(node);

	// detach node from locale
	if (m_Locale != NULL)
		node->onDetachLocale(m_Locale);

	Node* cur = m_First;
	Node* prev = NULL;
	Node* next = NULL;
	int types = 0;
	int deltypes = 0;

	while(cur != NULL) {
		next = node->m_Next;
		if (cur == node) {
			deltypes = getAllTypes(node);
			if (next != NULL)
				next->m_Prev = prev;
			else
				m_Last = prev;

			if (prev != NULL)
				prev->m_Next = next;
			else
				m_First = next;

			cur->m_Next = cur->m_Prev = NULL;
			cur->m_Parent = NULL;
			cur->release();
			break;
		}

		// get types of subnode plus all children (if applicable);
		types |= getAllTypes(cur);

		// go to next node
		prev = node;
		cur = next;
	}

	ASSERT(cur != NULL);

	cur = next;
	while(cur != NULL) {
		types |= getAllTypes(cur);
		cur = cur->m_Next;
	}

	// Did the types really disappear from the group?
	if ((types & deltypes) == deltypes)
		return;

	// set types
	m_ChildTypes = types;

	// Propagate type changes to parents
	Group *g = m_Parent;
	while(g != NULL) {
		types = 0;
		node = g->m_First;
		while(node != NULL) {
			types |= getAllTypes(node);
			node = node->m_Next;
		}
		if ((types & deltypes) == deltypes)
			return;
		g->m_ChildTypes = types;
		g = g->m_Parent;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Group::remove(const Group::Iterator& iter) {
	remove((Node*)iter);
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Group::clear() {
	Node* next;
	while(m_First != NULL) {
		next = m_First->m_Next;
		
		// Inform about removal
		onRemove(m_First);
		
		// Detach from locale if applicabale
		if (m_Locale != NULL)
			m_First->onDetachLocale(m_Locale);
			
		// Remove node from chain
		m_First->m_Parent = NULL;
		m_First->m_Prev = m_First->m_Next = NULL;
		m_First->release();
		m_First = next;
	}

	m_First = m_Last = NULL;
	m_ChildTypes = 0;

	// Propagate type changes to parents
	Group *g = m_Parent;
	while(g != NULL) {
		int types = 0;
		Node* node = g->m_First;
		while(node != NULL) {
			types |= getAllTypes(node);
			node = node->m_Next;
		}
		g->m_ChildTypes = types;
		g = g->m_Parent;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Group::onAttachLocale(Locale* l) {
	Node::onAttachLocale(l);
	Node* node = m_First;
	while (node != NULL) {
		node->onAttachLocale(l);
		node = node->m_Next;
	}	
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Group::onDetachLocale(Locale* l) {
	Node::onDetachLocale(l);
	Node* node = m_First;
	while (node != NULL) {
		node->onDetachLocale(l);
		node = node->m_Next;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Group::onEnable() {
	Node* node = m_First;
	while (node != NULL) {
		node->onEnable();
		node = node->m_Next;
	}	
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Group::onDisable() {
	Node* node = m_First;
	while (node != NULL) {
		node->onDisable();
		node = node->m_Next;
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Group::onInsert(Node* ) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Group::onRemove(Node* ) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Group::onTransform(const Transform3<float>& parent) {
	m_WorldTransform = parent;
}


}; };
