/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "scene3d.observer.h"
#include "scene3d.universe.h"
#include "scene3d.geometry.h"
#include "scene3d.locale.h"
#include "scene3d.root.h"
#include "scene3d.behaviour.h"
#include "scene3d.cameranode.h"
#include "scene3d.listenernode.h"
#include "scene3d.soundnode.h"
#include "scene3d.activatornode.h"
#include "scene3d.modifier.h"
#include "scene3d.transformgroup.h"

#include "object.pointer.inl"
#include "container.iterator.inl"
#include "container.chain.inl"
#include "container.scheduler.inl"
#include "math.vector.inl"
#include "math.vector3.inl"
#include "math.matrix.inl"
#include "graphics3d.camera.inl"
#include "geometry.space.inl"
#include "geometry.sphere.inl"
#include "geometry.cube.inl"
#include "geometry.frustum.inl"
#include "geometry.traverser.inl"
#include "geometry.bounds.inl"
#include "geometry.transform.inl"
#include "scene3d.geometry.inl"
#include "scene3d.graphics.inl"
#include "scene3d.observer.inl"
#include "scene3d.universe.inl"
#include "scene3d.behaviour.inl"
#include "scene3d.scheduler.inl"
#include "scene3d.locale.inl"
#include "scene3d.root.inl"
#include "scene3d.transformgroup.inl"
#include "scene3d.node.inl"
#include "scene3d.traverser.inl"
#include "scene3d.listenernode.inl"
#include "scene3d.activatornode.inl"
#include "scene3d.cameranode.inl"
#include "scene3d.audioqueue.inl"
#include "scene3d.renderqueue.inl"
#include "scene3d.scheduler.inl"

namespace magnum { namespace scene3d {

#pragma warning(disable: 4355)

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Observer::Observer(Universe* u) : 
	m_Universe(u), 
	m_GraphicsTraverser(this),
	m_AudioTraverser(this),
	m_BehaviourTraverser(this),
	m_AudioQueue(this),
	m_RenderQueue(this),
	m_Scheduler(this) {
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Observer::~Observer() {
	m_Scheduler.clear();
	m_RenderQueue.clear();
	m_AudioQueue.clear();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Observer::view() {
	if (m_Camera == NULL)
		return;

	ASSERT(m_Camera->getUniverse() == m_Universe);

	// Select camera
	m_Camera->onSelect();
	m_GraphicsTraverser.traverse();
	m_RenderQueue.view();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Observer::listen() {
	if (m_Listener == NULL)
		return;

	ASSERT(m_Listener->getUniverse() == m_Universe);

	m_AudioTraverser.traverse();
	m_AudioQueue.listen();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Observer::process() {
	if (m_Universe == NULL)
		return;

	m_BehaviourTraverser.traverse();
	m_Scheduler.process();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Observer::GraphicsTraverser::GraphicsTraverser(Observer* o) 
	: Traverser<GraphicsTraverser, Graphics>(o->getUniverse()) {
	m_Observer = o;
	m_RenderQueue = &o->m_RenderQueue;
	m_Camera = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Observer::GraphicsTraverser::traverse() {
	ASSERT(m_Observer != NULL);

	m_Camera = m_Observer->getCamera();

	// Check if we have a camera
	if (m_Camera != NULL) {
		ASSERT(m_Camera->getUniverse() == getUniverse());

		// Get Camera transformation
		setPreTransform(Transform3<float>(m_Camera->getWorldTransform()).invert());
		setPreTranslation(-m_Camera->getLocale()->getOrigin());
		m_Frustum = m_Camera->getFrustum();
	} else {
		setPreTransform(Transform3<float>::Identity);
		setPreTranslation(Vector3<double>::Null);
	}

	Traverser<GraphicsTraverser, Graphics>::traverse(m_Frustum, -getPreTranslation());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Observer::AudioTraverser::AudioTraverser(Observer* o) 
	: Traverser<AudioTraverser, Audio>(o->getUniverse()) {
	m_Observer = o;
	m_AudioQueue = &o->m_AudioQueue;
	m_Listener = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Observer::AudioTraverser::traverse() {
	ASSERT(m_Observer != NULL);

	m_Listener = m_Observer->getListener();

	// Check if we have a camera
	if (m_Listener != NULL) {
		ASSERT(m_Listener->getUniverse() == getUniverse());

		// Get Listener transformation
		setPreTransform(Transform3<float>(m_Listener->getWorldTransform()).invert());
		setPreTranslation(-m_Listener->getLocale()->getOrigin());
		m_Position = m_Listener->getWorldPosition();
	} else {
		setPreTransform(Transform3<float>::Identity);
		setPreTranslation(Vector3<double>::Null);
	}

	Traverser<AudioTraverser, Audio>::traverse(m_Position, -getPreTranslation());
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
Observer::BehaviourTraverser::BehaviourTraverser(Observer* o) 
	: Traverser<BehaviourTraverser, Behaviour>(o->getUniverse()) {
	m_Observer = o;
	m_Scheduler = &o->m_Scheduler;
	m_Activator = NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void Observer::BehaviourTraverser::traverse() {
	ASSERT(m_Observer != NULL);

	m_Activator = m_Observer->getActivator();

	// Check if we have a camera
	if (m_Activator != NULL) {
		ASSERT(m_Activator->getUniverse() == getUniverse());

		// Get Activator transformation
		setPreTransform(Transform3<float>(m_Activator->getWorldTransform()).invert());
		setPreTranslation(-m_Activator->getLocale()->getOrigin());
		m_Frustum = m_Activator->getFrustum();
	} else {
		setPreTransform(Transform3<float>::Identity);
		setPreTranslation(Vector3<double>::Null);
	}

	Traverser<BehaviourTraverser, Behaviour>::traverse(m_Frustum, -getPreTranslation());
}

}; };
